@@ -16,6 +16,7 @@ typedef THREAD_ENTRY ThreadEntry(LPVOID);
 typedef HANDLE Thread;
 typedef CRITICAL_SECTION Mutex;
 typedef CONDITION_VARIABLE Condition;
+typedef DWORD ThreadLocal;
 
 static inline int MutexInit(Mutex* mutex) {
 	InitializeCriticalSection(mutex);
@@ -88,4 +89,16 @@ static inline int ThreadSetName(const char* name) {
 	return -1;
 }
 
+static inline void ThreadLocalInitKey(ThreadLocal* key) {
+	*key = TlsAlloc();
+}
+
+static inline void ThreadLocalSetKey(ThreadLocal key, void* value) {
+	TlsSetValue(key, value);
+}
+
+static inline void* ThreadLocalGetValue(ThreadLocal key) {
+	return TlsGetValue(key);
+}
+
 #endif
