@@ -19,6 +19,10 @@ char* strndup(const char* start, size_t len);
 char* strdup(const char* str);
 #endif
 
+#ifndef HAVE_STRLCPY
+size_t strlcpy(char* restrict dst, const char* restrict src, size_t dstsize);
+#endif
+
 char* strnrstr(const char* restrict s1, const char* restrict s2, size_t len);
 bool endswith(const char* restrict s1, const char* restrict end);
 bool startswith(const char* restrict s1, const char* restrict start);
@@ -28,6 +32,7 @@ int utfcmp(const uint16_t* utf16, const char* utf8, size_t utf16Length, size_t u
 char* utf16to8(const uint16_t* utf16, size_t length);
 uint32_t utf8Char(const char** unicode, size_t* length);
 uint32_t utf16Char(const uint16_t** unicode, size_t* length);
+char* gbkToUtf8(const char* gbk, size_t length);
 
 int hexDigit(char digit);
 const char* hex32(const char* line, uint32_t* out);
@@ -39,6 +44,9 @@ const char* hex4(const char* line, uint8_t* out);
 
 void rtrim(char* string);
 
+ssize_t parseQuotedString(const char* unparsed, ssize_t unparsedLen, char* parsed, ssize_t parsedLen);
+bool wildcard(const char* search, const char* string);
+
 CXX_GUARD_END
 
 #endif
