@@ -36,6 +36,32 @@ typedef uint32_t color_t;
 #define M_RGB8_TO_BGR5(X) ((((X) & 0xF8) >> 3) | (((X) & 0xF800) >> 6) | (((X) & 0xF80000) >> 9))
 #define M_RGB8_TO_RGB5(X) ((((X) & 0xF8) << 7) | (((X) & 0xF800) >> 6) | (((X) & 0xF80000) >> 19))
 
+#ifndef COLOR_16_BIT
+#define M_COLOR_RED   0x000000FF
+#define M_COLOR_GREEN 0x0000FF00
+#define M_COLOR_BLUE  0x00FF0000
+#define M_COLOR_ALPHA 0xFF000000
+#define M_COLOR_WHITE 0x00FFFFFF
+
+#define M_RGB8_TO_NATIVE(X) (((X) & 0x00FF00) | (((X) & 0x0000FF) << 16) | (((X) & 0xFF0000) >> 16))
+#elif defined(COLOR_5_6_5)
+#define M_COLOR_RED   0x001F
+#define M_COLOR_GREEN 0x07E0
+#define M_COLOR_BLUE  0xF800
+#define M_COLOR_ALPHA 0x0000
+#define M_COLOR_WHITE 0xFFDF
+
+#define M_RGB8_TO_NATIVE(X) ((((X) & 0xF8) << 8) | (((X) & 0xFC00) >> 5) | (((X) & 0xF80000) >> 19))
+#else
+#define M_COLOR_RED   0x001F
+#define M_COLOR_GREEN 0x03E0
+#define M_COLOR_BLUE  0x7C00
+#define M_COLOR_ALPHA 0x1000
+#define M_COLOR_WHITE 0x7FFF
+
+#define M_RGB8_TO_NATIVE(X) M_RGB8_TO_BGR5(X)
+#endif
+
 #ifndef PYCPARSE
 static inline color_t mColorFrom555(uint16_t value) {
 #ifdef COLOR_16_BIT
@@ -53,6 +79,68 @@ static inline color_t mColorFrom555(uint16_t value) {
 #endif
 	return color;
 }
+
+ATTRIBUTE_UNUSED static unsigned mColorMix5Bit(int weightA, unsigned colorA, int weightB, unsigned colorB) {
+	unsigned c = 0;
+	unsigned a, b;
+#ifdef COLOR_16_BIT
+#ifdef COLOR_5_6_5
+	a = colorA & 0xF81F;
+	b = colorB & 0xF81F;
+	a |= (colorA & 0x7C0) << 16;
+	b |= (colorB & 0x7C0) << 16;
+	c = ((a * weightA + b * weightB) / 16);
+	if (c & 0x08000000) {
+		c = (c & ~0x0FC00000) | 0x07C00000;
+	}
+	if (c & 0x0020) {
+		c = (c & ~0x003F) | 0x001F;
+	}
+	if (c & 0x10000) {
+		c = (c & ~0x1F800) | 0xF800;
+	}
+	c = (c & 0xF81F) | ((c >> 16) & 0x07C0);
+#else
+	a = colorA & 0x7C1F;
+	b = colorB & 0x7C1F;
+	a |= (colorA & 0x3E0) << 16;
+	b |= (colorB & 0x3E0) << 16;
+	c = ((a * weightA + b * weightB) / 16);
+	if (c & 0x04000000) {
+		c = (c & ~0x07E00000) | 0x03E00000;
+	}
+	if (c & 0x0020) {
+		c = (c & ~0x003F) | 0x001F;
+	}
+	if (c & 0x8000) {
+		c = (c & ~0xF800) | 0x7C00;
+	}
+	c = (c & 0x7C1F) | ((c >> 16) & 0x03E0);
+#endif
+#else
+	a = colorA & 0xFF;
+	b = colorB & 0xFF;
+	c |= ((a * weightA + b * weightB) / 16) & 0x1FF;
+	if (c & 0x00000100) {
+		c = 0x000000FF;
+	}
+
+	a = colorA & 0xFF00;
+	b = colorB & 0xFF00;
+	c |= ((a * weightA + b * weightB) / 16) & 0x1FF00;
+	if (c & 0x00010000) {
+		c = (c & 0x000000FF) | 0x0000FF00;
+	}
+
+	a = colorA & 0xFF0000;
+	b = colorB & 0xFF0000;
+	c |= ((a * weightA + b * weightB) / 16) & 0x1FF0000;
+	if (c & 0x01000000) {
+		c = (c & 0x0000FFFF) | 0x00FF0000;
+	}
+#endif
+	return c;
+}
 #endif
 
 struct blip_t;
@@ -90,7 +178,9 @@ struct mCoreCallbacks {
 	void (*videoFrameEnded)(void* context);
 	void (*coreCrashed)(void* context);
 	void (*sleep)(void* context);
+	void (*shutdown)(void* context);
 	void (*keysRead)(void* context);
+	void (*savedataUpdated)(void* context);
 };
 
 DECLARE_VECTOR(mCoreCallbacksList, struct mCoreCallbacks);
@@ -175,6 +265,7 @@ enum mCoreMemoryBlockFlags {
 	mCORE_MEMORY_READ = 0x01,
 	mCORE_MEMORY_WRITE = 0x02,
 	mCORE_MEMORY_RW = 0x03,
+	mCORE_MEMORY_WORM = 0x04,
 	mCORE_MEMORY_MAPPED = 0x10,
 	mCORE_MEMORY_VIRTUAL = 0x20,
 };
