@@ -42,7 +42,7 @@ struct mLibrary {
 
 static void _mLibraryDeleteEntry(struct mLibrary* library, struct mLibraryEntry* entry);
 static void _mLibraryInsertEntry(struct mLibrary* library, struct mLibraryEntry* entry);
-static void _mLibraryAddEntry(struct mLibrary* library, const char* filename, const char* base, struct VFile* vf);
+static bool _mLibraryAddEntry(struct mLibrary* library, const char* filename, const char* base, struct VFile* vf);
 
 static void _bindConstraints(sqlite3_stmt* statement, const struct mLibraryEntry* constraints) {
 	if (!constraints) {
@@ -85,7 +85,7 @@ static void _bindConstraints(sqlite3_stmt* statement, const struct mLibraryEntry
 		sqlite3_bind_text(statement, index, constraints->internalCode, -1, SQLITE_TRANSIENT);
 	}
 
-	if (constraints->platform != PLATFORM_NONE) {
+	if (constraints->platform != mPLATFORM_NONE) {
 		useIndex = sqlite3_bind_parameter_index(statement, ":usePlatform");
 		index = sqlite3_bind_parameter_index(statement, ":platform");
 		sqlite3_bind_int(statement, useIndex, 1);
@@ -212,7 +212,7 @@ void mLibraryDestroy(struct mLibrary* library) {
 	free(library);
 }
 
-void mLibraryLoadDirectory(struct mLibrary* library, const char* base) {
+void mLibraryLoadDirectory(struct mLibrary* library, const char* base, bool recursive) {
 	struct VDir* dir = VDirOpenArchive(base);
 	if (!dir) {
 		dir = VDirOpen(base);
@@ -248,44 +248,55 @@ void mLibraryLoadDirectory(struct mLibrary* library, const char* base) {
 	dir->rewind(dir);
 	struct VDirEntry* dirent = dir->listNext(dir);
 	while (dirent) {
-		struct VFile* vf = dir->openFile(dir, dirent->name(dirent), O_RDONLY);
-		if (!vf) {
-			dirent = dir->listNext(dir);
-			continue;
+		const char* name = dirent->name(dirent);
+		struct VFile* vf = dir->openFile(dir, name, O_RDONLY);
+		bool wasAdded = false;
+
+		if (vf) {
+			wasAdded = _mLibraryAddEntry(library, name, base, vf);
+		}
+		if (!wasAdded && name[0] != '.') {
+			char newBase[PATH_MAX];
+			snprintf(newBase, sizeof(newBase), "%s" PATH_SEP "%s", base, name);
+
+			if (recursive) {
+				mLibraryLoadDirectory(library, newBase, recursive);
+			} else if (dirent->type(dirent) == VFS_FILE) {
+				mLibraryLoadDirectory(library, newBase, true); // This will add as an archive
+			}
 		}
-		_mLibraryAddEntry(library, dirent->name(dirent), base, vf);
 		dirent = dir->listNext(dir);
 	}
 	dir->close(dir);
 	sqlite3_exec(library->db, "COMMIT;", NULL, NULL, NULL);
 }
 
-void _mLibraryAddEntry(struct mLibrary* library, const char* filename, const char* base, struct VFile* vf) {
-	struct mCore* core;
+bool _mLibraryAddEntry(struct mLibrary* library, const char* filename, const char* base, struct VFile* vf) {
 	if (!vf) {
-		return;
+		return false;
 	}
-	core = mCoreFindVF(vf);
-	if (core) {
-		struct mLibraryEntry entry;
-		memset(&entry, 0, sizeof(entry));
-		core->init(core);
-		core->loadROM(core, vf);
-
-		core->getGameTitle(core, entry.internalTitle);
-		core->getGameCode(core, entry.internalCode);
-		core->checksum(core, &entry.crc32, CHECKSUM_CRC32);
-		entry.platform = core->platform(core);
-		entry.title = NULL;
-		entry.base = base;
-		entry.filename = filename;
-		entry.filesize = vf->size(vf);
-		_mLibraryInsertEntry(library, &entry);
-		// Note: this destroys the VFile
-		core->deinit(core);
-	} else {
+	struct mCore* core = mCoreFindVF(vf);
+	if (!core) {
 		vf->close(vf);
+		return false;
 	}
+	struct mLibraryEntry entry;
+	memset(&entry, 0, sizeof(entry));
+	core->init(core);
+	core->loadROM(core, vf);
+
+	core->getGameTitle(core, entry.internalTitle);
+	core->getGameCode(core, entry.internalCode);
+	core->checksum(core, &entry.crc32, mCHECKSUM_CRC32);
+	entry.platform = core->platform(core);
+	entry.title = NULL;
+	entry.base = base;
+	entry.filename = filename;
+	entry.filesize = vf->size(vf);
+	_mLibraryInsertEntry(library, &entry);
+	// Note: this destroys the VFile
+	core->deinit(core);
+	return true;
 }
 
 static void _mLibraryInsertEntry(struct mLibrary* library, struct mLibraryEntry* entry) {
@@ -445,6 +456,11 @@ struct VFile* mLibraryOpenVFile(struct mLibrary* library, const struct mLibraryE
 			break;
 		}
 	}
+	for (i = 0; i < mLibraryListingSize(&entries); ++i) {
+		struct mLibraryEntry* e = mLibraryListingGetPointer(&entries, i);
+		mLibraryEntryFree(e);
+	}
+	mLibraryListingDeinit(&entries);
 	return vf;
 }
 
