@@ -7,6 +7,8 @@
 
 #include <mgba/core/core.h>
 #include <mgba/gba/interface.h>
+#include <mgba/internal/gba/gba.h>
+#include <mgba-util/math.h>
 
 #include <libavcodec/version.h>
 #include <libavcodec/avcodec.h>
@@ -61,6 +63,7 @@ void FFmpegEncoderInit(struct FFmpegEncoder* encoder) {
 	encoder->iheight = GBA_VIDEO_VERTICAL_PIXELS;
 	encoder->frameskip = 1;
 	encoder->skipResidue = 0;
+	encoder->loop = false;
 	encoder->ipixFormat =
 #ifdef COLOR_16_BIT
 #ifdef COLOR_5_6_5
@@ -91,6 +94,7 @@ void FFmpegEncoderInit(struct FFmpegEncoder* encoder) {
 	encoder->source = NULL;
 	encoder->sink = NULL;
 	encoder->sinkFrame = NULL;
+	FFmpegEncoderSetInputFrameRate(encoder, VIDEO_TOTAL_LENGTH, GBA_ARM7TDMI_FREQUENCY);
 
 	int i;
 	for (i = 0; i < FFMPEG_FILTERS_MAX; ++i) {
@@ -160,7 +164,7 @@ bool FFmpegEncoderSetAudio(struct FFmpegEncoder* encoder, const char* acodec, un
 	return true;
 }
 
-bool FFmpegEncoderSetVideo(struct FFmpegEncoder* encoder, const char* vcodec, unsigned vbr, int frameskip) {
+bool FFmpegEncoderSetVideo(struct FFmpegEncoder* encoder, const char* vcodec, int vbr, int frameskip) {
 	static const struct {
 		enum AVPixelFormat format;
 		int priority;
@@ -177,10 +181,12 @@ bool FFmpegEncoderSetVideo(struct FFmpegEncoder* encoder, const char* vcodec, un
 		{ AV_PIX_FMT_0BGR, 3 },
 		{ AV_PIX_FMT_0RGB, 3 },
 #endif
-		{ AV_PIX_FMT_YUV422P, 4 },
+		{ AV_PIX_FMT_RGB32, 4},
+		{ AV_PIX_FMT_BGR32, 4},
 		{ AV_PIX_FMT_YUV444P, 5 },
-		{ AV_PIX_FMT_YUV420P, 6 },
-		{ AV_PIX_FMT_PAL8, 7 },
+		{ AV_PIX_FMT_YUV422P, 6 },
+		{ AV_PIX_FMT_YUV420P, 7 },
+		{ AV_PIX_FMT_PAL8, 8 },
 	};
 
 	if (!vcodec) {
@@ -208,6 +214,9 @@ bool FFmpegEncoderSetVideo(struct FFmpegEncoder* encoder, const char* vcodec, un
 	if (encoder->pixFormat == AV_PIX_FMT_NONE) {
 		return false;
 	}
+	if (vbr < 0 && !av_opt_find(&codec->priv_class, "crf", NULL, 0, 0)) {
+		return false;
+	}
 	encoder->videoCodec = vcodec;
 	encoder->videoBitrate = vbr;
 	encoder->frameskip = frameskip + 1;
@@ -228,6 +237,10 @@ void FFmpegEncoderSetDimensions(struct FFmpegEncoder* encoder, int width, int he
 	encoder->height = height > 0 ? height : GBA_VIDEO_VERTICAL_PIXELS;
 }
 
+void FFmpegEncoderSetLooping(struct FFmpegEncoder* encoder, bool loop) {
+	encoder->loop = loop;
+}
+
 bool FFmpegEncoderVerifyContainer(struct FFmpegEncoder* encoder) {
 	AVOutputFormat* oformat = av_guess_format(encoder->containerFormat, 0, 0);
 	AVCodec* acodec = avcodec_find_encoder_by_name(encoder->audioCodec);
@@ -309,6 +322,7 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		encoder->audioFrame->nb_samples = encoder->audio->frame_size;
 		encoder->audioFrame->format = encoder->audio->sample_fmt;
 		encoder->audioFrame->pts = 0;
+		encoder->audioFrame->channel_layout = AV_CH_LAYOUT_STEREO;
 #ifdef USE_LIBAVRESAMPLE
 		encoder->resampleContext = avresample_alloc_context();
 		av_opt_set_int(encoder->resampleContext, "in_channel_layout", AV_CH_LAYOUT_STEREO, 0);
@@ -325,14 +339,12 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 #endif
 		encoder->audioBufferSize = (encoder->audioFrame->nb_samples * PREFERRED_SAMPLE_RATE / encoder->sampleRate) * 4;
 		encoder->audioBuffer = av_malloc(encoder->audioBufferSize);
-		encoder->postaudioBufferSize = av_samples_get_buffer_size(0, encoder->audio->channels, encoder->audio->frame_size, encoder->audio->sample_fmt, 0);
-		encoder->postaudioBuffer = av_malloc(encoder->postaudioBufferSize);
-		avcodec_fill_audio_frame(encoder->audioFrame, encoder->audio->channels, encoder->audio->sample_fmt, (const uint8_t*) encoder->postaudioBuffer, encoder->postaudioBufferSize, 0);
+		av_frame_get_buffer(encoder->audioFrame, 0);
 
 		if (encoder->audio->codec->id == AV_CODEC_ID_AAC &&
-		    (strcasecmp(encoder->containerFormat, "mp4") ||
-		        strcasecmp(encoder->containerFormat, "m4v") ||
-		        strcasecmp(encoder->containerFormat, "mov"))) {
+		    (strcasecmp(encoder->containerFormat, "mp4") == 0||
+		        strcasecmp(encoder->containerFormat, "m4v") == 0 ||
+		        strcasecmp(encoder->containerFormat, "mov") == 0)) {
 			// MP4 container doesn't support the raw ADTS AAC format that the encoder spits out
 #ifdef FFMPEG_USE_NEW_BSF
 			av_bsf_alloc(av_bsf_get_by_name("aac_adtstoasc"), &encoder->absf);
@@ -358,8 +370,10 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		encoder->video->bit_rate = encoder->videoBitrate;
 		encoder->video->width = encoder->width;
 		encoder->video->height = encoder->height;
-		encoder->video->time_base = (AVRational) { VIDEO_TOTAL_LENGTH * encoder->frameskip, GBA_ARM7TDMI_FREQUENCY };
-		encoder->video->framerate = (AVRational) { GBA_ARM7TDMI_FREQUENCY, VIDEO_TOTAL_LENGTH * encoder->frameskip };
+		encoder->video->time_base = (AVRational) { encoder->frameCycles * encoder->frameskip, encoder->cycles };
+		encoder->video->framerate = (AVRational) { encoder->cycles, encoder->frameCycles * encoder->frameskip };
+		encoder->videoStream->time_base = encoder->video->time_base;
+		encoder->videoStream->avg_frame_rate = encoder->video->framerate;
 		encoder->video->pix_fmt = encoder->pixFormat;
 		encoder->video->gop_size = 60;
 		encoder->video->max_b_frames = 3;
@@ -372,9 +386,9 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		}
 
 		if (encoder->video->codec->id == AV_CODEC_ID_H264 &&
-		    (strcasecmp(encoder->containerFormat, "mp4") ||
-		        strcasecmp(encoder->containerFormat, "m4v") ||
-		        strcasecmp(encoder->containerFormat, "mov"))) {
+		    (strcasecmp(encoder->containerFormat, "mp4") == 0 ||
+		        strcasecmp(encoder->containerFormat, "m4v") == 0 ||
+		        strcasecmp(encoder->containerFormat, "mov") == 0)) {
 			// QuickTime and a few other things require YUV420
 			encoder->video->pix_fmt = AV_PIX_FMT_YUV420P;
 		}
@@ -390,7 +404,13 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		if (encoder->video->codec->id == AV_CODEC_ID_PNG) {
 			encoder->video->compression_level = 8;
 		}
-		if (strcmp(vcodec->name, "libx264") == 0) {
+#if LIBAVCODEC_VERSION_INT >= AV_VERSION_INT(58, 48, 100)
+		if (encoder->video->codec->id == AV_CODEC_ID_ZMBV) {
+			encoder->video->compression_level = 5;
+			encoder->video->pix_fmt = AV_PIX_FMT_BGR0;
+		}
+#endif
+		if (strcmp(vcodec->name, "libx264") == 0 || strcmp(vcodec->name, "libx264rgb") == 0) {
 			// Try to adaptively figure out when you can use a slower encoder
 			if (encoder->width * encoder->height > 1000000) {
 				av_opt_set(encoder->video->priv_data, "preset", "superfast", 0);
@@ -399,10 +419,26 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 			} else {
 				av_opt_set(encoder->video->priv_data, "preset", "faster", 0);
 			}
+			av_opt_set(encoder->video->priv_data, "tune", "zerolatency", 0);
 			if (encoder->videoBitrate == 0) {
 				av_opt_set(encoder->video->priv_data, "qp", "0", 0);
-				encoder->video->pix_fmt = AV_PIX_FMT_YUV444P;
+				if (strcmp(vcodec->name, "libx264") == 0) {
+					encoder->video->pix_fmt = AV_PIX_FMT_YUV444P;
+				}
+			} else if (encoder->videoBitrate < 0) {
+				av_opt_set_int(encoder->video->priv_data, "crf", -encoder->videoBitrate, 0);
 			}
+		} else if (encoder->videoBitrate < 0) {
+			if (strcmp(vcodec->name, "libvpx") == 0 || strcmp(vcodec->name, "libvpx-vp9") == 0 || strcmp(vcodec->name, "libx265") == 0) {
+				av_opt_set_int(encoder->video->priv_data, "crf", -encoder->videoBitrate, 0);
+			} else {
+				FFmpegEncoderClose(encoder);
+				return false;
+			}
+		}
+		if (strncmp(vcodec->name, "libvpx", 6) == 0) {
+			av_opt_set_int(encoder->video->priv_data, "cpu-used", 2, 0);
+			av_opt_set(encoder->video->priv_data, "deadline", "realtime", 0);
 		}
 		if (strcmp(vcodec->name, "libvpx-vp9") == 0 && encoder->videoBitrate == 0) {
 			av_opt_set_int(encoder->video->priv_data, "lossless", 1, 0);
@@ -410,6 +446,10 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 			encoder->video->gop_size = 120;
 			encoder->video->pix_fmt = AV_PIX_FMT_GBRP;
 		}
+		if (strcmp(vcodec->name, "libwebp_anim") == 0 && encoder->videoBitrate == 0) {
+			av_opt_set(encoder->video->priv_data, "lossless", "1", 0);
+			encoder->video->pix_fmt = AV_PIX_FMT_RGB32;
+		}
 
 		if (encoder->pixFormat == AV_PIX_FMT_PAL8) {
 			encoder->graph = avfilter_graph_alloc();
@@ -476,12 +516,20 @@ bool FFmpegEncoderOpen(struct FFmpegEncoder* encoder, const char* outfile) {
 		encoder->videoFrame->height = encoder->video->height;
 		encoder->videoFrame->pts = 0;
 		_ffmpegSetVideoDimensions(&encoder->d, encoder->iwidth, encoder->iheight);
-		av_image_alloc(encoder->videoFrame->data, encoder->videoFrame->linesize, encoder->videoFrame->width, encoder->videoFrame->height, encoder->videoFrame->format, 32);
+		av_frame_get_buffer(encoder->videoFrame, 32);
 #ifdef FFMPEG_USE_CODECPAR
 		avcodec_parameters_from_context(encoder->videoStream->codecpar, encoder->video);
 #endif
 	}
 
+	if (strcmp(encoder->containerFormat, "gif") == 0) {
+		av_opt_set(encoder->context->priv_data, "loop", encoder->loop ? "0" : "-1", 0);
+	} else if (strcmp(encoder->containerFormat, "apng") == 0) {
+		av_opt_set(encoder->context->priv_data, "plays", encoder->loop ? "0" : "1", 0);
+	} else if (strcmp(encoder->containerFormat, "webp") == 0) {
+		av_opt_set(encoder->context->priv_data, "loop", encoder->loop ? "0" : "1", 0);
+	}
+
 	AVDictionary* opts = 0;
 	av_dict_set(&opts, "strict", "-2", 0);
 	bool res = avio_open(&encoder->context->pb, outfile, AVIO_FLAG_WRITE) < 0 || avformat_write_header(encoder->context, &opts) < 0;
@@ -543,8 +591,12 @@ void FFmpegEncoderClose(struct FFmpegEncoder* encoder) {
 #endif
 	}
 	if (encoder->audio) {
+#ifdef FFMPEG_USE_CODECPAR
+		avcodec_free_context(&encoder->audio);
+#else
 		avcodec_close(encoder->audio);
 		encoder->audio = NULL;
+#endif
 	}
 
 	if (encoder->resampleContext) {
@@ -566,6 +618,7 @@ void FFmpegEncoderClose(struct FFmpegEncoder* encoder) {
 	}
 
 	if (encoder->videoFrame) {
+		av_freep(encoder->videoFrame->data);
 #if LIBAVCODEC_VERSION_MAJOR >= 55
 		av_frame_free(&encoder->videoFrame);
 #else
@@ -583,8 +636,12 @@ void FFmpegEncoderClose(struct FFmpegEncoder* encoder) {
 	}
 
 	if (encoder->video) {
+#ifdef FFMPEG_USE_CODECPAR
+		avcodec_free_context(&encoder->video);
+#else
 		avcodec_close(encoder->video);
 		encoder->video = NULL;
+#endif
 	}
 
 	if (encoder->scaleContext) {
@@ -634,7 +691,6 @@ void _ffmpegPostAudioFrame(struct mAVStream* stream, int16_t left, int16_t right
 		return;
 	}
 
-	int channelSize = 2 * av_get_bytes_per_sample(encoder->audio->sample_fmt);
 	encoder->currentAudioSample = 0;
 #ifdef USE_LIBAVRESAMPLE
 	avresample_convert(encoder->resampleContext, 0, 0, 0,
@@ -646,7 +702,7 @@ void _ffmpegPostAudioFrame(struct mAVStream* stream, int16_t left, int16_t right
 #if LIBAVCODEC_VERSION_MAJOR >= 55
 	av_frame_make_writable(encoder->audioFrame);
 #endif
-	int samples = avresample_read(encoder->resampleContext, encoder->audioFrame->data, encoder->postaudioBufferSize / channelSize);
+	int samples = avresample_read(encoder->resampleContext, encoder->audioFrame->data, encoder->audioFrame->nb_samples);
 #else
 #if LIBAVCODEC_VERSION_MAJOR >= 55
 	av_frame_make_writable(encoder->audioFrame);
@@ -655,7 +711,7 @@ void _ffmpegPostAudioFrame(struct mAVStream* stream, int16_t left, int16_t right
 		swr_convert(encoder->resampleContext, NULL, 0, (const uint8_t**) &encoder->audioBuffer, encoder->audioBufferSize / 4);
 		return;
 	}
-	int samples = swr_convert(encoder->resampleContext, encoder->audioFrame->data, encoder->postaudioBufferSize / channelSize,
+	int samples = swr_convert(encoder->resampleContext, encoder->audioFrame->data, encoder->audioFrame->nb_samples,
 	                          (const uint8_t**) &encoder->audioBuffer, encoder->audioBufferSize / 4);
 #endif
 
@@ -666,61 +722,76 @@ void _ffmpegPostAudioFrame(struct mAVStream* stream, int16_t left, int16_t right
 }
 
 bool _ffmpegWriteAudioFrame(struct FFmpegEncoder* encoder, struct AVFrame* audioFrame) {
-	AVPacket packet;
-	av_init_packet(&packet);
-	packet.data = 0;
-	packet.size = 0;
+	AVPacket* packet;
+#ifdef FFMPEG_USE_PACKET_UNREF
+	packet = av_packet_alloc();
+#else
+	packet = av_malloc(sizeof(*packet));
+	av_init_packet(packet);
+#endif
+	packet->data = 0;
+	packet->size = 0;
 
 	int gotData;
 #ifdef FFMPEG_USE_PACKETS
 	avcodec_send_frame(encoder->audio, audioFrame);
-	gotData = avcodec_receive_packet(encoder->audio, &packet);
-	gotData = (gotData == 0) && packet.size;
+	gotData = avcodec_receive_packet(encoder->audio, packet);
+	gotData = (gotData == 0) && packet->size;
 #else
-	avcodec_encode_audio2(encoder->audio, &packet, audioFrame, &gotData);
+	avcodec_encode_audio2(encoder->audio, packet, audioFrame, &gotData);
 #endif
-	packet.pts = av_rescale_q(packet.pts, encoder->audio->time_base, encoder->audioStream->time_base);
-	packet.dts = packet.pts;
+	packet->pts = av_rescale_q(packet->pts, encoder->audio->time_base, encoder->audioStream->time_base);
+	packet->dts = packet->pts;
 
 	if (gotData) {
 		if (encoder->absf) {
-			AVPacket tempPacket;
+			AVPacket* tempPacket;
+#ifdef FFMPEG_USE_PACKETS
+			tempPacket = av_packet_alloc();
+#else
+			tempPacket = av_malloc(sizeof(*tempPacket));
+			av_init_packet(tempPacket);
+#endif
 
 #ifdef FFMPEG_USE_NEW_BSF
-			int success = av_bsf_send_packet(encoder->absf, &packet);
+			int success = av_bsf_send_packet(encoder->absf, packet);
 			if (success >= 0) {
-				success = av_bsf_receive_packet(encoder->absf, &tempPacket);
+				success = av_bsf_receive_packet(encoder->absf, tempPacket);
 			}
 #else
 			int success = av_bitstream_filter_filter(encoder->absf, encoder->audio, 0,
-			    &tempPacket.data, &tempPacket.size,
-			    packet.data, packet.size, 0);
+			    &tempPacket->data, &tempPacket->size,
+			    packet->data, packet->size, 0);
 #endif
 
 			if (success >= 0) {
 #if LIBAVUTIL_VERSION_MAJOR >= 53
-				tempPacket.buf = av_buffer_create(tempPacket.data, tempPacket.size, av_buffer_default_free, 0, 0);
+				tempPacket->buf = av_buffer_create(tempPacket->data, tempPacket->size, av_buffer_default_free, 0, 0);
 #endif
 
 #ifdef FFMPEG_USE_PACKET_UNREF
-				av_packet_move_ref(&packet, &tempPacket);
+				av_packet_move_ref(packet, tempPacket);
+				av_packet_free(&packet);
 #else
-				av_free_packet(&packet);
+				av_free_packet(packet);
+				av_freep(&packet);
 				packet = tempPacket;
 #endif
 
-				packet.stream_index = encoder->audioStream->index;
-				av_interleaved_write_frame(encoder->context, &packet);
+				packet->stream_index = encoder->audioStream->index;
+				av_interleaved_write_frame(encoder->context, packet);
 			}
 		} else {
-			packet.stream_index = encoder->audioStream->index;
-			av_interleaved_write_frame(encoder->context, &packet);
+			packet->stream_index = encoder->audioStream->index;
+			av_interleaved_write_frame(encoder->context, packet);
 		}
 	}
 #ifdef FFMPEG_USE_PACKET_UNREF
-	av_packet_unref(&packet);
+	av_packet_unref(packet);
+	av_packet_free(&packet);
 #else
-	av_free_packet(&packet);
+	av_free_packet(packet);
+	av_freep(&packet);
 #endif
 	return gotData;
 }
@@ -739,7 +810,12 @@ void _ffmpegPostVideoFrame(struct mAVStream* stream, const color_t* pixels, size
 #if LIBAVCODEC_VERSION_MAJOR >= 55
 	av_frame_make_writable(encoder->videoFrame);
 #endif
-	encoder->videoFrame->pts = av_rescale_q(encoder->currentVideoFrame, encoder->video->time_base, encoder->videoStream->time_base);
+	if (encoder->video->codec->id == AV_CODEC_ID_WEBP) {
+		// TODO: Figure out why WebP is rescaling internally (should video frames not be rescaled externally?)
+		encoder->videoFrame->pts = encoder->currentVideoFrame;
+	} else {
+		encoder->videoFrame->pts = av_rescale_q(encoder->currentVideoFrame, encoder->video->time_base, encoder->videoStream->time_base);
+	}
 	++encoder->currentVideoFrame;
 
 	sws_scale(encoder->scaleContext, (const uint8_t* const*) &pixels, (const int*) &stride, 0, encoder->iheight, encoder->videoFrame->data, encoder->videoFrame->linesize);
@@ -762,32 +838,39 @@ void _ffmpegPostVideoFrame(struct mAVStream* stream, const color_t* pixels, size
 }
 
 bool _ffmpegWriteVideoFrame(struct FFmpegEncoder* encoder, struct AVFrame* videoFrame) {
-	AVPacket packet;
+	AVPacket* packet;
 
-	av_init_packet(&packet);
-	packet.data = 0;
-	packet.size = 0;
+#ifdef FFMPEG_USE_PACKET_UNREF
+	packet = av_packet_alloc();
+#else
+	packet = av_malloc(sizeof(*packet));
+	av_init_packet(packet);
+#endif
+	packet->data = 0;
+	packet->size = 0;
 
 	int gotData;
 #ifdef FFMPEG_USE_PACKETS
 	avcodec_send_frame(encoder->video, videoFrame);
-	gotData = avcodec_receive_packet(encoder->video, &packet) == 0;
+	gotData = avcodec_receive_packet(encoder->video, packet) == 0;
 #else
-	avcodec_encode_video2(encoder->video, &packet, videoFrame, &gotData);
+	avcodec_encode_video2(encoder->video, packet, videoFrame, &gotData);
 #endif
 	if (gotData) {
 #ifndef FFMPEG_USE_PACKET_UNREF
 		if (encoder->video->coded_frame->key_frame) {
-			packet.flags |= AV_PKT_FLAG_KEY;
+			packet->flags |= AV_PKT_FLAG_KEY;
 		}
 #endif
-		packet.stream_index = encoder->videoStream->index;
-		av_interleaved_write_frame(encoder->context, &packet);
+		packet->stream_index = encoder->videoStream->index;
+		av_interleaved_write_frame(encoder->context, packet);
 	}
 #ifdef FFMPEG_USE_PACKET_UNREF
-	av_packet_unref(&packet);
+	av_packet_unref(packet);
+	av_packet_free(&packet);
 #else
-	av_free_packet(&packet);
+	av_free_packet(packet);
+	av_freep(&packet);
 #endif
 
 	return gotData;
@@ -807,3 +890,12 @@ static void _ffmpegSetVideoDimensions(struct mAVStream* stream, unsigned width,
 	    encoder->videoFrame->width, encoder->videoFrame->height, encoder->videoFrame->format,
 	    SWS_POINT, 0, 0, 0);
 }
+
+void FFmpegEncoderSetInputFrameRate(struct FFmpegEncoder* encoder, int numerator, int denominator) {
+	reduceFraction(&numerator, &denominator);
+	encoder->frameCycles = numerator;
+	encoder->cycles = denominator;
+	if (encoder->video) {
+		encoder->video->framerate = (AVRational) { denominator, numerator * encoder->frameskip };
+	}
+}
