@@ -40,6 +40,11 @@ enum {
 
 #define RUNNER_STATE(X) ((X) << 16)
 
+enum {
+	SCREENSHOT_VALID = 0x10000,
+	SCREENSHOT_INVALID = 0x20000,
+};
+
 static const struct mInputPlatformInfo _mGUIKeyInfo = {
 	.platformName = "gui",
 	.keyId = (const char*[GUI_INPUT_MAX]) {
@@ -56,6 +61,7 @@ static const struct mInputPlatformInfo _mGUIKeyInfo = {
 		[mGUI_INPUT_SCREENSHOT] = "Take screenshot",
 		[mGUI_INPUT_FAST_FORWARD_HELD] = "Fast forward (held)",
 		[mGUI_INPUT_FAST_FORWARD_TOGGLE] = "Fast forward (toggle)",
+		[mGUI_INPUT_MUTE_TOGGLE] = "Mute (toggle)",
 	},
 	.nKeys = GUI_INPUT_MAX
 };
@@ -104,39 +110,49 @@ static void _drawBackground(struct GUIBackground* background, void* context) {
 
 static void _drawState(struct GUIBackground* background, void* id) {
 	struct mGUIBackground* gbaBackground = (struct mGUIBackground*) background;
-	int stateId = ((int) id) >> 16;
+	unsigned stateId = ((uint32_t) id) >> 16;
 	if (gbaBackground->p->drawScreenshot) {
 		unsigned w, h;
 		gbaBackground->p->core->desiredVideoDimensions(gbaBackground->p->core, &w, &h);
-		if (gbaBackground->screenshot && gbaBackground->screenshotId == (int) id) {
-			gbaBackground->p->drawScreenshot(gbaBackground->p, gbaBackground->screenshot, w, h, true);
+		size_t size = w * h * BYTES_PER_PIXEL;
+		if (size != gbaBackground->imageSize) {
+			mappedMemoryFree(gbaBackground->image, gbaBackground->imageSize);
+			gbaBackground->image = NULL;
+		}
+		if (gbaBackground->image && gbaBackground->screenshotId == (stateId | SCREENSHOT_VALID)) {
+			gbaBackground->p->drawScreenshot(gbaBackground->p, gbaBackground->image, w, h, true);
 			return;
-		}
-		struct VFile* vf = mCoreGetState(gbaBackground->p->core, stateId, false);
-		color_t* pixels = gbaBackground->screenshot;
-		if (!pixels) {
-			pixels = anonymousMemoryMap(w * h * 4);
-			gbaBackground->screenshot = pixels;
-		}
-		bool success = false;
-		if (vf && isPNG(vf) && pixels) {
-			png_structp png = PNGReadOpen(vf, PNG_HEADER_BYTES);
-			png_infop info = png_create_info_struct(png);
-			png_infop end = png_create_info_struct(png);
-			if (png && info && end) {
-				success = PNGReadHeader(png, info);
-				success = success && PNGReadPixels(png, info, pixels, w, h, w);
-				success = success && PNGReadFooter(png, end);
+		} else if (gbaBackground->screenshotId != (stateId | SCREENSHOT_INVALID)) {
+			struct VFile* vf = mCoreGetState(gbaBackground->p->core, stateId, false);
+			color_t* pixels = gbaBackground->image;
+			if (!pixels) {
+				pixels = anonymousMemoryMap(size);
+				gbaBackground->image = pixels;
+				gbaBackground->imageSize = size;
+			}
+			bool success = false;
+			if (vf && isPNG(vf) && pixels) {
+				png_structp png = PNGReadOpen(vf, PNG_HEADER_BYTES);
+				png_infop info = png_create_info_struct(png);
+				png_infop end = png_create_info_struct(png);
+				if (png && info && end) {
+					success = PNGReadHeader(png, info);
+					success = success && PNGReadPixels(png, info, pixels, w, h, w);
+					success = success && PNGReadFooter(png, end);
+				}
+				PNGReadClose(png, info, end);
+			}
+			if (vf) {
+				vf->close(vf);
+			}
+			if (success) {
+				gbaBackground->p->drawScreenshot(gbaBackground->p, pixels, w, h, true);
+				gbaBackground->screenshotId = stateId | SCREENSHOT_VALID;
+			} else {
+				gbaBackground->screenshotId = stateId | SCREENSHOT_INVALID;
 			}
-			PNGReadClose(png, info, end);
 		}
-		if (vf) {
-			vf->close(vf);
-		}
-		if (success) {
-			gbaBackground->p->drawScreenshot(gbaBackground->p, pixels, w, h, true);
-			gbaBackground->screenshotId = (int) id;
-		} else if (gbaBackground->p->drawFrame) {
+		if (gbaBackground->p->drawFrame && gbaBackground->screenshotId == (stateId | SCREENSHOT_INVALID)) {
 			gbaBackground->p->drawFrame(gbaBackground->p, true);
 		}
 	}
@@ -165,10 +181,10 @@ static void _tryAutosave(struct mGUIRunner* runner) {
 #ifdef DISABLE_THREADING
 	mCoreSaveState(runner->core, 0, SAVESTATE_SAVEDATA | SAVESTATE_RTC | SAVESTATE_METADATA);
 #else
+	MutexLock(&runner->autosave.mutex);
 	if (!runner->autosave.buffer) {
 		runner->autosave.buffer = VFileMemChunk(NULL, 0);
 	}
-	MutexLock(&runner->autosave.mutex);
 	runner->autosave.core = runner->core;
 	mCoreSaveStateNamed(runner->core, runner->autosave.buffer, SAVESTATE_SAVEDATA | SAVESTATE_RTC | SAVESTATE_METADATA);
 	ConditionWake(&runner->autosave.cond);
@@ -201,6 +217,7 @@ void mGUIInit(struct mGUIRunner* runner, const char* port) {
 #else
 	mCoreConfigSetDefaultIntValue(&runner->config, "autosave", true);
 #endif
+	mCoreConfigSetDefaultIntValue(&runner->config, "showOSD", true);
 	mCoreConfigLoad(&runner->config);
 	mCoreConfigGetIntValue(&runner->config, "logLevel", &logger.logLevel);
 
@@ -287,14 +304,13 @@ static void _log(struct mLogger* logger, int category, enum mLogLevel level, con
 			guiLogger->vf = NULL;
 		}
 	}
+#ifdef GEKKO
+	puts(log2);
+#endif
 }
 
 static void _updateLoading(size_t read, size_t size, void* context) {
 	struct mGUIRunner* runner = context;
-	if (read & 0x3FFFF) {
-		return;
-	}
-
 	runner->params.drawStart();
 	if (runner->params.guiPrepare) {
 		runner->params.guiPrepare();
@@ -312,7 +328,7 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 			.draw = _drawState
 		},
 		.p = runner,
-		.screenshot = 0,
+		.image = 0,
 		.screenshotId = 0
 	};
 	struct GUIMenu pauseMenu = {
@@ -382,8 +398,25 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 		mCoreInitConfig(runner->core, runner->port);
 		mInputMapInit(&runner->core->inputMap, &GBAInputInfo);
 
-		found = mCorePreloadFileCB(runner->core, path, _updateLoading, runner);
+		struct VFile* rom = mDirectorySetOpenPath(&runner->core->dirs, path, runner->core->isROM);
+		if (runner->setFrameLimiter) {
+			runner->setFrameLimiter(runner, false);
+		}
+		found = mCorePreloadVFCB(runner->core, rom, _updateLoading, runner);
+		if (runner->setFrameLimiter) {
+			runner->setFrameLimiter(runner, true);
+		}
+
+#ifdef FIXED_ROM_BUFFER
+		extern size_t romBufferSize;
+		if (!found && rom && (size_t) rom->size(rom) > romBufferSize) {
+			found = runner->core->loadROM(runner->core, rom);
+		}
+#endif
 		if (!found) {
+			if (rom) {
+				rom->close(rom);
+			}
 			mLOG(GUI_RUNNER, WARN, "Failed to load %s!", path);
 			mCoreConfigDeinit(&runner->core->config);
 			runner->core->deinit(runner->core);
@@ -395,7 +428,7 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 		GUIShowMessageBox(&runner->params, GUI_MESSAGE_BOX_OK, 240, "Load failed!");
 		return;
 	}
-	if (runner->core->platform(runner->core) == PLATFORM_GBA) {
+	if (runner->core->platform(runner->core) == mPLATFORM_GBA) {
 		runner->core->setPeripheral(runner->core, mPERIPH_GBA_LUMINANCE, &runner->luminanceSource.d);
 	}
 	mLOG(GUI_RUNNER, DEBUG, "Loading config...");
@@ -426,6 +459,12 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 		mCoreLoadState(runner->core, 0, SAVESTATE_SCREENSHOT | SAVESTATE_RTC);
 	}
 
+	int showOSD = true;
+	mCoreConfigGetIntValue(&runner->config, "showOSD", &showOSD);
+
+	int drawFps = false;
+	mCoreConfigGetIntValue(&runner->config, "fpsCounter", &drawFps);
+
 	bool running = true;
 
 #ifndef DISABLE_THREADING
@@ -487,6 +526,11 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 					runner->setFrameLimiter(runner, true);
 				}
 			}
+			if (guiKeys & (1 << mGUI_INPUT_MUTE_TOGGLE)) {
+				int mute = !runner->core->opts.mute;
+				mCoreConfigSetUIntValue(&runner->config, "mute", mute);
+				runner->core->reloadConfigOption(runner->core, "mute", &runner->config);
+			}
 			uint16_t keys = runner->pollGameInput(runner);
 			if (runner->prepareForFrame) {
 				runner->prepareForFrame(runner);
@@ -494,16 +538,29 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 			runner->core->setKeys(runner->core, keys);
 			runner->core->runFrame(runner->core);
 			if (runner->drawFrame) {
-				int drawFps = false;
-				mCoreConfigGetIntValue(&runner->config, "fpsCounter", &drawFps);
-
 				runner->params.drawStart();
 				runner->drawFrame(runner, false);
-				if (drawFps) {
+				if (showOSD || drawFps) {
 					if (runner->params.guiPrepare) {
 						runner->params.guiPrepare();
 					}
-					GUIFontPrintf(runner->params.font, 0, GUIFontHeight(runner->params.font), GUI_ALIGN_LEFT, 0x7FFFFFFF, "%.2f fps", runner->fps);
+					if (drawFps) {
+						GUIFontPrintf(runner->params.font, 0, GUIFontHeight(runner->params.font), GUI_ALIGN_LEFT, 0x7FFFFFFF, "%.2f fps", runner->fps);
+					}
+					if (showOSD) {
+						unsigned origin = runner->params.width - GUIFontHeight(runner->params.font) / 2;
+						unsigned w;
+						if (fastForward || (heldKeys & (1 << mGUI_INPUT_FAST_FORWARD_HELD))) {
+							GUIFontDrawIcon(runner->params.font, origin, GUIFontHeight(runner->params.font) / 2, GUI_ALIGN_RIGHT, 0, 0x7FFFFFFF, GUI_ICON_STATUS_FAST_FORWARD);
+							GUIFontIconMetrics(runner->params.font, GUI_ICON_STATUS_FAST_FORWARD, &w, NULL);
+							origin -= w + GUIFontHeight(runner->params.font) / 2;
+						}
+						if (runner->core->opts.mute) {
+							GUIFontDrawIcon(runner->params.font, origin, GUIFontHeight(runner->params.font) / 2, GUI_ALIGN_RIGHT, 0, 0x7FFFFFFF, GUI_ICON_STATUS_MUTE);
+							GUIFontIconMetrics(runner->params.font, GUI_ICON_STATUS_MUTE, &w, NULL);
+							origin -= w + GUIFontHeight(runner->params.font) / 2;
+						}
+					}
 					if (runner->params.guiFinish) {
 						runner->params.guiFinish();
 					}
@@ -563,10 +620,10 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 				runner->core->reset(runner->core);
 				break;
 			case RUNNER_SAVE_STATE:
-				mCoreSaveState(runner->core, ((int) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_SAVEDATA | SAVESTATE_RTC | SAVESTATE_METADATA);
+				mCoreSaveState(runner->core, ((uint32_t) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_SAVEDATA | SAVESTATE_RTC | SAVESTATE_METADATA);
 				break;
 			case RUNNER_LOAD_STATE:
-				mCoreLoadState(runner->core, ((int) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_RTC);
+				mCoreLoadState(runner->core, ((uint32_t) item->data) >> 16, SAVESTATE_SCREENSHOT | SAVESTATE_RTC);
 				break;
 			case RUNNER_SCREENSHOT:
 				mCoreTakeScreenshot(runner->core);
@@ -581,19 +638,26 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 		int frames = 0;
 		GUIPollInput(&runner->params, 0, &keys);
 		while (keys && frames < 30) {
-			++frames;
-			runner->params.drawStart();
-			runner->drawFrame(runner, true);
-			runner->params.drawEnd();
 #ifdef _3DS
-			// XXX: Why does this fix #1294?
-			usleep(1000);
+			if (!frames) {
 #endif
+				runner->params.drawStart();
+				runner->drawFrame(runner, true);
+				runner->params.drawEnd();
+#ifdef _3DS
+			} else {
+				// XXX: Why does this fix #1294?
+				usleep(15000);
+			}
+#endif
+			++frames;
 			GUIPollInput(&runner->params, 0, &keys);
 		}
 		if (runner->unpaused) {
 			runner->unpaused(runner);
 		}
+		mCoreConfigGetIntValue(&runner->config, "fpsCounter", &drawFps);
+		mCoreConfigGetIntValue(&runner->config, "showOSD", &showOSD);
 	}
 	mLOG(GUI_RUNNER, DEBUG, "Shutting down...");
 	if (runner->gameUnloaded) {
@@ -614,10 +678,8 @@ void mGUIRun(struct mGUIRunner* runner, const char* path) {
 	mLOG(GUI_RUNNER, DEBUG, "Unloading game...");
 	runner->core->unloadROM(runner->core);
 	drawState.screenshotId = 0;
-	if (drawState.screenshot) {
-		unsigned w, h;
-		runner->core->desiredVideoDimensions(runner->core, &w, &h);
-		mappedMemoryFree(drawState.screenshot, w * h * 4);
+	if (drawState.image) {
+		mappedMemoryFree(drawState.image, drawState.imageSize);
 	}
 
 	if (runner->config.port) {
@@ -677,7 +739,13 @@ THREAD_ENTRY mGUIAutosaveThread(void* context) {
 	while (autosave->running) {
 		ConditionWait(&autosave->cond, &autosave->mutex);
 		if (autosave->running && autosave->core) {
+			if (!autosave->buffer) {
+				continue;
+			}
 			struct VFile* vf = mCoreGetState(autosave->core, 0, true);
+			if (!vf) {
+				continue;
+			}
 			void* mem = autosave->buffer->map(autosave->buffer, autosave->buffer->size(autosave->buffer), MAP_READ);
 			vf->write(vf, mem, autosave->buffer->size(autosave->buffer));
 			autosave->buffer->unmap(autosave->buffer, mem, autosave->buffer->size(autosave->buffer));
