@@ -26,9 +26,9 @@
 #include <mgba-util/vfs.h>
 
 static const struct mCoreChannelInfo _GBVideoLayers[] = {
-	{ 0, "bg", "Background", NULL },
-	{ 1, "bgwin", "Window", NULL },
-	{ 2, "obj", "Objects", NULL },
+	{ GB_LAYER_BACKGROUND, "bg", "Background", NULL },
+	{ GB_LAYER_WINDOW, "bgwin", "Window", NULL },
+	{ GB_LAYER_OBJ, "obj", "Objects", NULL },
 };
 
 static const struct mCoreChannelInfo _GBAudioChannels[] = {
@@ -40,7 +40,7 @@ static const struct mCoreChannelInfo _GBAudioChannels[] = {
 
 static const struct mCoreMemoryBlock _GBMemoryBlocks[] = {
 	{ -1, "mem", "All", "All", 0, 0x10000, 0x10000, mCORE_MEMORY_VIRTUAL },
-	{ GB_REGION_CART_BANK0, "cart0", "ROM Bank", "Game Pak (32kiB)", GB_BASE_CART_BANK0, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 * 2, 0x800000, mCORE_MEMORY_READ | mCORE_MEMORY_MAPPED, 511, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 },
+	{ GB_REGION_CART_BANK0, "cart0", "ROM Bank", "Game Pak (32kiB)", GB_BASE_CART_BANK0, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 * 2, 0x800000, mCORE_MEMORY_READ | mCORE_MEMORY_WORM | mCORE_MEMORY_MAPPED, 511, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 },
 	{ GB_REGION_VRAM, "vram", "VRAM", "Video RAM (8kiB)", GB_BASE_VRAM, GB_BASE_VRAM + GB_SIZE_VRAM, GB_SIZE_VRAM, mCORE_MEMORY_RW | mCORE_MEMORY_MAPPED },
 	{ GB_REGION_EXTERNAL_RAM, "sram", "SRAM", "External RAM (8kiB)", GB_BASE_EXTERNAL_RAM, GB_BASE_EXTERNAL_RAM + GB_SIZE_EXTERNAL_RAM, GB_SIZE_EXTERNAL_RAM * 4, mCORE_MEMORY_RW | mCORE_MEMORY_MAPPED, 3 },
 	{ GB_REGION_WORKING_RAM_BANK0, "wram", "WRAM", "Working RAM (8kiB)", GB_BASE_WORKING_RAM_BANK0, GB_BASE_WORKING_RAM_BANK0 + GB_SIZE_WORKING_RAM_BANK0 * 2 , GB_SIZE_WORKING_RAM_BANK0 * 2, mCORE_MEMORY_RW | mCORE_MEMORY_MAPPED },
@@ -51,7 +51,7 @@ static const struct mCoreMemoryBlock _GBMemoryBlocks[] = {
 
 static const struct mCoreMemoryBlock _GBCMemoryBlocks[] = {
 	{ -1, "mem", "All", "All", 0, 0x10000, 0x10000, mCORE_MEMORY_VIRTUAL },
-	{ GB_REGION_CART_BANK0, "cart0", "ROM Bank", "Game Pak (32kiB)", GB_BASE_CART_BANK0, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 * 2, 0x800000, mCORE_MEMORY_READ | mCORE_MEMORY_MAPPED, 511, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 },
+	{ GB_REGION_CART_BANK0, "cart0", "ROM Bank", "Game Pak (32kiB)", GB_BASE_CART_BANK0, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 * 2, 0x800000, mCORE_MEMORY_READ | mCORE_MEMORY_WORM | mCORE_MEMORY_MAPPED, 511, GB_BASE_CART_BANK0 + GB_SIZE_CART_BANK0 },
 	{ GB_REGION_VRAM, "vram", "VRAM", "Video RAM (8kiB)", GB_BASE_VRAM, GB_BASE_VRAM + GB_SIZE_VRAM, GB_SIZE_VRAM * 2, mCORE_MEMORY_RW | mCORE_MEMORY_MAPPED, 1 },
 	{ GB_REGION_EXTERNAL_RAM, "sram", "SRAM", "External RAM (8kiB)", GB_BASE_EXTERNAL_RAM, GB_BASE_EXTERNAL_RAM + GB_SIZE_EXTERNAL_RAM, GB_SIZE_EXTERNAL_RAM * 4, mCORE_MEMORY_RW | mCORE_MEMORY_MAPPED, 3 },
 	{ GB_REGION_WORKING_RAM_BANK0, "wram", "WRAM", "Working RAM (8kiB)", GB_BASE_WORKING_RAM_BANK0, GB_BASE_WORKING_RAM_BANK0 + GB_SIZE_WORKING_RAM_BANK0 * 2, GB_SIZE_WORKING_RAM_BANK0 * 8, mCORE_MEMORY_RW | mCORE_MEMORY_MAPPED, 7, GB_BASE_WORKING_RAM_BANK0 + GB_SIZE_WORKING_RAM_BANK0 },
@@ -63,15 +63,19 @@ static const struct mCoreMemoryBlock _GBCMemoryBlocks[] = {
 struct mVideoLogContext;
 struct GBCore {
 	struct mCore d;
+	struct GBVideoRenderer dummyRenderer;
 	struct GBVideoSoftwareRenderer renderer;
+#ifndef MINIMAL_CORE
 	struct GBVideoProxyRenderer proxyRenderer;
 	struct mVideoLogContext* logContext;
+#endif
 	struct mCoreCallbacks logCallbacks;
 	uint8_t keys;
 	struct mCPUComponent* components[CPU_COMPONENT_MAX];
 	const struct Configuration* overrides;
 	struct mDebuggerPlatform* debuggerPlatform;
 	struct mCheatDevice* cheatDevice;
+	struct mCoreMemoryBlock memoryBlocks[8];
 };
 
 static bool _GBCoreInit(struct mCore* core) {
@@ -90,6 +94,10 @@ static bool _GBCoreInit(struct mCore* core) {
 	gbcore->overrides = NULL;
 	gbcore->debuggerPlatform = NULL;
 	gbcore->cheatDevice = NULL;
+#ifndef MINIMAL_CORE
+	gbcore->logContext = NULL;
+#endif
+	memcpy(gbcore->memoryBlocks, _GBMemoryBlocks, sizeof(_GBMemoryBlocks));
 
 	GBCreate(gb);
 	memset(gbcore->components, 0, sizeof(gbcore->components));
@@ -98,9 +106,16 @@ static bool _GBCoreInit(struct mCore* core) {
 	mRTCGenericSourceInit(&core->rtc, core);
 	gb->memory.rtc = &core->rtc.d;
 
+	GBVideoDummyRendererCreate(&gbcore->dummyRenderer);
+	GBVideoAssociateRenderer(&gb->video, &gbcore->dummyRenderer);
+
 	GBVideoSoftwareRendererCreate(&gbcore->renderer);
 	gbcore->renderer.outputBuffer = NULL;
 
+#ifndef MINIMAL_CORE
+	gbcore->proxyRenderer.logger = NULL;
+#endif
+
 	gbcore->keys = 0;
 	gb->keySource = &gbcore->keys;
 
@@ -130,14 +145,13 @@ static void _GBCoreDeinit(struct mCore* core) {
 	if (gbcore->cheatDevice) {
 		mCheatDeviceDestroy(gbcore->cheatDevice);
 	}
-	free(gbcore->cheatDevice);
 	mCoreConfigFreeOpts(&core->opts);
 	free(core);
 }
 
 static enum mPlatform _GBCorePlatform(const struct mCore* core) {
 	UNUSED(core);
-	return PLATFORM_GB;
+	return mPLATFORM_GB;
 }
 
 static bool _GBCoreSupportsFeature(const struct mCore* core, enum mCoreFeature feature) {
@@ -208,6 +222,8 @@ static void _GBCoreLoadConfig(struct mCore* core, const struct mCoreConfig* conf
 	mCoreConfigCopyValue(&core->config, config, "gb.model");
 	mCoreConfigCopyValue(&core->config, config, "sgb.model");
 	mCoreConfigCopyValue(&core->config, config, "cgb.model");
+	mCoreConfigCopyValue(&core->config, config, "cgb.hybridModel");
+	mCoreConfigCopyValue(&core->config, config, "cgb.sgbModel");
 	mCoreConfigCopyValue(&core->config, config, "useCgbColors");
 	mCoreConfigCopyValue(&core->config, config, "allowOpposingDirections");
 
@@ -285,8 +301,8 @@ static void _GBCoreReloadConfigOption(struct mCore* core, const char* option, co
 	}
 }
 
-static void _GBCoreDesiredVideoDimensions(struct mCore* core, unsigned* width, unsigned* height) {
-	struct GB* gb = core->board;
+static void _GBCoreDesiredVideoDimensions(const struct mCore* core, unsigned* width, unsigned* height) {
+	const struct GB* gb = core->board;
 	if (gb && (!(gb->model & GB_MODEL_SGB) || !gb->video.sgbBorders)) {
 		*width = GB_VIDEO_HORIZONTAL_PIXELS;
 		*height = GB_VIDEO_VERTICAL_PIXELS;
@@ -406,7 +422,7 @@ static void _GBCoreUnloadROM(struct mCore* core) {
 static void _GBCoreChecksum(const struct mCore* core, void* data, enum mCoreChecksumType type) {
 	struct GB* gb = (struct GB*) core->board;
 	switch (type) {
-	case CHECKSUM_CRC32:
+	case mCHECKSUM_CRC32:
 		memcpy(data, &gb->romCrc32, sizeof(gb->romCrc32));
 		break;
 	}
@@ -427,22 +443,47 @@ static void _GBCoreReset(struct mCore* core) {
 		struct GBCartridgeOverride override;
 		const struct GBCartridge* cart = (const struct GBCartridge*) &gb->memory.rom[0x100];
 		override.headerCrc32 = doCrc32(cart, sizeof(*cart));
-		if (GBOverrideFind(gbcore->overrides, &override) || (doColorOverride && GBOverrideColorFind(&override))) {
+		bool modelOverride = GBOverrideFind(gbcore->overrides, &override) || (doColorOverride && GBOverrideColorFind(&override));
+		if (modelOverride) {
 			GBOverrideApply(gb, &override);
 		}
-	}
-
-	const char* modelGB = mCoreConfigGetValue(&core->config, "gb.model");
-	const char* modelCGB = mCoreConfigGetValue(&core->config, "cgb.model");
-	const char* modelSGB = mCoreConfigGetValue(&core->config, "sgb.model");
-	if (modelGB || modelCGB || modelSGB) {
-		GBDetectModel(gb);
-		if (gb->model == GB_MODEL_DMG && modelGB) {
-			gb->model = GBNameToModel(modelGB);
-		} else if ((gb->model & GB_MODEL_CGB) && modelCGB) {
-			gb->model = GBNameToModel(modelCGB);
-		} else if ((gb->model & GB_MODEL_SGB) && modelSGB) {
-			gb->model = GBNameToModel(modelSGB);
+		if (!modelOverride || override.model == GB_MODEL_AUTODETECT) {
+			const char* modelGB = mCoreConfigGetValue(&core->config, "gb.model");
+			const char* modelSGB = mCoreConfigGetValue(&core->config, "sgb.model");
+			const char* modelCGB = mCoreConfigGetValue(&core->config, "cgb.model");
+			const char* modelCGBHybrid = mCoreConfigGetValue(&core->config, "cgb.hybridModel");
+			const char* modelCGBSGB = mCoreConfigGetValue(&core->config, "cgb.sgbModel");
+			if (modelGB || modelCGB || modelSGB || modelCGBHybrid || modelCGBSGB) {
+				int models = GBValidModels(gb->memory.rom);
+				switch (models) {
+				case GB_MODEL_SGB | GB_MODEL_MGB:
+					if (modelSGB) {
+						gb->model = GBNameToModel(modelSGB);
+					}
+					break;
+				case GB_MODEL_MGB:
+					if (modelGB) {
+						gb->model = GBNameToModel(modelGB);
+					}
+					break;
+				case GB_MODEL_MGB | GB_MODEL_CGB:
+					if (modelCGBHybrid) {
+						gb->model = GBNameToModel(modelCGBHybrid);
+					}
+					break;
+				case GB_MODEL_SGB | GB_MODEL_CGB: // TODO: Do these even exist?
+				case GB_MODEL_MGB | GB_MODEL_SGB | GB_MODEL_CGB:
+					if (modelCGBSGB) {
+						gb->model = GBNameToModel(modelCGBSGB);
+					}
+					break;
+				case GB_MODEL_CGB:
+					if (modelCGB) {
+						gb->model = GBNameToModel(modelCGB);
+					}
+					break;
+				}
+			}
 		}
 	}
 
@@ -522,6 +563,26 @@ static void _GBCoreReset(struct mCore* core) {
 	}
 #endif
 
+	if (gb->model < GB_MODEL_CGB) {
+		memcpy(gbcore->memoryBlocks, _GBMemoryBlocks, sizeof(_GBMemoryBlocks));
+	} else {
+		memcpy(gbcore->memoryBlocks, _GBCMemoryBlocks, sizeof(_GBCMemoryBlocks));
+	}
+
+	size_t i;
+	for (i = 0; i < sizeof(gbcore->memoryBlocks) / sizeof(*gbcore->memoryBlocks); ++i) {
+		if (gbcore->memoryBlocks[i].id == GB_REGION_CART_BANK0) {
+			gbcore->memoryBlocks[i].maxSegment = gb->memory.romSize / GB_SIZE_CART_BANK0;
+		} else if (gbcore->memoryBlocks[i].id == GB_REGION_EXTERNAL_RAM) {
+			gbcore->memoryBlocks[i].maxSegment = gb->sramSize / GB_SIZE_EXTERNAL_RAM;
+		} else {
+			continue;
+		}
+		if (gbcore->memoryBlocks[i].maxSegment) {
+			--gbcore->memoryBlocks[i].maxSegment;
+		}
+	}
+
 	SM83Reset(core->cpu);
 
 	if (core->opts.skipBios) {
@@ -695,20 +756,9 @@ static void _GBCoreRawWrite32(struct mCore* core, uint32_t address, int segment,
 }
 
 size_t _GBListMemoryBlocks(const struct mCore* core, const struct mCoreMemoryBlock** blocks) {
-	const struct GB* gb = core->board;
-	switch (gb->model) {
-	case GB_MODEL_DMG:
-	case GB_MODEL_MGB:
-	case GB_MODEL_SGB:
-	case GB_MODEL_SGB2:
-	default:
-		*blocks = _GBMemoryBlocks;
-		return sizeof(_GBMemoryBlocks) / sizeof(*_GBMemoryBlocks);
-	case GB_MODEL_CGB:
-	case GB_MODEL_AGB:
-		*blocks = _GBCMemoryBlocks;
-		return sizeof(_GBCMemoryBlocks) / sizeof(*_GBCMemoryBlocks);
-	}
+	struct GBCore* gbcore = (struct GBCore*) core;
+	*blocks = gbcore->memoryBlocks;
+	return sizeof(gbcore->memoryBlocks) / sizeof(*gbcore->memoryBlocks);
 }
 
 void* _GBGetMemoryBlock(struct mCore* core, size_t id, size_t* sizeOut) {
@@ -721,13 +771,13 @@ void* _GBGetMemoryBlock(struct mCore* core, size_t id, size_t* sizeOut) {
 		*sizeOut = gb->memory.romSize;
 		return gb->memory.rom;
 	case GB_REGION_VRAM:
-		*sizeOut = GB_SIZE_WORKING_RAM_BANK0 * (isCgb ? 1 : 2);
+		*sizeOut = GB_SIZE_VRAM_BANK0 * (isCgb ? 1 : 2);
 		return gb->video.vram;
 	case GB_REGION_EXTERNAL_RAM:
 		*sizeOut = gb->sramSize;
 		return gb->memory.sram;
 	case GB_REGION_WORKING_RAM_BANK0:
-		*sizeOut = GB_SIZE_VRAM * (isCgb ? 8 : 2);
+		*sizeOut = GB_SIZE_WORKING_RAM_BANK0 * (isCgb ? 8 : 2);
 		return gb->memory.wram;
 	case GB_BASE_OAM:
 		*sizeOut = GB_SIZE_OAM;
@@ -800,7 +850,7 @@ static bool _GBCoreLookupIdentifier(struct mCore* core, const char* name, int32_
 	UNUSED(core);
 	*segment = -1;
 	int i;
-	for (i = 0; i < REG_MAX; ++i) {
+	for (i = 0; i < GB_REG_MAX; ++i) {
 		const char* reg = GBIORegisterNames[i];
 		if (reg && strcasecmp(reg, name) == 0) {
 			*value = GB_BASE_IO | i;
@@ -878,13 +928,13 @@ static size_t _GBCoreListAudioChannels(const struct mCore* core, const struct mC
 static void _GBCoreEnableVideoLayer(struct mCore* core, size_t id, bool enable) {
 	struct GB* gb = core->board;
 	switch (id) {
-	case 0:
+	case GB_LAYER_BACKGROUND:
 		gb->video.renderer->disableBG = !enable;
 		break;
-	case 1:
+	case GB_LAYER_WINDOW:
 		gb->video.renderer->disableWIN = !enable;
 		break;
-	case 2:
+	case GB_LAYER_OBJ:
 		gb->video.renderer->disableOBJ = !enable;
 		break;
 	default:
@@ -909,15 +959,15 @@ static void _GBCoreEnableAudioChannel(struct mCore* core, size_t id, bool enable
 static void _GBCoreAdjustVideoLayer(struct mCore* core, size_t id, int32_t x, int32_t y) {
 	struct GBCore* gbcore = (struct GBCore*) core;
 	switch (id) {
-	case 0:
+	case GB_LAYER_BACKGROUND:
 		gbcore->renderer.offsetScx = x;
 		gbcore->renderer.offsetScy = y;
 		break;
-	case 1:
+	case GB_LAYER_WINDOW:
 		gbcore->renderer.offsetWx = x;
 		gbcore->renderer.offsetWy = y;
 		break;
-	case 2:
+	case GB_LAYER_OBJ:
 		gbcore->renderer.objOffsetX = x;
 		gbcore->renderer.objOffsetY = y;
 		break;
@@ -1074,7 +1124,7 @@ static bool _GBVLPInit(struct mCore* core) {
 static void _GBVLPDeinit(struct mCore* core) {
 	struct GBCore* gbcore = (struct GBCore*) core;
 	if (gbcore->logContext) {
-		mVideoLogContextDestroy(core, gbcore->logContext);
+		mVideoLogContextDestroy(core, gbcore->logContext, true);
 	}
 	_GBCoreDeinit(core);
 }
@@ -1094,16 +1144,16 @@ static void _GBVLPReset(struct mCore* core) {
 	GBVideoProxyRendererShim(&gb->video, &gbcore->proxyRenderer);
 
 	// Make sure CPU loop never spins
-	GBHalt(gb->cpu);
 	gb->memory.ie = 0;
 	gb->memory.ime = false;
+	GBHalt(gb->cpu);
 }
 
 static bool _GBVLPLoadROM(struct mCore* core, struct VFile* vf) {
 	struct GBCore* gbcore = (struct GBCore*) core;
 	gbcore->logContext = mVideoLogContextCreate(NULL);
 	if (!mVideoLogContextLoad(gbcore->logContext, vf)) {
-		mVideoLogContextDestroy(core, gbcore->logContext);
+		mVideoLogContextDestroy(core, gbcore->logContext, false);
 		gbcore->logContext = NULL;
 		return false;
 	}
@@ -1125,11 +1175,14 @@ static bool _GBVLPLoadState(struct mCore* core, const void* buffer) {
 	GBVideoDeserialize(&gb->video, state);
 	GBIODeserialize(gb, state);
 	GBAudioReset(&gb->audio);
+	if (gb->model & GB_MODEL_SGB) {
+		GBSGBDeserialize(gb, state);
+	}
 
 	// Make sure CPU loop never spins
-	GBHalt(gb->cpu);
 	gb->memory.ie = 0;
 	gb->memory.ime = false;
+	GBHalt(gb->cpu);
 
 	return true;
 }
