@@ -36,6 +36,9 @@ static void _GBHuC3(struct GB*, uint16_t address, uint8_t value);
 static void _GBPocketCam(struct GB* gb, uint16_t address, uint8_t value);
 static void _GBTAMA5(struct GB* gb, uint16_t address, uint8_t value);
 static void _GBWisdomTree(struct GB* gb, uint16_t address, uint8_t value);
+static void _GBPKJD(struct GB* gb, uint16_t address, uint8_t value);
+static void _GBBBD(struct GB* gb, uint16_t address, uint8_t value);
+static void _GBHitek(struct GB* gb, uint16_t address, uint8_t value);
 
 static uint8_t _GBMBC2Read(struct GBMemory*, uint16_t address);
 static uint8_t _GBMBC6Read(struct GBMemory*, uint16_t address);
@@ -43,10 +46,15 @@ static uint8_t _GBMBC7Read(struct GBMemory*, uint16_t address);
 static void _GBMBC7Write(struct GBMemory* memory, uint16_t address, uint8_t value);
 
 static uint8_t _GBTAMA5Read(struct GBMemory*, uint16_t address);
+static uint8_t _GBPKJDRead(struct GBMemory*, uint16_t address);
+static uint8_t _GBBBDRead(struct GBMemory*, uint16_t address);
+static uint8_t _GBHitekRead(struct GBMemory*, uint16_t address);
 
 static uint8_t _GBPocketCamRead(struct GBMemory*, uint16_t address);
 static void _GBPocketCamCapture(struct GBMemory*);
 
+static void _GBMBC6MapChip(struct GB*, int half, uint8_t value);
+
 void GBMBCSwitchBank(struct GB* gb, int bank) {
 	size_t bankStart = bank * GB_SIZE_CART_BANK0;
 	if (bankStart + GB_SIZE_CART_BANK0 > gb->memory.romSize) {
@@ -68,6 +76,7 @@ void GBMBCSwitchBank0(struct GB* gb, int bank) {
 		bankStart &= (gb->memory.romSize - 1);
 	}
 	gb->memory.romBase = &gb->memory.rom[bankStart];
+	gb->memory.currentBank0 = bank;
 	if (gb->cpu->pc < GB_SIZE_CART_BANK0) {
 		gb->cpu->memory.setActiveRegion(gb->cpu, gb->cpu->pc);
 	}
@@ -75,19 +84,37 @@ void GBMBCSwitchBank0(struct GB* gb, int bank) {
 
 void GBMBCSwitchHalfBank(struct GB* gb, int half, int bank) {
 	size_t bankStart = bank * GB_SIZE_CART_HALFBANK;
-	if (bankStart + GB_SIZE_CART_HALFBANK > gb->memory.romSize) {
-		mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid ROM bank: %0X", bank);
-		bankStart &= (gb->memory.romSize - 1);
-		bank = bankStart / GB_SIZE_CART_HALFBANK;
-		if (!bank) {
-			++bank;
+	bool isFlash = half ? gb->memory.mbcState.mbc6.flashBank1 : gb->memory.mbcState.mbc6.flashBank0;
+	if (isFlash) {
+		if (bankStart + GB_SIZE_CART_HALFBANK > GB_SIZE_MBC6_FLASH) {
+			mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid Flash bank: %0X", bank);
+			bankStart &= GB_SIZE_MBC6_FLASH - 1;
+			bank = bankStart / GB_SIZE_CART_HALFBANK;
+		}
+		bankStart += gb->sramSize - GB_SIZE_MBC6_FLASH;
+	} else {
+		if (bankStart + GB_SIZE_CART_HALFBANK > gb->memory.romSize) {
+			mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid ROM bank: %0X", bank);
+			bankStart &= gb->memory.romSize - 1;
+			bank = bankStart / GB_SIZE_CART_HALFBANK;
+			if (!bank) {
+				++bank;
+			}
 		}
 	}
 	if (!half) {
-		gb->memory.romBank = &gb->memory.rom[bankStart];
+		if (isFlash) {
+			gb->memory.romBank = &gb->memory.sram[bankStart];
+		} else {
+			gb->memory.romBank = &gb->memory.rom[bankStart];
+		}
 		gb->memory.currentBank = bank;
 	} else {
-		gb->memory.mbcState.mbc6.romBank1 = &gb->memory.rom[bankStart];
+		if (isFlash) {
+			gb->memory.mbcState.mbc6.romBank1 = &gb->memory.sram[bankStart];
+		} else {
+			gb->memory.mbcState.mbc6.romBank1 = &gb->memory.rom[bankStart];
+		}
 		gb->memory.mbcState.mbc6.currentBank1 = bank;
 	}
 	if (gb->cpu->pc < GB_BASE_VRAM) {
@@ -143,6 +170,31 @@ static bool _isWisdomTree(const uint8_t* mem, size_t size) {
 	return false;
 }
 
+static enum GBMemoryBankControllerType _detectUnlMBC(const uint8_t* mem, size_t size) {
+	const struct GBCartridge* cart = (const struct GBCartridge*) &mem[0x100];
+
+	switch (cart->type) {
+		case 0:
+		if (_isWisdomTree(mem, size)) {
+			return GB_UNL_WISDOM_TREE;
+		}
+		break;
+	}
+
+	uint32_t secondaryLogo = doCrc32(&mem[0x184], 0x30);
+	switch (secondaryLogo) {
+	case 0x4fdab691:
+		return GB_UNL_HITEK;
+	case 0xc7d8c1df:
+	case 0x6d1ea662: // Garou
+		if (mem[0x7FFF] != 0x01) { // Make sure we're not using a "fixed" version
+			return GB_UNL_BBD;
+		}
+	}
+
+	return GB_MBC_AUTODETECT;
+}
+
 void GBMBCSwitchSramBank(struct GB* gb, int bank) {
 	size_t bankStart = bank * GB_SIZE_EXTERNAL_RAM;
 	if (bankStart + GB_SIZE_EXTERNAL_RAM > gb->sramSize) {
@@ -156,9 +208,10 @@ void GBMBCSwitchSramBank(struct GB* gb, int bank) {
 
 void GBMBCSwitchSramHalfBank(struct GB* gb, int half, int bank) {
 	size_t bankStart = bank * GB_SIZE_EXTERNAL_RAM_HALFBANK;
-	if (bankStart + GB_SIZE_EXTERNAL_RAM_HALFBANK > gb->sramSize) {
+	size_t sramSize = gb->sramSize - GB_SIZE_MBC6_FLASH;
+	if (bankStart + GB_SIZE_EXTERNAL_RAM_HALFBANK > sramSize) {
 		mLOG(GB_MBC, GAME_ERROR, "Attempting to switch to an invalid RAM bank: %0X", bank);
-		bankStart &= (gb->sramSize - 1);
+		bankStart &= (sramSize - 1);
 		bank = bankStart / GB_SIZE_EXTERNAL_RAM_HALFBANK;
 	}
 	if (!half) {
@@ -172,7 +225,7 @@ void GBMBCSwitchSramHalfBank(struct GB* gb, int half, int bank) {
 
 void GBMBCInit(struct GB* gb) {
 	const struct GBCartridge* cart = (const struct GBCartridge*) &gb->memory.rom[0x100];
-	if (gb->memory.rom) {
+	if (gb->memory.rom && gb->memory.romSize) {
 		if (gb->memory.romSize >= 0x8000) {
 			const struct GBCartridge* cartFooter = (const struct GBCartridge*) &gb->memory.rom[gb->memory.romSize - 0x7F00];
 			if (doCrc32(cartFooter->logo, sizeof(cartFooter->logo)) == GB_LOGO_HASH && cartFooter->type >= 0x0B && cartFooter->type <= 0x0D) {
@@ -183,9 +236,6 @@ void GBMBCInit(struct GB* gb) {
 		case 0:
 			gb->sramSize = 0;
 			break;
-		case 1:
-			gb->sramSize = 0x800;
-			break;
 		default:
 		case 2:
 			gb->sramSize = 0x2000;
@@ -200,15 +250,13 @@ void GBMBCInit(struct GB* gb) {
 			gb->sramSize = 0x10000;
 			break;
 		}
+		if (gb->memory.mbcType == GB_MBC_AUTODETECT) {
+			gb->memory.mbcType = _detectUnlMBC(gb->memory.rom, gb->memory.romSize);
+		}
 
 		if (gb->memory.mbcType == GB_MBC_AUTODETECT) {
 			switch (cart->type) {
 			case 0:
-				if (_isWisdomTree(gb->memory.rom, gb->memory.romSize)) {
-					gb->memory.mbcType = GB_UNL_WISDOM_TREE;
-					break;
-				}
-				// Fall through
 			case 8:
 			case 9:
 				gb->memory.mbcType = GB_MBC_NONE;
@@ -273,6 +321,7 @@ void GBMBCInit(struct GB* gb) {
 		gb->memory.mbcType = GB_MBC_NONE;
 	}
 	gb->memory.mbcRead = NULL;
+	gb->memory.directSramAccess = true;
 	switch (gb->memory.mbcType) {
 	case GB_MBC_NONE:
 		gb->memory.mbcWrite = _GBMBCNone;
@@ -288,6 +337,7 @@ void GBMBCInit(struct GB* gb) {
 	case GB_MBC2:
 		gb->memory.mbcWrite = _GBMBC2;
 		gb->memory.mbcRead = _GBMBC2Read;
+		gb->memory.directSramAccess = false;
 		gb->sramSize = 0x100;
 		break;
 	case GB_MBC3:
@@ -300,9 +350,13 @@ void GBMBCInit(struct GB* gb) {
 		gb->memory.mbcWrite = _GBMBC5;
 		break;
 	case GB_MBC6:
-		mLOG(GB_MBC, WARN, "unimplemented MBC: MBC6");
 		gb->memory.mbcWrite = _GBMBC6;
 		gb->memory.mbcRead = _GBMBC6Read;
+		gb->memory.directSramAccess = false;
+		if (!gb->sramSize) {
+			gb->sramSize = GB_SIZE_EXTERNAL_RAM; // Force minimum size for convenience
+		}
+		gb->sramSize += GB_SIZE_MBC6_FLASH; // Flash is concatenated at the end
 		break;
 	case GB_MBC7:
 		gb->memory.mbcWrite = _GBMBC7;
@@ -335,6 +389,9 @@ void GBMBCInit(struct GB* gb) {
 	case GB_POCKETCAM:
 		gb->memory.mbcWrite = _GBPocketCam;
 		gb->memory.mbcRead = _GBPocketCamRead;
+		if (!gb->sramSize) {
+			gb->sramSize = GB_SIZE_EXTERNAL_RAM; // Force minimum size for convenience
+		}
 		if (gb->memory.cam && gb->memory.cam->startRequestImage) {
 			gb->memory.cam->startRequestImage(gb->memory.cam, GBCAM_WIDTH, GBCAM_HEIGHT, mCOLOR_ANY);
 		}
@@ -342,6 +399,20 @@ void GBMBCInit(struct GB* gb) {
 	case GB_UNL_WISDOM_TREE:
 		gb->memory.mbcWrite = _GBWisdomTree;
 		break;
+	case GB_UNL_BBD:
+		gb->memory.mbcWrite = _GBBBD;
+		gb->memory.mbcRead = _GBBBDRead;
+		break;
+	case GB_UNL_HITEK:
+		gb->memory.mbcWrite = _GBHitek;
+		gb->memory.mbcRead = _GBHitekRead;
+		gb->memory.mbcState.bbd.dataSwapMode = 7;
+		gb->memory.mbcState.bbd.bankSwapMode = 7;
+		break;
+	case GB_UNL_PKJD:
+		gb->memory.mbcWrite = _GBPKJD;
+		gb->memory.mbcRead = _GBPKJDRead;
+		break;
 	}
 
 	gb->memory.currentBank = 1;
@@ -368,6 +439,34 @@ void GBMBCInit(struct GB* gb) {
 	}
 }
 
+void GBMBCReset(struct GB* gb) {
+	gb->memory.currentBank0 = 0;
+	gb->memory.romBank = &gb->memory.rom[GB_SIZE_CART_BANK0];
+
+	memset(&gb->memory.mbcState, 0, sizeof(gb->memory.mbcState));
+	GBMBCInit(gb);
+	switch (gb->memory.mbcType) {
+	case GB_MBC1:
+		gb->memory.mbcState.mbc1.mode = 0;
+		gb->memory.mbcState.mbc1.bankLo = 1;
+		break;
+	case GB_MBC6:
+		GBMBCSwitchHalfBank(gb, 0, 2);
+		GBMBCSwitchHalfBank(gb, 1, 3);
+		gb->memory.mbcState.mbc6.sramAccess = false;
+		GBMBCSwitchSramHalfBank(gb, 0, 0);
+		GBMBCSwitchSramHalfBank(gb, 0, 1);
+		break;
+	case GB_MMM01:
+		GBMBCSwitchBank0(gb, gb->memory.romSize / GB_SIZE_CART_BANK0 - 2);
+		GBMBCSwitchBank(gb, gb->memory.romSize / GB_SIZE_CART_BANK0 - 1);
+		break;
+	default:
+		break;
+	}
+	gb->memory.sramBank = gb->memory.sram;
+}
+
 static void _latchRtc(struct mRTCSource* rtc, uint8_t* rtcRegs, time_t* rtcLastLatch) {
 	time_t t;
 	if (rtc) {
@@ -419,10 +518,28 @@ static void _latchRtc(struct mRTCSource* rtc, uint8_t* rtcRegs, time_t* rtcLastL
 	}
 }
 
+static void _GBMBC1Update(struct GB* gb) {
+	struct GBMBC1State* state = &gb->memory.mbcState.mbc1;
+	int bank = state->bankLo;
+	bank &= (1 << state->multicartStride) - 1;
+	bank |= state->bankHi << state->multicartStride;
+	if (state->mode) {
+		GBMBCSwitchBank0(gb, state->bankHi << state->multicartStride);
+		GBMBCSwitchSramBank(gb, state->bankHi & 3);
+	} else {
+		GBMBCSwitchBank0(gb, 0);
+		GBMBCSwitchSramBank(gb, 0);
+	}
+	if (!(state->bankLo & 0x1F)) {
+		++state->bankLo;
+		++bank;
+	}
+	GBMBCSwitchBank(gb, bank);
+}
+
 void _GBMBC1(struct GB* gb, uint16_t address, uint8_t value) {
 	struct GBMemory* memory = &gb->memory;
 	int bank = value & 0x1F;
-	int stride = 1 << memory->mbcState.mbc1.multicartStride;
 	switch (address >> 13) {
 	case 0x0:
 		switch (value & 0xF) {
@@ -440,28 +557,17 @@ void _GBMBC1(struct GB* gb, uint16_t address, uint8_t value) {
 		}
 		break;
 	case 0x1:
-		if (!bank) {
-			++bank;
-		}
-		bank &= stride - 1;
-		GBMBCSwitchBank(gb, bank | (memory->currentBank & (3 * stride)));
+		memory->mbcState.mbc1.bankLo = bank;
+		_GBMBC1Update(gb);
 		break;
 	case 0x2:
 		bank &= 3;
-		if (memory->mbcState.mbc1.mode) {
-			GBMBCSwitchBank0(gb, bank << gb->memory.mbcState.mbc1.multicartStride);
-			GBMBCSwitchSramBank(gb, bank);
-		}
-		GBMBCSwitchBank(gb, (bank << memory->mbcState.mbc1.multicartStride) | (memory->currentBank & (stride - 1)));
+		memory->mbcState.mbc1.bankHi = bank;
+		_GBMBC1Update(gb);
 		break;
 	case 0x3:
 		memory->mbcState.mbc1.mode = value & 1;
-		if (memory->mbcState.mbc1.mode) {
-			GBMBCSwitchBank0(gb, memory->currentBank & ~((1 << memory->mbcState.mbc1.multicartStride) - 1));
-		} else {
-			GBMBCSwitchBank0(gb, 0);
-			GBMBCSwitchSramBank(gb, 0);
-		}
+		_GBMBC1Update(gb);
 		break;
 	default:
 		// TODO
@@ -527,7 +633,7 @@ void _GBMBC3(struct GB* gb, uint16_t address, uint8_t value) {
 	int bank = value;
 	switch (address >> 13) {
 	case 0x0:
-		switch (value) {
+		switch (value & 0xF) {
 		case 0:
 			memory->sramAccess = false;
 			break;
@@ -551,11 +657,12 @@ void _GBMBC3(struct GB* gb, uint16_t address, uint8_t value) {
 		GBMBCSwitchBank(gb, bank);
 		break;
 	case 0x2:
-		if (value < 8) {
+		bank &= 0xF;
+		if (bank < 8) {
 			GBMBCSwitchSramBank(gb, value);
 			memory->rtcAccess = false;
-		} else if (value <= 0xC) {
-			memory->activeRtcReg = value - 8;
+		} else if (bank <= 0xC) {
+			memory->activeRtcReg = bank - 8;
 			memory->rtcAccess = true;
 		}
 		break;
@@ -620,10 +727,10 @@ void _GBMBC6(struct GB* gb, uint16_t address, uint8_t value) {
 	case 0:
 		switch (value) {
 		case 0:
-			memory->mbcState.mbc6.sramAccess = false;
+			memory->sramAccess = false;
 			break;
 		case 0xA:
-			memory->mbcState.mbc6.sramAccess = true;
+			memory->sramAccess = true;
 			break;
 		default:
 			// TODO
@@ -637,19 +744,33 @@ void _GBMBC6(struct GB* gb, uint16_t address, uint8_t value) {
 	case 0x2:
 		GBMBCSwitchSramHalfBank(gb, 1, bank);
 		break;
+	case 0x3:
+		mLOG(GB_MBC, STUB, "MBC6 unimplemented flash OE write: %04X:%02X", address, value);
+		break;
+	case 0x4:
+		mLOG(GB_MBC, STUB, "MBC6 unimplemented flash WE write: %04X:%02X", address, value);
+		break;
 	case 0x8:
 	case 0x9:
 		GBMBCSwitchHalfBank(gb, 0, bank);
 		break;
+	case 0xA:
+	case 0xB:
+		_GBMBC6MapChip(gb, 0, value);
+		break;
 	case 0xC:
 	case 0xD:
 		GBMBCSwitchHalfBank(gb, 1, bank);
 		break;
+	case 0xE:
+	case 0xF:
+		_GBMBC6MapChip(gb, 1, value);
+		break;
 	case 0x28:
 	case 0x29:
 	case 0x2A:
 	case 0x2B:
-		if (memory->mbcState.mbc6.sramAccess) {
+		if (memory->sramAccess) {
 			memory->sramBank[address & (GB_SIZE_EXTERNAL_RAM_HALFBANK - 1)] = value;
 		}
 		break;
@@ -657,7 +778,7 @@ void _GBMBC6(struct GB* gb, uint16_t address, uint8_t value) {
 	case 0x2D:
 	case 0x2E:
 	case 0x2F:
-		if (memory->mbcState.mbc6.sramAccess) {
+		if (memory->sramAccess) {
 			memory->mbcState.mbc6.sramBank1[address & (GB_SIZE_EXTERNAL_RAM_HALFBANK - 1)] = value;
 		}
 		break;
@@ -668,7 +789,7 @@ void _GBMBC6(struct GB* gb, uint16_t address, uint8_t value) {
 }
 
 uint8_t _GBMBC6Read(struct GBMemory* memory, uint16_t address) {
-	if (!memory->mbcState.mbc6.sramAccess) {
+	if (!memory->sramAccess) {
 		return 0xFF;
 	}
 	switch (address >> 12) {
@@ -680,6 +801,16 @@ uint8_t _GBMBC6Read(struct GBMemory* memory, uint16_t address) {
 	return 0xFF;
 }
 
+static void _GBMBC6MapChip(struct GB* gb, int half, uint8_t value) {
+	if (!half) {
+		gb->memory.mbcState.mbc6.flashBank0 = !!(value & 0x08);
+		GBMBCSwitchHalfBank(gb, half, gb->memory.currentBank);
+	} else {
+		gb->memory.mbcState.mbc6.flashBank1 = !!(value & 0x08);
+		GBMBCSwitchHalfBank(gb, half, gb->memory.mbcState.mbc6.currentBank1);
+	}
+}
+
 void _GBMBC7(struct GB* gb, uint16_t address, uint8_t value) {
 	int bank = value & 0x7F;
 	switch (address >> 13) {
@@ -1223,6 +1354,191 @@ void _GBWisdomTree(struct GB* gb, uint16_t address, uint8_t value) {
 	}
 }
 
+void _GBPKJD(struct GB* gb, uint16_t address, uint8_t value) {
+	struct GBMemory* memory = &gb->memory;
+	switch (address >> 13) {
+	case 0x2:
+		if (value < 8) {
+			memory->directSramAccess = true;
+			memory->activeRtcReg = 0;
+		} else if (value >= 0xD && value <= 0xF) {
+			memory->directSramAccess = false;
+			memory->rtcAccess = false;
+			memory->activeRtcReg = value - 8;
+		}
+		break;
+	case 0x5:
+		if (!memory->sramAccess) {
+			return;
+		}
+		switch (memory->activeRtcReg) {
+		case 0:
+			memory->sramBank[address & (GB_SIZE_EXTERNAL_RAM - 1)] = value;
+			break;
+		case 5:
+		case 6:
+			memory->mbcState.pkjd.reg[memory->activeRtcReg - 5] = value;
+			break;
+		case 7:
+			switch (value) {
+			case 0x11:
+				memory->mbcState.pkjd.reg[0]--;
+				break;
+			case 0x12:
+				memory->mbcState.pkjd.reg[1]--;
+				break;
+			case 0x41:
+				memory->mbcState.pkjd.reg[0] += memory->mbcState.pkjd.reg[1];
+				break;
+			case 0x42:
+				memory->mbcState.pkjd.reg[1] += memory->mbcState.pkjd.reg[0];
+				break;
+			case 0x51:
+				memory->mbcState.pkjd.reg[0]++;
+				break;
+			case 0x52:
+				memory->mbcState.pkjd.reg[1]--;
+				break;
+			}
+			break;
+		}
+		return;
+	}
+	_GBMBC3(gb, address, value);
+}
+
+static uint8_t _GBPKJDRead(struct GBMemory* memory, uint16_t address) {
+	if (!memory->sramAccess) {
+		return 0xFF;
+	}
+	switch (memory->activeRtcReg) {
+	case 0:
+		return memory->sramBank[address & (GB_SIZE_EXTERNAL_RAM - 1)];
+	case 5:
+	case 6:
+		return memory->mbcState.pkjd.reg[memory->activeRtcReg - 5];
+	default:
+		return 0;
+	}
+}
+
+static uint8_t _reorderBits(uint8_t input, const uint8_t* reorder) {
+	uint8_t newbyte = 0;
+	int i;
+	for(i = 0; i < 8; ++i) {
+		int oldbit = reorder[i];
+		int newbit = i;
+		newbyte += ((input >> oldbit) & 1) << newbit;
+	}
+
+	return newbyte;
+}
+
+static const uint8_t _bbdDataReordering[8][8] = {
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 00 - Normal
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 01 - NOT KNOWN YET
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 02 - NOT KNOWN YET
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 03 - NOT KNOWN YET
+	{ 0, 5, 1, 3, 4, 2, 6, 7 }, // 04 - Garou
+	{ 0, 4, 2, 3, 1, 5, 6, 7 }, // 05 - Harry
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 06 - NOT KNOWN YET
+	{ 0, 1, 5, 3, 4, 2, 6, 7 }, // 07 - Digimon
+};
+
+static const uint8_t _bbdBankReordering[8][8] = {
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 00 - Normal
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 01 - NOT KNOWN YET
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 02 - NOT KNOWN YET
+	{ 3, 4, 2, 0, 1, 5, 6, 7 }, // 03 - 0,1 unconfirmed. Digimon/Garou
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 04 - NOT KNOWN YET
+	{ 1, 2, 3, 4, 0, 5, 6, 7 }, // 05 - 0,1 unconfirmed. Harry
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 06 - NOT KNOWN YET
+	{ 0, 1, 2, 3, 4, 5, 6, 7 }, // 07 - NOT KNOWN YET
+};
+
+void  _GBBBD(struct GB* gb, uint16_t address, uint8_t value) {
+	struct GBMemory* memory = &gb->memory;
+	switch (address & 0xF0FF) {
+	case 0x2000:
+		value = _reorderBits(value, _bbdBankReordering[memory->mbcState.bbd.bankSwapMode]);
+		break;
+	case 0x2001:
+		memory->mbcState.bbd.dataSwapMode = value & 0x07;
+		if (!(memory->mbcState.bbd.dataSwapMode == 0x07 || memory->mbcState.bbd.dataSwapMode == 0x05 || memory->mbcState.bbd.dataSwapMode == 0x04 || memory->mbcState.bbd.dataSwapMode == 0x00)) {
+			mLOG(GB_MBC, STUB, "Bitswap mode unsupported: %X", memory->mbcState.bbd.dataSwapMode);
+		}
+		break;
+	case 0x2080:
+		memory->mbcState.bbd.bankSwapMode = value & 0x07;
+		if (!(memory->mbcState.bbd.bankSwapMode == 0x03 || memory->mbcState.bbd.bankSwapMode == 0x05 || memory->mbcState.bbd.bankSwapMode == 0x00)) {
+			mLOG(GB_MBC, STUB, "Bankswap mode unsupported: %X", memory->mbcState.bbd.dataSwapMode);
+		}
+		break;
+	}
+	_GBMBC5(gb, address, value);
+}
+
+uint8_t _GBBBDRead(struct GBMemory* memory, uint16_t address) {
+	switch (address >> 14) {
+	case 0:
+	default:
+		return memory->romBank[address & (GB_SIZE_CART_BANK0 - 1)];
+	case 1:
+		return _reorderBits(memory->romBank[address & (GB_SIZE_CART_BANK0 - 1)], _bbdDataReordering[memory->mbcState.bbd.dataSwapMode]);
+	}
+}
+
+static const uint8_t _hitekDataReordering[8][8] = {
+	{ 0, 1, 2, 3, 4, 5, 6, 7 },
+	{ 0, 6, 5, 3, 4, 1, 2, 7 },
+	{ 0, 5, 6, 3, 4, 2, 1, 7 },
+	{ 0, 6, 2, 3, 4, 5, 1, 7 },
+	{ 0, 6, 1, 3, 4, 5, 2, 7 },
+	{ 0, 1, 6, 3, 4, 5, 2, 7 },
+	{ 0, 2, 6, 3, 4, 1, 5, 7 },
+	{ 0, 6, 2, 3, 4, 1, 5, 7 },
+};
+
+static const uint8_t _hitekBankReordering[8][8] = {
+	{ 0, 1, 2, 3, 4, 5, 6, 7 },
+	{ 3, 2, 1, 0, 4, 5, 6, 7 },
+	{ 2, 1, 0, 3, 4, 5, 6, 7 },
+	{ 1, 0, 3, 2, 4, 5, 6, 7 },
+	{ 0, 3, 2, 1, 4, 5, 6, 7 },
+	{ 2, 3, 0, 1, 4, 5, 6, 7 },
+	{ 3, 0, 1, 2, 4, 5, 6, 7 },
+	{ 2, 0, 3, 1, 4, 5, 6, 7 },
+};
+
+void  _GBHitek(struct GB* gb, uint16_t address, uint8_t value) {
+	struct GBMemory* memory = &gb->memory;
+	switch (address & 0xF0FF) {
+	case 0x2000:
+		value = _reorderBits(value, _hitekBankReordering[memory->mbcState.bbd.bankSwapMode]);
+		break;
+	case 0x2001:
+		memory->mbcState.bbd.dataSwapMode = value & 0x07;
+		break;
+	case 0x2080:
+		memory->mbcState.bbd.bankSwapMode = value & 0x07;
+		break;
+	case 0x300:
+		// See hhugboy src/memory/mbc/MbcUnlHitek.cpp for commentary on this return
+		return;	
+	}
+	_GBMBC5(gb, address, value);
+}
+
+uint8_t _GBHitekRead(struct GBMemory* memory, uint16_t address) {
+	switch (address >> 14) {
+	case 0:
+	default:
+		return memory->romBank[address & (GB_SIZE_CART_BANK0 - 1)];
+	case 1:
+		return _reorderBits(memory->romBank[address & (GB_SIZE_CART_BANK0 - 1)], _hitekDataReordering[memory->mbcState.bbd.dataSwapMode]);
+	}
+}
+
 void GBMBCRTCRead(struct GB* gb) {
 	struct GBMBCRTCSaveBuffer rtcBuffer;
 	struct VFile* vf = gb->sramVf;
