@@ -11,6 +11,15 @@
 #include <mgba-util/math.h>
 #include <mgba-util/memory.h>
 
+#define PAL_BG 0
+#define PAL_OBJ 0x20
+#define PAL_HIGHLIGHT 0x80
+#define PAL_HIGHLIGHT_BG (PAL_HIGHLIGHT | PAL_BG)
+#define PAL_HIGHLIGHT_OBJ (PAL_HIGHLIGHT | PAL_OBJ)
+#define PAL_SGB_BORDER 0x40
+#define OBJ_PRIORITY 0x100
+#define OBJ_PRIO_MASK 0x0FF
+
 static void GBVideoSoftwareRendererInit(struct GBVideoRenderer* renderer, enum GBModel model, bool borders);
 static void GBVideoSoftwareRendererDeinit(struct GBVideoRenderer* renderer);
 static uint8_t GBVideoSoftwareRendererWriteVideoRegister(struct GBVideoRenderer* renderer, uint16_t address, uint8_t value);
@@ -18,15 +27,15 @@ static void GBVideoSoftwareRendererWriteSGBPacket(struct GBVideoRenderer* render
 static void GBVideoSoftwareRendererWritePalette(struct GBVideoRenderer* renderer, int index, uint16_t value);
 static void GBVideoSoftwareRendererWriteVRAM(struct GBVideoRenderer* renderer, uint16_t address);
 static void GBVideoSoftwareRendererWriteOAM(struct GBVideoRenderer* renderer, uint16_t oam);
-static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, int startX, int endX, int y, struct GBObj* obj, size_t oamMax);
+static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, int startX, int endX, int y);
 static void GBVideoSoftwareRendererFinishScanline(struct GBVideoRenderer* renderer, int y);
 static void GBVideoSoftwareRendererFinishFrame(struct GBVideoRenderer* renderer);
 static void GBVideoSoftwareRendererEnableSGBBorder(struct GBVideoRenderer* renderer, bool enable);
 static void GBVideoSoftwareRendererGetPixels(struct GBVideoRenderer* renderer, size_t* stride, const void** pixels);
 static void GBVideoSoftwareRendererPutPixels(struct GBVideoRenderer* renderer, size_t stride, const void* pixels);
 
-static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer* renderer, uint8_t* maps, int startX, int endX, int sx, int sy);
-static void GBVideoSoftwareRendererDrawObj(struct GBVideoSoftwareRenderer* renderer, struct GBObj* obj, int startX, int endX, int y);
+static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer* renderer, uint8_t* maps, int startX, int endX, int sx, int sy, bool highlight);
+static void GBVideoSoftwareRendererDrawObj(struct GBVideoSoftwareRenderer* renderer, struct GBVideoRendererSprite* obj, int startX, int endX, int y);
 
 static void _clearScreen(struct GBVideoSoftwareRenderer* renderer) {
 	size_t sgbOffset = 0;
@@ -51,7 +60,7 @@ static void _regenerateSGBBorder(struct GBVideoSoftwareRenderer* renderer) {
 	for (i = 0; i < 0x40; ++i) {
 		uint16_t color;
 		LOAD_16LE(color, 0x800 + i * 2, renderer->d.sgbMapRam);
-		renderer->d.writePalette(&renderer->d, i + 0x40, color);
+		renderer->d.writePalette(&renderer->d, i + PAL_SGB_BORDER, color);
 	}
 	int x, y;
 	for (y = 0; y < 224; ++y) {
@@ -180,6 +189,15 @@ void GBVideoSoftwareRendererCreate(struct GBVideoSoftwareRenderer* renderer) {
 	renderer->d.disableOBJ = false;
 	renderer->d.disableWIN = false;
 
+	renderer->d.highlightBG = false;
+	renderer->d.highlightWIN = false;
+	int i;
+	for (i = 0; i < GB_VIDEO_MAX_OBJ; ++i) {
+		renderer->d.highlightOBJ[i] = false;
+	}
+	renderer->d.highlightColor = M_COLOR_WHITE;
+	renderer->d.highlightAmount = 0;
+
 	renderer->temporaryBuffer = 0;
 }
 
@@ -206,8 +224,8 @@ static void GBVideoSoftwareRendererInit(struct GBVideoRenderer* renderer, enum G
 	softwareRenderer->offsetWx = 0;
 	softwareRenderer->offsetWy = 0;
 
-	int i;
-	for (i = 0; i < 64; ++i) {
+	size_t i;
+	for (i = 0; i < (sizeof(softwareRenderer->lookup) / sizeof(*softwareRenderer->lookup)); ++i) {
 		softwareRenderer->lookup[i] = i;
 		softwareRenderer->lookup[i] = i;
 		softwareRenderer->lookup[i] = i;
@@ -215,6 +233,8 @@ static void GBVideoSoftwareRendererInit(struct GBVideoRenderer* renderer, enum G
 	}
 
 	memset(softwareRenderer->palette, 0, sizeof(softwareRenderer->palette));
+
+	softwareRenderer->lastHighlightAmount = 0;
 }
 
 static void GBVideoSoftwareRendererDeinit(struct GBVideoRenderer* renderer) {
@@ -257,41 +277,53 @@ static uint8_t GBVideoSoftwareRendererWriteVideoRegister(struct GBVideoRenderer*
 	bool wasWindow = _inWindow(softwareRenderer);
 	uint8_t wy = softwareRenderer->wy;
 	switch (address) {
-	case REG_LCDC:
+	case GB_REG_LCDC:
 		softwareRenderer->lcdc = value;
 		GBVideoSoftwareRendererUpdateWindow(softwareRenderer, wasWindow, _inWindow(softwareRenderer), wy);
 		break;
-	case REG_SCY:
+	case GB_REG_SCY:
 		softwareRenderer->scy = value;
 		break;
-	case REG_SCX:
+	case GB_REG_SCX:
 		softwareRenderer->scx = value;
 		break;
-	case REG_WY:
+	case GB_REG_WY:
 		softwareRenderer->wy = value;
 		GBVideoSoftwareRendererUpdateWindow(softwareRenderer, wasWindow, _inWindow(softwareRenderer), wy);
 		break;
-	case REG_WX:
+	case GB_REG_WX:
 		softwareRenderer->wx = value;
 		GBVideoSoftwareRendererUpdateWindow(softwareRenderer, wasWindow, _inWindow(softwareRenderer), wy);
 		break;
-	case REG_BGP:
+	case GB_REG_BGP:
 		softwareRenderer->lookup[0] = value & 3;
 		softwareRenderer->lookup[1] = (value >> 2) & 3;
 		softwareRenderer->lookup[2] = (value >> 4) & 3;
 		softwareRenderer->lookup[3] = (value >> 6) & 3;
+		softwareRenderer->lookup[PAL_HIGHLIGHT_BG + 0] = PAL_HIGHLIGHT + (value & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_BG + 1] = PAL_HIGHLIGHT + ((value >> 2) & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_BG + 2] = PAL_HIGHLIGHT + ((value >> 4) & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_BG + 3] = PAL_HIGHLIGHT + ((value >> 6) & 3);
 		break;
-	case REG_OBP0:
-		softwareRenderer->lookup[0x20 + 0] = value & 3;
-		softwareRenderer->lookup[0x20 + 1] = (value >> 2) & 3;
-		softwareRenderer->lookup[0x20 + 2] = (value >> 4) & 3;
-		softwareRenderer->lookup[0x20 + 3] = (value >> 6) & 3;
+	case GB_REG_OBP0:
+		softwareRenderer->lookup[PAL_OBJ + 0] = value & 3;
+		softwareRenderer->lookup[PAL_OBJ + 1] = (value >> 2) & 3;
+		softwareRenderer->lookup[PAL_OBJ + 2] = (value >> 4) & 3;
+		softwareRenderer->lookup[PAL_OBJ + 3] = (value >> 6) & 3;
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 0] = PAL_HIGHLIGHT + (value & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 1] = PAL_HIGHLIGHT + ((value >> 2) & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 2] = PAL_HIGHLIGHT + ((value >> 4) & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 3] = PAL_HIGHLIGHT + ((value >> 6) & 3);
 		break;
-	case REG_OBP1:
-		softwareRenderer->lookup[0x24 + 0] = value & 3;
-		softwareRenderer->lookup[0x24 + 1] = (value >> 2) & 3;
-		softwareRenderer->lookup[0x24 + 2] = (value >> 4) & 3;
-		softwareRenderer->lookup[0x24 + 3] = (value >> 6) & 3;
+	case GB_REG_OBP1:
+		softwareRenderer->lookup[PAL_OBJ + 4] = value & 3;
+		softwareRenderer->lookup[PAL_OBJ + 5] = (value >> 2) & 3;
+		softwareRenderer->lookup[PAL_OBJ + 6] = (value >> 4) & 3;
+		softwareRenderer->lookup[PAL_OBJ + 7] = (value >> 6) & 3;
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 4] = PAL_HIGHLIGHT + (value & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 5] = PAL_HIGHLIGHT + ((value >> 2) & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 6] = PAL_HIGHLIGHT + ((value >> 4) & 3);
+		softwareRenderer->lookup[PAL_HIGHLIGHT_OBJ + 7] = PAL_HIGHLIGHT + ((value >> 6) & 3);
 		break;
 	}
 	return value;
@@ -442,8 +474,10 @@ static void GBVideoSoftwareRendererWritePalette(struct GBVideoRenderer* renderer
 	if (softwareRenderer->model & GB_MODEL_SGB) {
 		if (index < 0x10 && index && !(index & 3)) {
 			color = softwareRenderer->palette[0];
-		} else if (index >= 0x40 && !(index & 0xF)) {
+		} else if (index >= PAL_SGB_BORDER && !(index & 0xF)) {
 			color = softwareRenderer->palette[0];
+		} else if (index > PAL_HIGHLIGHT && index < PAL_HIGHLIGHT_OBJ && !(index & 3)) {
+			color = softwareRenderer->palette[PAL_HIGHLIGHT_BG];			
 		}
 	}
 	if (renderer->cache) {
@@ -472,6 +506,9 @@ static void GBVideoSoftwareRendererWritePalette(struct GBVideoRenderer* renderer
 #endif
 	}
 	softwareRenderer->palette[index] = color;
+	if (index < PAL_SGB_BORDER && (index < PAL_OBJ || (index & 3))) {
+		softwareRenderer->palette[index + PAL_HIGHLIGHT] = mColorMix5Bit(0x10 - softwareRenderer->lastHighlightAmount, color, softwareRenderer->lastHighlightAmount, renderer->highlightColor);
+	}
 
 	if (softwareRenderer->model & GB_MODEL_SGB && !index && GBRegisterLCDCIsEnable(softwareRenderer->lcdc)) {
 		renderer->writePalette(renderer, 0x04, value);
@@ -499,16 +536,44 @@ static void GBVideoSoftwareRendererWriteOAM(struct GBVideoRenderer* renderer, ui
 	// Nothing to do
 }
 
-static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, int startX, int endX, int y, struct GBObj* obj, size_t oamMax) {
+static void _cleanOAM(struct GBVideoSoftwareRenderer* renderer, int y) {
+	// TODO: GBC differences
+	// TODO: Optimize
+	int spriteHeight = 8;
+	if (GBRegisterLCDCIsObjSize(renderer->lcdc)) {
+		spriteHeight = 16;
+	}
+	int o = 0;
+	int i;
+	for (i = 0; i < GB_VIDEO_MAX_OBJ && o < GB_VIDEO_MAX_LINE_OBJ; ++i) {
+		uint8_t oy = renderer->d.oam->obj[i].y;
+		if (y < oy - 16 || y >= oy - 16 + spriteHeight) {
+			continue;
+		}
+		// TODO: Sort
+		renderer->obj[o].obj = renderer->d.oam->obj[i];
+		renderer->obj[o].index = i;
+		++o;
+		if (o == 10) {
+			break;
+		}
+	}
+	renderer->objMax = o;
+}
+
+static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, int startX, int endX, int y) {
 	struct GBVideoSoftwareRenderer* softwareRenderer = (struct GBVideoSoftwareRenderer*) renderer;
 	softwareRenderer->lastY = y;
 	softwareRenderer->lastX = endX;
+	if (startX >= endX) {
+		return;
+	}
 	uint8_t* maps = &softwareRenderer->d.vram[GB_BASE_MAP];
 	if (GBRegisterLCDCIsTileMap(softwareRenderer->lcdc)) {
 		maps += GB_SIZE_MAP;
 	}
 	if (softwareRenderer->d.disableBG) {
-		memset(&softwareRenderer->row[startX], 0, endX - startX);
+		memset(&softwareRenderer->row[startX], 0, (endX - startX) * sizeof(softwareRenderer->row[0]));
 	}
 	if (GBRegisterLCDCIsBgEnable(softwareRenderer->lcdc) || softwareRenderer->model >= GB_MODEL_CGB) {
 		int wy = softwareRenderer->wy + softwareRenderer->currentWy;
@@ -518,25 +583,40 @@ static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, i
 		}
 		if (GBRegisterLCDCIsWindow(softwareRenderer->lcdc) && softwareRenderer->hasWindow && wx <= endX && !softwareRenderer->d.disableWIN) {
 			if (wx > 0 && !softwareRenderer->d.disableBG) {
-				GBVideoSoftwareRendererDrawBackground(softwareRenderer, maps, startX, wx, softwareRenderer->scx - softwareRenderer->offsetScx, softwareRenderer->scy + y - softwareRenderer->offsetScy);
+				GBVideoSoftwareRendererDrawBackground(softwareRenderer, maps, startX, wx, softwareRenderer->scx - softwareRenderer->offsetScx, softwareRenderer->scy + y - softwareRenderer->offsetScy, renderer->highlightBG);
 			}
 
 			maps = &softwareRenderer->d.vram[GB_BASE_MAP];
 			if (GBRegisterLCDCIsWindowTileMap(softwareRenderer->lcdc)) {
 				maps += GB_SIZE_MAP;
 			}
-			GBVideoSoftwareRendererDrawBackground(softwareRenderer, maps, wx, endX, -wx - softwareRenderer->offsetWx, y - wy - softwareRenderer->offsetWy);
+			GBVideoSoftwareRendererDrawBackground(softwareRenderer, maps, wx, endX, -wx - softwareRenderer->offsetWx, y - wy - softwareRenderer->offsetWy, renderer->highlightWIN);
 		} else if (!softwareRenderer->d.disableBG) {
-			GBVideoSoftwareRendererDrawBackground(softwareRenderer, maps, startX, endX, softwareRenderer->scx - softwareRenderer->offsetScx, softwareRenderer->scy + y - softwareRenderer->offsetScy);
+			GBVideoSoftwareRendererDrawBackground(softwareRenderer, maps, startX, endX, softwareRenderer->scx - softwareRenderer->offsetScx, softwareRenderer->scy + y - softwareRenderer->offsetScy, renderer->highlightBG);
 		}
 	} else if (!softwareRenderer->d.disableBG) {
-		memset(&softwareRenderer->row[startX], 0, endX - startX);
+		memset(&softwareRenderer->row[startX], 0, (endX - startX) * sizeof(softwareRenderer->row[0]));
 	}
 
+	if (startX == 0) {
+		_cleanOAM(softwareRenderer, y);
+	}
 	if (GBRegisterLCDCIsObjEnable(softwareRenderer->lcdc) && !softwareRenderer->d.disableOBJ) {
-		size_t i;
-		for (i = 0; i < oamMax; ++i) {
-			GBVideoSoftwareRendererDrawObj(softwareRenderer, &obj[i], startX, endX, y);
+		int i;
+		for (i = 0; i < softwareRenderer->objMax; ++i) {
+			GBVideoSoftwareRendererDrawObj(softwareRenderer, &softwareRenderer->obj[i], startX, endX, y);
+		}
+	}
+
+	unsigned highlightAmount = (renderer->highlightAmount + 6) >> 4;
+	if (softwareRenderer->lastHighlightAmount != highlightAmount) {
+		softwareRenderer->lastHighlightAmount = highlightAmount;
+		int i;
+		for (i = 0; i < PAL_SGB_BORDER; ++i) {
+			if (i >= PAL_OBJ && (i & 3) == 0) {
+				continue;
+			}
+			softwareRenderer->palette[i + PAL_HIGHLIGHT] = mColorMix5Bit(0x10 - highlightAmount, softwareRenderer->palette[i], highlightAmount, renderer->highlightColor);
 		}
 	}
 
@@ -556,7 +636,7 @@ static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, i
 			p <<= 2;
 		}
 		for (; x < ((startX + 7) & ~7) && x < endX; ++x) {
-			row[x] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x] & 0x7F]];
+			row[x] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x] & OBJ_PRIO_MASK]];
 		}
 		for (; x + 7 < (endX & ~7); x += 8) {
 			if (softwareRenderer->model & GB_MODEL_SGB) {
@@ -565,14 +645,14 @@ static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, i
 				p &= 3;
 				p <<= 2;
 			}
-			row[x + 0] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x] & 0x7F]];
-			row[x + 1] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 1] & 0x7F]];
-			row[x + 2] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 2] & 0x7F]];
-			row[x + 3] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 3] & 0x7F]];
-			row[x + 4] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 4] & 0x7F]];
-			row[x + 5] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 5] & 0x7F]];
-			row[x + 6] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 6] & 0x7F]];
-			row[x + 7] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 7] & 0x7F]];
+			row[x + 0] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x] & OBJ_PRIO_MASK]];
+			row[x + 1] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 1] & OBJ_PRIO_MASK]];
+			row[x + 2] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 2] & OBJ_PRIO_MASK]];
+			row[x + 3] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 3] & OBJ_PRIO_MASK]];
+			row[x + 4] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 4] & OBJ_PRIO_MASK]];
+			row[x + 5] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 5] & OBJ_PRIO_MASK]];
+			row[x + 6] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 6] & OBJ_PRIO_MASK]];
+			row[x + 7] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x + 7] & OBJ_PRIO_MASK]];
 		}
 		if (softwareRenderer->model & GB_MODEL_SGB) {
 			p = softwareRenderer->d.sgbAttributes[(x >> 5) + 5 * (y >> 3)];
@@ -581,7 +661,7 @@ static void GBVideoSoftwareRendererDrawRange(struct GBVideoRenderer* renderer, i
 			p <<= 2;
 		}
 		for (; x < endX; ++x) {
-			row[x] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x] & 0x7F]];
+			row[x] = softwareRenderer->palette[p | softwareRenderer->lookup[softwareRenderer->row[x] & OBJ_PRIO_MASK]];
 		}
 		break;
 	case 1:
@@ -700,6 +780,9 @@ static void GBVideoSoftwareRendererFinishFrame(struct GBVideoRenderer* renderer)
 		case SGB_ATTR_SET:
 			if (softwareRenderer->sgbPacket[1] & 0x40) {
 				renderer->sgbRenderMode = 0;
+				if (softwareRenderer->sgbBorders) {
+					_regenerateSGBBorder(softwareRenderer);
+				}
 			}
 			break;
 		case SGB_PAL_TRN:
@@ -742,7 +825,7 @@ static void GBVideoSoftwareRendererEnableSGBBorder(struct GBVideoRenderer* rende
 	}
 }
 
-static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer* renderer, uint8_t* maps, int startX, int endX, int sx, int sy) {
+static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer* renderer, uint8_t* maps, int startX, int endX, int sx, int sy, bool highlight) {
 	uint8_t* data = renderer->d.vram;
 	uint8_t* attr = &maps[GB_SIZE_VRAM_BANK0];
 	if (!GBRegisterLCDCIsTileData(renderer->lcdc)) {
@@ -767,12 +850,12 @@ static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer
 			} else {
 				bgTile = ((int8_t*) maps)[topX + topY];
 			}
-			int p = 0;
+			int p = highlight ? PAL_HIGHLIGHT_BG : PAL_BG;
 			if (renderer->model >= GB_MODEL_CGB) {
 				GBObjAttributes attrs = attr[topX + topY];
-				p = GBObjAttributesGetCGBPalette(attrs) * 4;
+				p |= GBObjAttributesGetCGBPalette(attrs) * 4;
 				if (GBObjAttributesIsPriority(attrs) && GBRegisterLCDCIsBgEnable(renderer->lcdc)) {
-					p |= 0x80;
+					p |= OBJ_PRIORITY;
 				}
 				if (GBObjAttributesIsBank(attrs)) {
 					localData += GB_SIZE_VRAM_BANK0;
@@ -802,12 +885,12 @@ static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer
 		} else {
 			bgTile = ((int8_t*) maps)[topX + topY];
 		}
-		int p = 0;
+		int p = highlight ? PAL_HIGHLIGHT_BG : PAL_BG;
 		if (renderer->model >= GB_MODEL_CGB) {
 			GBObjAttributes attrs = attr[topX + topY];
-			p = GBObjAttributesGetCGBPalette(attrs) * 4;
+			p |= GBObjAttributesGetCGBPalette(attrs) * 4;
 			if (GBObjAttributesIsPriority(attrs) && GBRegisterLCDCIsBgEnable(renderer->lcdc)) {
-				p |= 0x80;
+				p |= OBJ_PRIORITY;
 			}
 			if (GBObjAttributesIsBank(attrs)) {
 				localData += GB_SIZE_VRAM_BANK0;
@@ -842,8 +925,8 @@ static void GBVideoSoftwareRendererDrawBackground(struct GBVideoSoftwareRenderer
 	}
 }
 
-static void GBVideoSoftwareRendererDrawObj(struct GBVideoSoftwareRenderer* renderer, struct GBObj* obj, int startX, int endX, int y) {
-	int objX = obj->x + renderer->objOffsetX;
+static void GBVideoSoftwareRendererDrawObj(struct GBVideoSoftwareRenderer* renderer, struct GBVideoRendererSprite* obj, int startX, int endX, int y) {
+	int objX = obj->obj.x + renderer->objOffsetX;
 	int ix = objX - 8;
 	if (endX < ix || startX >= ix + 8) {
 		return;
@@ -860,8 +943,8 @@ static void GBVideoSoftwareRendererDrawObj(struct GBVideoSoftwareRenderer* rende
 	uint8_t* data = renderer->d.vram;
 	int tileOffset = 0;
 	int bottomY;
-	int objY = obj->y + renderer->objOffsetY;
-	if (GBObjAttributesIsYFlip(obj->attr)) {
+	int objY = obj->obj.y + renderer->objOffsetY;
+	if (GBObjAttributesIsYFlip(obj->obj.attr)) {
 		bottomY = 7 - ((y - objY - 16) & 7);
 		if (GBRegisterLCDCIsObjSize(renderer->lcdc) && y - objY < -8) {
 			++tileOffset;
@@ -872,115 +955,113 @@ static void GBVideoSoftwareRendererDrawObj(struct GBVideoSoftwareRenderer* rende
 			++tileOffset;
 		}
 	}
-	if (GBRegisterLCDCIsObjSize(renderer->lcdc) && obj->tile & 1) {
+	if (GBRegisterLCDCIsObjSize(renderer->lcdc) && obj->obj.tile & 1) {
 		--tileOffset;
 	}
-	uint8_t mask = GBObjAttributesIsPriority(obj->attr) ? 0x63 : 0x60;
-	uint8_t mask2 = GBObjAttributesIsPriority(obj->attr) ? 0 : 0x83;
-	int p;
+	unsigned mask = GBObjAttributesIsPriority(obj->obj.attr) ? 0x63 : 0x60;
+	unsigned mask2 = GBObjAttributesIsPriority(obj->obj.attr) ? 0 : (OBJ_PRIORITY | 3);
+	int p = renderer->d.highlightOBJ[obj->index] ? PAL_HIGHLIGHT_OBJ : PAL_OBJ;
 	if (renderer->model >= GB_MODEL_CGB) {
-		p = (GBObjAttributesGetCGBPalette(obj->attr) + 8) * 4;
-		if (GBObjAttributesIsBank(obj->attr)) {
+		p |= GBObjAttributesGetCGBPalette(obj->obj.attr) * 4;
+		if (GBObjAttributesIsBank(obj->obj.attr)) {
 			data += GB_SIZE_VRAM_BANK0;
 		}
 		if (!GBRegisterLCDCIsBgEnable(renderer->lcdc)) {
 			mask = 0x60;
-			mask2 = 0x83;
+			mask2 = OBJ_PRIORITY | 3;
 		}
 	} else {
-		p = (GBObjAttributesGetPalette(obj->attr) + 8) * 4;
+		p |= (GBObjAttributesGetPalette(obj->obj.attr) + 8) * 4;
 	}
 	int bottomX;
 	int x = startX;
+	int objTile = obj->obj.tile + tileOffset;
 	if ((x - objX) & 7) {
 		for (; x < endX; ++x) {
-			if (GBObjAttributesIsXFlip(obj->attr)) {
+			if (GBObjAttributesIsXFlip(obj->obj.attr)) {
 				bottomX = (x - objX) & 7;
 			} else {
 				bottomX = 7 - ((x - objX) & 7);
 			}
-			int objTile = obj->tile + tileOffset;
 			uint8_t tileDataLower = data[(objTile * 8 + bottomY) * 2];
 			uint8_t tileDataUpper = data[(objTile * 8 + bottomY) * 2 + 1];
 			tileDataUpper >>= bottomX;
 			tileDataLower >>= bottomX;
-			color_t current = renderer->row[x];
-			if (((tileDataUpper | tileDataLower) & 1) && !(current & mask) && (current & mask2) <= 0x80) {
+			unsigned current = renderer->row[x];
+			if (((tileDataUpper | tileDataLower) & 1) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 				renderer->row[x] = p | ((tileDataUpper & 1) << 1) | (tileDataLower & 1);
 			}
 		}
-	} else if (GBObjAttributesIsXFlip(obj->attr)) {
-		int objTile = obj->tile + tileOffset;
+	} else if (GBObjAttributesIsXFlip(obj->obj.attr)) {
 		uint8_t tileDataLower = data[(objTile * 8 + bottomY) * 2];
 		uint8_t tileDataUpper = data[(objTile * 8 + bottomY) * 2 + 1];
-		color_t current;
+		unsigned current;
 		current = renderer->row[x];
-		if (((tileDataUpper | tileDataLower) & 1) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 1) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x] = p | ((tileDataUpper & 1) << 1) | (tileDataLower & 1);
 		}
 		current = renderer->row[x + 1];
-		if (((tileDataUpper | tileDataLower) & 2) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 2) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 1] = p | (tileDataUpper & 2) | ((tileDataLower & 2) >> 1);
 		}
 		current = renderer->row[x + 2];
-		if (((tileDataUpper | tileDataLower) & 4) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 4) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 2] = p | ((tileDataUpper & 4) >> 1) | ((tileDataLower & 4) >> 2);
 		}
 		current = renderer->row[x + 3];
-		if (((tileDataUpper | tileDataLower) & 8) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 8) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 3] = p | ((tileDataUpper & 8) >> 2) | ((tileDataLower & 8) >> 3);
 		}
 		current = renderer->row[x + 4];
-		if (((tileDataUpper | tileDataLower) & 16) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 16) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 4] = p | ((tileDataUpper & 16) >> 3) | ((tileDataLower & 16) >> 4);
 		}
 		current = renderer->row[x + 5];
-		if (((tileDataUpper | tileDataLower) & 32) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 32) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 5] = p | ((tileDataUpper & 32) >> 4) | ((tileDataLower & 32) >> 5);
 		}
 		current = renderer->row[x + 6];
-		if (((tileDataUpper | tileDataLower) & 64) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 64) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 6] = p | ((tileDataUpper & 64) >> 5) | ((tileDataLower & 64) >> 6);
 		}
 		current = renderer->row[x + 7];
-		if (((tileDataUpper | tileDataLower) & 128) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 128) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 7] = p | ((tileDataUpper & 128) >> 6) | ((tileDataLower & 128) >> 7);
 		}
 	} else {
-		int objTile = obj->tile + tileOffset;
 		uint8_t tileDataLower = data[(objTile * 8 + bottomY) * 2];
 		uint8_t tileDataUpper = data[(objTile * 8 + bottomY) * 2 + 1];
-		color_t current;
+		unsigned current;
 		current = renderer->row[x + 7];
-		if (((tileDataUpper | tileDataLower) & 1) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 1) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 7] = p | ((tileDataUpper & 1) << 1) | (tileDataLower & 1);
 		}
 		current = renderer->row[x + 6];
-		if (((tileDataUpper | tileDataLower) & 2) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 2) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 6] = p | (tileDataUpper & 2) | ((tileDataLower & 2) >> 1);
 		}
 		current = renderer->row[x + 5];
-		if (((tileDataUpper | tileDataLower) & 4) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 4) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 5] = p | ((tileDataUpper & 4) >> 1) | ((tileDataLower & 4) >> 2);
 		}
 		current = renderer->row[x + 4];
-		if (((tileDataUpper | tileDataLower) & 8) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 8) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 4] = p | ((tileDataUpper & 8) >> 2) | ((tileDataLower & 8) >> 3);
 		}
 		current = renderer->row[x + 3];
-		if (((tileDataUpper | tileDataLower) & 16) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 16) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 3] = p | ((tileDataUpper & 16) >> 3) | ((tileDataLower & 16) >> 4);
 		}
 		current = renderer->row[x + 2];
-		if (((tileDataUpper | tileDataLower) & 32) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 32) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 2] = p | ((tileDataUpper & 32) >> 4) | ((tileDataLower & 32) >> 5);
 		}
 		current = renderer->row[x + 1];
-		if (((tileDataUpper | tileDataLower) & 64) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 64) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x + 1] = p | ((tileDataUpper & 64) >> 5) | ((tileDataLower & 64) >> 6);
 		}
 		current = renderer->row[x];
-		if (((tileDataUpper | tileDataLower) & 128) && !(current & mask) && (current & mask2) <= 0x80) {
+		if (((tileDataUpper | tileDataLower) & 128) && !(current & mask) && (current & mask2) <= OBJ_PRIORITY) {
 			renderer->row[x] = p | ((tileDataUpper & 128) >> 6) | ((tileDataLower & 128) >> 7);
 		}
 	}
