@@ -14,28 +14,29 @@ void _GBTimerIRQ(struct mTiming* timing, void* context, uint32_t cyclesLate) {
 	UNUSED(timing);
 	UNUSED(cyclesLate);
 	struct GBTimer* timer = context;
-	timer->p->memory.io[REG_TIMA] = timer->p->memory.io[REG_TMA];
-	timer->p->memory.io[REG_IF] |= (1 << GB_IRQ_TIMER);
+	timer->p->memory.io[GB_REG_TIMA] = timer->p->memory.io[GB_REG_TMA];
+	timer->p->memory.io[GB_REG_IF] |= (1 << GB_IRQ_TIMER);
 	GBUpdateIRQs(timer->p);
 }
 
 static void _GBTimerDivIncrement(struct GBTimer* timer, uint32_t cyclesLate) {
-	while (timer->nextDiv >= GB_DMG_DIV_PERIOD) {
-		timer->nextDiv -= GB_DMG_DIV_PERIOD;
+	int tMultiplier = 2 - timer->p->doubleSpeed;
+	while (timer->nextDiv >= GB_DMG_DIV_PERIOD * tMultiplier) {
+		timer->nextDiv -= GB_DMG_DIV_PERIOD * tMultiplier;
 
 		// Make sure to trigger when the correct bit is a falling edge
 		if (timer->timaPeriod > 0 && (timer->internalDiv & (timer->timaPeriod - 1)) == timer->timaPeriod - 1) {
-			++timer->p->memory.io[REG_TIMA];
-			if (!timer->p->memory.io[REG_TIMA]) {
-				mTimingSchedule(&timer->p->timing, &timer->irq, 7 - ((timer->p->cpu->executionState - cyclesLate) & 3));
+			++timer->p->memory.io[GB_REG_TIMA];
+			if (!timer->p->memory.io[GB_REG_TIMA]) {
+				mTimingSchedule(&timer->p->timing, &timer->irq, 7 * tMultiplier - ((timer->p->cpu->executionState * tMultiplier - cyclesLate) & (3 * tMultiplier)));
 			}
 		}
-		unsigned timingFactor = 0x3FF >> !timer->p->doubleSpeed;
+		unsigned timingFactor = (0x200 << timer->p->doubleSpeed) - 1;
 		if ((timer->internalDiv & timingFactor) == timingFactor) {
-			GBAudioUpdateFrame(&timer->p->audio, &timer->p->timing);
+			GBAudioUpdateFrame(&timer->p->audio);
 		}
 		++timer->internalDiv;
-		timer->p->memory.io[REG_DIV] = timer->internalDiv >> 4;
+		timer->p->memory.io[GB_REG_DIV] = timer->internalDiv >> 4;
 	}
 }
 
@@ -52,7 +53,7 @@ void _GBTimerUpdate(struct mTiming* timing, void* context, uint32_t cyclesLate)
 	if (timaToGo < divsToGo) {
 		divsToGo = timaToGo;
 	}
-	timer->nextDiv = GB_DMG_DIV_PERIOD * divsToGo;
+	timer->nextDiv = GB_DMG_DIV_PERIOD * divsToGo * (2 - timer->p->doubleSpeed);
 	mTimingSchedule(timing, &timer->event, timer->nextDiv - cyclesLate);
 }
 
@@ -66,7 +67,7 @@ void GBTimerReset(struct GBTimer* timer) {
 	timer->irq.callback = _GBTimerIRQ;
 	timer->event.priority = 0x21;
 
-	timer->nextDiv = GB_DMG_DIV_PERIOD; // TODO: GBC differences
+	timer->nextDiv = GB_DMG_DIV_PERIOD * 2;
 	timer->timaPeriod = 1024 >> 4;
 }
 
@@ -74,27 +75,27 @@ void GBTimerDivReset(struct GBTimer* timer) {
 	timer->nextDiv -= mTimingUntil(&timer->p->timing, &timer->event);
 	mTimingDeschedule(&timer->p->timing, &timer->event);
 	_GBTimerDivIncrement(timer, 0);
-	if (((timer->internalDiv << 1) | ((timer->nextDiv >> 3) & 1)) & timer->timaPeriod) {
-		++timer->p->memory.io[REG_TIMA];
-		if (!timer->p->memory.io[REG_TIMA]) {
-			mTimingSchedule(&timer->p->timing, &timer->irq, 7 - (timer->p->cpu->executionState & 3));
+	int tMultiplier = 2 - timer->p->doubleSpeed;
+	if (((timer->internalDiv << 1) | ((timer->nextDiv >> (4 - timer->p->doubleSpeed)) & 1)) & timer->timaPeriod) {
+		++timer->p->memory.io[GB_REG_TIMA];
+		if (!timer->p->memory.io[GB_REG_TIMA]) {
+			mTimingSchedule(&timer->p->timing, &timer->irq, (7 - (timer->p->cpu->executionState & 3)) * tMultiplier);
 		}
 	}
-	unsigned timingFactor = 0x400 >> !timer->p->doubleSpeed;
-	if (timer->internalDiv & timingFactor) {
-		GBAudioUpdateFrame(&timer->p->audio, &timer->p->timing);
+	if (timer->internalDiv & (0x200 << timer->p->doubleSpeed)) {
+		GBAudioUpdateFrame(&timer->p->audio);
 	}
-	timer->p->memory.io[REG_DIV] = 0;
+	timer->p->memory.io[GB_REG_DIV] = 0;
 	timer->internalDiv = 0;
-	timer->nextDiv = GB_DMG_DIV_PERIOD;
-	mTimingSchedule(&timer->p->timing, &timer->event, timer->nextDiv - ((timer->p->cpu->executionState + 1) & 3));
+	timer->nextDiv = GB_DMG_DIV_PERIOD * (2 - timer->p->doubleSpeed);
+	mTimingSchedule(&timer->p->timing, &timer->event, timer->nextDiv - ((timer->p->cpu->executionState + 1) & 3) * tMultiplier);
 }
 
 uint8_t GBTimerUpdateTAC(struct GBTimer* timer, GBRegisterTAC tac) {
 	if (GBRegisterTACIsRun(tac)) {
 		timer->nextDiv -= mTimingUntil(&timer->p->timing, &timer->event);
 		mTimingDeschedule(&timer->p->timing, &timer->event);
-		_GBTimerDivIncrement(timer, (timer->p->cpu->executionState + 2) & 3);
+		_GBTimerDivIncrement(timer, ((timer->p->cpu->executionState + 2) & 3) * (2 - timer->p->doubleSpeed));
 
 		switch (GBRegisterTACGetClock(tac)) {
 		case 0:
@@ -111,7 +112,7 @@ uint8_t GBTimerUpdateTAC(struct GBTimer* timer, GBRegisterTAC tac) {
 			break;
 		}
 
-		timer->nextDiv += GB_DMG_DIV_PERIOD;
+		timer->nextDiv += GB_DMG_DIV_PERIOD * (2 - timer->p->doubleSpeed);
 		mTimingSchedule(&timer->p->timing, &timer->event, timer->nextDiv);
 	} else {
 		timer->timaPeriod = 0;
@@ -140,8 +141,10 @@ void GBTimerDeserialize(struct GBTimer* timer, const struct GBSerializedState* s
 
 	GBSerializedTimerFlags flags = state->timer.flags;
 
+	LOAD_32LE(when, 0, &state->timer.nextIRQ);
 	if (GBSerializedTimerFlagsIsIrqPending(flags)) {
-		LOAD_32LE(when, 0, &state->timer.nextIRQ);
 		mTimingSchedule(&timer->p->timing, &timer->irq, when);
+	} else {
+		timer->irq.when = when + mTimingCurrentTime(&timer->p->timing);
 	}
 }
