@@ -230,24 +230,20 @@ static bool _addPAR3Special(struct GBACheatSet* cheats, uint32_t op2) {
 		break;
 	}
 	if (romPatch >= 0) {
-		while (cheats->romPatches[romPatch].exists) {
-			++romPatch;
-			if (romPatch >= MAX_ROM_PATCHES) {
-				break;
-			}
-		}
-		cheats->romPatches[romPatch].address = BASE_CART0 | ((op2 & 0xFFFFFF) << 1);
-		cheats->romPatches[romPatch].applied = false;
-		cheats->romPatches[romPatch].exists = true;
-		cheats->incompletePatch = &cheats->romPatches[romPatch];
+		struct mCheatPatch* patch = mCheatPatchListAppend(&cheats->d.romPatches);
+		patch->address = BASE_CART0 | ((op2 & 0xFFFFFF) << 1);
+		patch->applied = false;
+		patch->check = false;
+		patch->width = 2;
+		cheats->incompletePatch = patch;
 	}
 	return true;
 }
 
 bool GBACheatAddProActionReplayRaw(struct GBACheatSet* cheats, uint32_t op1, uint32_t op2) {
 	if (cheats->incompletePatch) {
-		cheats->incompletePatch->newValue = op1;
-		cheats->incompletePatch = 0;
+		cheats->incompletePatch->value = op1;
+		cheats->incompletePatch = NULL;
 		return true;
 	}
 	if (cheats->incompleteCheat != COMPLETE) {
