@@ -13,7 +13,6 @@
 #include <mgba/internal/gba/cheats.h>
 #include <mgba/internal/gba/io.h>
 #include <mgba/internal/gba/overrides.h>
-#include <mgba/internal/gba/rr/rr.h>
 
 #include <mgba-util/patch.h>
 #include <mgba-util/crc32.h>
@@ -52,7 +51,7 @@ static void _triggerIRQ(struct mTiming*, void* user, uint32_t cyclesLate);
 
 #ifdef USE_DEBUGGERS
 static bool _setSoftwareBreakpoint(struct ARMDebugger*, uint32_t address, enum ExecutionMode mode, uint32_t* opcode);
-static bool _clearSoftwareBreakpoint(struct ARMDebugger*, uint32_t address, enum ExecutionMode mode, uint32_t opcode);
+static void _clearSoftwareBreakpoint(struct ARMDebugger*, const struct ARMDebugBreakpoint*);
 #endif
 
 #ifdef FIXED_ROM_BUFFER
@@ -96,7 +95,6 @@ static void GBAInit(void* cpu, struct mCPUComponent* component) {
 	gba->luminanceSource = 0;
 	gba->rtcSource = 0;
 	gba->rumble = 0;
-	gba->rr = 0;
 
 	gba->romVf = 0;
 	gba->biosVf = 0;
@@ -110,6 +108,7 @@ static void GBAInit(void* cpu, struct mCPUComponent* component) {
 	gba->idleOptimization = IDLE_LOOP_REMOVE;
 	gba->idleLoop = IDLE_LOOP_NONE;
 
+	gba->vbaBugCompat = false;
 	gba->hardCrash = true;
 	gba->allowOpposingDirections = true;
 
@@ -132,7 +131,7 @@ void GBAUnloadROM(struct GBA* gba) {
 		if (gba->yankedRomSize) {
 			gba->yankedRomSize = 0;
 		}
-#if !defined(FIXED_ROM_BUFFER) && !defined(__wii__)
+#ifndef FIXED_ROM_BUFFER
 		mappedMemoryFree(gba->memory.rom, SIZE_CART0);
 #endif
 	}
@@ -172,7 +171,6 @@ void GBADestroy(struct GBA* gba) {
 	GBAVideoDeinit(&gba->video);
 	GBAAudioDeinit(&gba->audio);
 	GBASIODeinit(&gba->sio);
-	gba->rr = 0;
 	mTimingDeinit(&gba->timing);
 	mCoreCallbacksListDeinit(&gba->coreCallbacks);
 }
@@ -198,13 +196,13 @@ void GBAReset(struct ARMCore* cpu) {
 	cpu->gprs[ARM_SP] = SP_BASE_SYSTEM;
 
 	struct GBA* gba = (struct GBA*) cpu->master;
-	if (!gba->rr || (!gba->rr->isPlaying(gba->rr) && !gba->rr->isRecording(gba->rr))) {
-		gba->memory.savedata.maskWriteback = false;
-		GBASavedataUnmask(&gba->memory.savedata);
-	}
+	gba->memory.savedata.maskWriteback = false;
+	GBASavedataUnmask(&gba->memory.savedata);
 
 	gba->cpuBlocked = false;
 	gba->earlyExit = false;
+	gba->dmaPC = 0;
+	gba->biosStall = 0;
 	if (gba->yankedRomSize) {
 		gba->memory.romSize = gba->yankedRomSize;
 		gba->memory.romMask = toPow2(gba->memory.romSize) - 1;
@@ -219,6 +217,12 @@ void GBAReset(struct ARMCore* cpu) {
 
 	GBASIOReset(&gba->sio);
 
+	// GB Player SIO control should not be engaged before detection, even if we already know it's GBP
+	gba->memory.hw.devices &= ~HW_GB_PLAYER;
+	if (gba->sio.drivers.normal == &gba->memory.hw.gbpDriver.d) {
+		GBASIOSetDriver(&gba->sio, NULL, SIO_NORMAL_32);
+	}
+
 	bool isELF = false;
 #ifdef USE_ELF
 	struct ELF* elf = ELFOpen(gba->romVf);
@@ -283,13 +287,16 @@ static void GBAProcessEvents(struct ARMCore* cpu) {
 		do {
 			int32_t cycles = cpu->cycles;
 			cpu->cycles = 0;
+#ifdef USE_DEBUGGERS
+			gba->timing.globalCycles += cycles < nextEvent ? nextEvent : cycles;
+#endif
 #ifndef NDEBUG
 			if (cycles < 0) {
 				mLOG(GBA, FATAL, "Negative cycles passed: %i", cycles);
 			}
 #endif
-			nextEvent = mTimingTick(&gba->timing, nextEvent + cycles);
-		} while (gba->cpuBlocked);
+			nextEvent = mTimingTick(&gba->timing, cycles < nextEvent ? nextEvent : cycles);
+		} while (gba->cpuBlocked && !gba->earlyExit);
 
 		cpu->nextEvent = nextEvent;
 		if (cpu->halted) {
@@ -308,11 +315,9 @@ static void GBAProcessEvents(struct ARMCore* cpu) {
 		}
 	}
 	gba->earlyExit = false;
-#ifndef NDEBUG
 	if (gba->cpuBlocked) {
-		mLOG(GBA, FATAL, "CPU is blocked!");
+		cpu->cycles = cpu->nextEvent;
 	}
-#endif
 }
 
 #ifdef USE_DEBUGGERS
@@ -406,6 +411,7 @@ bool GBALoadROM(struct GBA* gba, struct VFile* vf) {
 		gba->memory.romSize = gba->pristineRomSize;
 	}
 	if (!gba->memory.rom) {
+		gba->romVf = NULL;
 		mLOG(GBA, WARN, "Couldn't map ROM");
 		return false;
 	}
@@ -434,7 +440,12 @@ bool GBALoadROM(struct GBA* gba, struct VFile* vf) {
 }
 
 bool GBALoadSave(struct GBA* gba, struct VFile* sav) {
+	enum SavedataType type = gba->memory.savedata.type;
+	GBASavedataDeinit(&gba->memory.savedata);
 	GBASavedataInit(&gba->memory.savedata, sav);
+	if (type != SAVEDATA_AUTODETECT) {
+		GBASavedataForceType(&gba->memory.savedata, type);
+	}
 	return sav;
 }
 
@@ -490,7 +501,11 @@ void GBAApplyPatch(struct GBA* gba, struct Patch* patch) {
 	}
 	if (gba->romVf) {
 #ifndef FIXED_ROM_BUFFER
-		gba->romVf->unmap(gba->romVf, gba->memory.rom, gba->pristineRomSize);
+		if (!gba->isPristine) {
+			mappedMemoryFree(gba->memory.rom, SIZE_CART0);
+		} else {
+			gba->romVf->unmap(gba->romVf, gba->memory.rom, gba->pristineRomSize);
+		}
 #endif
 		gba->romVf->close(gba->romVf);
 		gba->romVf = NULL;
@@ -499,7 +514,7 @@ void GBAApplyPatch(struct GBA* gba, struct Patch* patch) {
 	gba->memory.rom = newRom;
 	gba->memory.hw.gpioBase = &((uint16_t*) gba->memory.rom)[GPIO_REG_DATA >> 1];
 	gba->memory.romSize = patchedSize;
-	gba->memory.romMask = SIZE_CART0 - 1;
+	gba->memory.romMask = toPow2(patchedSize) - 1;
 	gba->romCrc32 = doCrc32(gba->memory.rom, gba->memory.romSize);
 }
 
@@ -527,11 +542,15 @@ void GBAHalt(struct GBA* gba) {
 }
 
 void GBAStop(struct GBA* gba) {
+	int validIrqs = (1 << IRQ_GAMEPAK) | (1 << IRQ_KEYPAD) | (1 << IRQ_SIO);
+	int sleep = gba->memory.io[REG_IE >> 1] & validIrqs;
 	size_t c;
 	for (c = 0; c < mCoreCallbacksListSize(&gba->coreCallbacks); ++c) {
 		struct mCoreCallbacks* callbacks = mCoreCallbacksListGetPointer(&gba->coreCallbacks, c);
-		if (callbacks->sleep) {
+		if (sleep && callbacks->sleep) {
 			callbacks->sleep(callbacks->context);
+		} else if (callbacks->shutdown) {
+			callbacks->shutdown(callbacks->context);
 		}
 	}
 	gba->cpu->nextEvent = gba->cpu->cycles;
@@ -741,14 +760,14 @@ void GBAHitStub(struct ARMCore* cpu, uint32_t opcode) {
 
 void GBAIllegal(struct ARMCore* cpu, uint32_t opcode) {
 	struct GBA* gba = (struct GBA*) cpu->master;
+	if (cpu->executionMode == MODE_THUMB && (opcode & 0xFFC0) == 0xE800) {
+		mLOG(GBA, INFO, "Hit Wii U VC opcode: %08x", opcode);
+		return;
+	}
 	if (!gba->yankedRomSize) {
 		// TODO: More sensible category?
 		mLOG(GBA, WARN, "Illegal opcode: %08x", opcode);
 	}
-	if (cpu->executionMode == MODE_THUMB && (opcode & 0xFFC0) == 0xE800) {
-		mLOG(GBA, DEBUG, "Hit Wii U VC opcode: %08x", opcode);
-		return;
-	}
 #ifdef USE_DEBUGGERS
 	if (gba->debugger) {
 		struct mDebuggerEntryInfo info = {
@@ -756,11 +775,9 @@ void GBAIllegal(struct ARMCore* cpu, uint32_t opcode) {
 			.type.bp.opcode = opcode
 		};
 		mDebuggerEnter(gba->debugger->d.p, DEBUGGER_ENTER_ILLEGAL_OP, &info);
-	} else
-#endif
-	{
-		ARMRaiseUndefined(cpu);
 	}
+#endif
+	ARMRaiseUndefined(cpu);
 }
 
 void GBABreakpoint(struct ARMCore* cpu, int immediate) {
@@ -820,12 +837,9 @@ void GBAFrameStarted(struct GBA* gba) {
 }
 
 void GBAFrameEnded(struct GBA* gba) {
+	int wasDirty = gba->memory.savedata.dirty;
 	GBASavedataClean(&gba->memory.savedata, gba->video.frameCounter);
 
-	if (gba->rr) {
-		gba->rr->nextFrame(gba->rr);
-	}
-
 	if (gba->cpu->components && gba->cpu->components[CPU_COMPONENT_CHEAT_DEVICE]) {
 		struct mCheatDevice* device = (struct mCheatDevice*) gba->cpu->components[CPU_COMPONENT_CHEAT_DEVICE];
 		size_t i;
@@ -854,6 +868,9 @@ void GBAFrameEnded(struct GBA* gba) {
 		if (callbacks->videoFrameEnded) {
 			callbacks->videoFrameEnded(callbacks->context);
 		}
+		if (callbacks->savedataUpdated && wasDirty && !gba->memory.savedata.dirty) {
+			callbacks->savedataUpdated(callbacks->context);
+		}
 	}
 }
 
@@ -934,8 +951,7 @@ static bool _setSoftwareBreakpoint(struct ARMDebugger* debugger, uint32_t addres
 	return true;
 }
 
-static bool _clearSoftwareBreakpoint(struct ARMDebugger* debugger, uint32_t address, enum ExecutionMode mode, uint32_t opcode) {
-	GBAClearBreakpoint((struct GBA*) debugger->cpu->master, address, mode, opcode);
-	return true;
+static void _clearSoftwareBreakpoint(struct ARMDebugger* debugger, const struct ARMDebugBreakpoint* breakpoint) {
+	GBAClearBreakpoint((struct GBA*) debugger->cpu->master, breakpoint->d.address, breakpoint->sw.mode, breakpoint->sw.opcode);
 }
-#endif
\ No newline at end of file
+#endif
