@@ -10,6 +10,7 @@
 #include <mgba/internal/gba/serialize.h>
 #include <mgba-util/formatting.h>
 #include <mgba-util/hash.h>
+#include <mgba-util/memory.h>
 
 mLOG_DEFINE_CATEGORY(GBA_HW, "GBA Pak Hardware", "gba.hardware");
 
@@ -49,6 +50,8 @@ static const int RTC_BYTES[8] = {
 
 void GBAHardwareInit(struct GBACartridgeHardware* hw, uint16_t* base) {
 	hw->gpioBase = base;
+	hw->eReaderDots = NULL;
+	memset(hw->eReaderCards, 0, sizeof(hw->eReaderCards));
 	GBAHardwareClear(hw);
 
 	hw->gbpCallback.d.readKeys = _gbpRead;
@@ -71,6 +74,20 @@ void GBAHardwareClear(struct GBACartridgeHardware* hw) {
 	hw->pinState = 0;
 	hw->direction = 0;
 
+	if (hw->eReaderDots) {
+		mappedMemoryFree(hw->eReaderDots, EREADER_DOTCODE_SIZE);
+		hw->eReaderDots = NULL;
+	}
+	int i;
+	for (i = 0; i < EREADER_CARDS_MAX; ++i) {
+		if (!hw->eReaderCards[i].data) {
+			continue;
+		}
+		free(hw->eReaderCards[i].data);
+		hw->eReaderCards[i].data = NULL;
+		hw->eReaderCards[i].size = 0;
+	}
+
 	if (hw->p->sio.drivers.normal == &hw->gbpDriver.d) {
 		GBASIOSetDriver(&hw->p->sio, 0, SIO_NORMAL_32);
 	}
@@ -82,8 +99,12 @@ void GBAHardwareGPIOWrite(struct GBACartridgeHardware* hw, uint32_t address, uin
 	}
 	switch (address) {
 	case GPIO_REG_DATA:
-		hw->pinState &= ~hw->direction;
-		hw->pinState |= value & hw->direction;
+		if (!hw->p->vbaBugCompat) {
+			hw->pinState &= ~hw->direction;
+			hw->pinState |= value & hw->direction;
+		} else {
+			hw->pinState = value;
+		}
 		_readPins(hw);
 		break;
 	case GPIO_REG_DIRECTION:
@@ -221,6 +242,7 @@ void _rtcProcessByte(struct GBACartridgeHardware* hw) {
 
 			hw->rtc.bytesRemaining = RTC_BYTES[RTCCommandDataGetCommand(command)];
 			hw->rtc.commandActive = hw->rtc.bytesRemaining > 0;
+			mLOG(GBA_HW, DEBUG, "Got RTC command %x", RTCCommandDataGetCommand(command));
 			switch (RTCCommandDataGetCommand(command)) {
 			case RTC_RESET:
 				hw->rtc.control = 0;
@@ -278,6 +300,9 @@ unsigned _rtcOutput(struct GBACartridgeHardware* hw) {
 		break;
 	}
 	unsigned output = (outputByte >> hw->rtc.bitsRead) & 1;
+	if (hw->rtc.bitsRead == 0) {
+		mLOG(GBA_HW, DEBUG, "RTC output byte %02X", outputByte);
+	}
 	return output;
 }
 
@@ -525,13 +550,14 @@ void GBAHardwarePlayerUpdate(struct GBA* gba) {
 		gba->memory.hw.gbpTxPosition = 0;
 		return;
 	}
-	if (gba->keyCallback || gba->sio.drivers.normal) {
+	if (gba->keyCallback) {
 		return;
 	}
 	if (GBAHardwarePlayerCheckScreen(&gba->video)) {
 		gba->memory.hw.devices |= HW_GB_PLAYER;
 		gba->memory.hw.gbpInputsPosted = 0;
 		gba->keyCallback = &gba->memory.hw.gbpCallback.d;
+		// TODO: Check if the SIO driver is actually used first
 		GBASIOSetDriver(&gba->sio, &gba->memory.hw.gbpDriver.d, SIO_NORMAL_32);
 	}
 }
@@ -601,14 +627,14 @@ void GBAHardwareSerialize(const struct GBACartridgeHardware* hw, struct GBASeria
 	STORE_16(hw->direction, 0, &state->hw.pinDirection);
 	state->hw.devices = hw->devices;
 
-	STORE_32(hw->rtc.bytesRemaining, 0, &state->hw.rtc.bytesRemaining);
-	STORE_32(hw->rtc.transferStep, 0, &state->hw.rtc.transferStep);
-	STORE_32(hw->rtc.bitsRead, 0, &state->hw.rtc.bitsRead);
-	STORE_32(hw->rtc.bits, 0, &state->hw.rtc.bits);
-	STORE_32(hw->rtc.commandActive, 0, &state->hw.rtc.commandActive);
-	STORE_32(hw->rtc.command, 0, &state->hw.rtc.command);
-	STORE_32(hw->rtc.control, 0, &state->hw.rtc.control);
-	memcpy(state->hw.rtc.time, hw->rtc.time, sizeof(state->hw.rtc.time));
+	STORE_32(hw->rtc.bytesRemaining, 0, &state->hw.rtcBytesRemaining);
+	STORE_32(hw->rtc.transferStep, 0, &state->hw.rtcTransferStep);
+	STORE_32(hw->rtc.bitsRead, 0, &state->hw.rtcBitsRead);
+	STORE_32(hw->rtc.bits, 0, &state->hw.rtcBits);
+	STORE_32(hw->rtc.commandActive, 0, &state->hw.rtcCommandActive);
+	STORE_32(hw->rtc.command, 0, &state->hw.rtcCommand);
+	STORE_32(hw->rtc.control, 0, &state->hw.rtcControl);
+	memcpy(state->hw.time, hw->rtc.time, sizeof(state->hw.time));
 
 	STORE_16(hw->gyroSample, 0, &state->hw.gyroSample);
 	flags1 = GBASerializedHWFlags1SetGyroEdge(flags1, hw->gyroEdge);
@@ -633,14 +659,14 @@ void GBAHardwareDeserialize(struct GBACartridgeHardware* hw, const struct GBASer
 	LOAD_16(hw->direction, 0, &state->hw.pinDirection);
 	hw->devices = state->hw.devices;
 
-	LOAD_32(hw->rtc.bytesRemaining, 0, &state->hw.rtc.bytesRemaining);
-	LOAD_32(hw->rtc.transferStep, 0, &state->hw.rtc.transferStep);
-	LOAD_32(hw->rtc.bitsRead, 0, &state->hw.rtc.bitsRead);
-	LOAD_32(hw->rtc.bits, 0, &state->hw.rtc.bits);
-	LOAD_32(hw->rtc.commandActive, 0, &state->hw.rtc.commandActive);
-	LOAD_32(hw->rtc.command, 0, &state->hw.rtc.command);
-	LOAD_32(hw->rtc.control, 0, &state->hw.rtc.control);
-	memcpy(hw->rtc.time, state->hw.rtc.time, sizeof(hw->rtc.time));
+	LOAD_32(hw->rtc.bytesRemaining, 0, &state->hw.rtcBytesRemaining);
+	LOAD_32(hw->rtc.transferStep, 0, &state->hw.rtcTransferStep);
+	LOAD_32(hw->rtc.bitsRead, 0, &state->hw.rtcBitsRead);
+	LOAD_32(hw->rtc.bits, 0, &state->hw.rtcBits);
+	LOAD_32(hw->rtc.commandActive, 0, &state->hw.rtcCommandActive);
+	LOAD_32(hw->rtc.command, 0, &state->hw.rtcCommand);
+	LOAD_32(hw->rtc.control, 0, &state->hw.rtcControl);
+	memcpy(hw->rtc.time, state->hw.time, sizeof(hw->rtc.time));
 
 	LOAD_16(hw->gyroSample, 0, &state->hw.gyroSample);
 	hw->gyroEdge = GBASerializedHWFlags1GetGyroEdge(flags1);
