@@ -8,7 +8,6 @@
 #include <mgba/internal/arm/macros.h>
 #include <mgba/internal/gba/dma.h>
 #include <mgba/internal/gba/gba.h>
-#include <mgba/internal/gba/rr/rr.h>
 #include <mgba/internal/gba/serialize.h>
 
 mLOG_DEFINE_CATEGORY(GBA_IO, "GBA I/O", "gba.io");
@@ -227,7 +226,7 @@ static const int _isValidRegister[REG_MAX >> 1] = {
 	1, 1, 1, 1, 1, 0, 0, 0,
 	1, 1, 1, 0, 0, 0, 0, 0,
 	1, 0, 0, 0, 0, 0, 0, 0,
-	1, 0, 1, 0, 1, 0, 0, 0,
+	1, 1, 1, 1, 1, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
@@ -269,7 +268,7 @@ static const int _isRSpecialRegister[REG_MAX >> 1] = {
 	1, 1, 1, 1, 1, 0, 0, 0,
 	1, 1, 1, 0, 0, 0, 0, 0,
 	1, 0, 0, 0, 0, 0, 0, 0,
-	1, 0, 1, 0, 1, 0, 0, 0,
+	1, 1, 1, 1, 1, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
@@ -310,7 +309,7 @@ static const int _isWSpecialRegister[REG_MAX >> 1] = {
 	1, 1, 1, 1, 1, 0, 0, 0,
 	1, 1, 1, 0, 0, 0, 0, 0,
 	1, 0, 0, 0, 0, 0, 0, 0,
-	1, 0, 1, 0, 1, 0, 0, 0,
+	1, 1, 1, 1, 1, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
 	0, 0, 0, 0, 0, 0, 0, 0,
@@ -342,247 +341,256 @@ void GBAIOInit(struct GBA* gba) {
 }
 
 void GBAIOWrite(struct GBA* gba, uint32_t address, uint16_t value) {
-	if (address < REG_SOUND1CNT_LO && (address > REG_VCOUNT || address == REG_DISPCNT)) {
-		value = gba->video.renderer->writeVideoRegister(gba->video.renderer, address, value);
-	} else {
-		switch (address) {
-		// Video
-		case REG_DISPSTAT:
-			value &= 0xFFF8;
-			GBAVideoWriteDISPSTAT(&gba->video, value);
+	if (address < REG_SOUND1CNT_LO && (address > REG_VCOUNT || address < REG_DISPSTAT)) {
+		gba->memory.io[address >> 1] = gba->video.renderer->writeVideoRegister(gba->video.renderer, address, value);
+		return;
+	}
+
+	if (address >= REG_SOUND1CNT_LO && address <= REG_SOUNDCNT_LO && !gba->audio.enable) {
+		// Ignore writes to most audio registers if the hardware is off.
+		return;
+	}
+
+	switch (address) {
+	// Video
+	case REG_DISPSTAT:
+		value &= 0xFFF8;
+		GBAVideoWriteDISPSTAT(&gba->video, value);
+		return;
+
+	case REG_VCOUNT:
+		mLOG(GBA_IO, GAME_ERROR, "Write to read-only I/O register: %03X", address);
+		return;
+
+	// Audio
+	case REG_SOUND1CNT_LO:
+		GBAAudioWriteSOUND1CNT_LO(&gba->audio, value);
+		value &= 0x007F;
+		break;
+	case REG_SOUND1CNT_HI:
+		GBAAudioWriteSOUND1CNT_HI(&gba->audio, value);
+		break;
+	case REG_SOUND1CNT_X:
+		GBAAudioWriteSOUND1CNT_X(&gba->audio, value);
+		value &= 0x47FF;
+		break;
+	case REG_SOUND2CNT_LO:
+		GBAAudioWriteSOUND2CNT_LO(&gba->audio, value);
+		break;
+	case REG_SOUND2CNT_HI:
+		GBAAudioWriteSOUND2CNT_HI(&gba->audio, value);
+		value &= 0x47FF;
+		break;
+	case REG_SOUND3CNT_LO:
+		GBAAudioWriteSOUND3CNT_LO(&gba->audio, value);
+		value &= 0x00E0;
+		break;
+	case REG_SOUND3CNT_HI:
+		GBAAudioWriteSOUND3CNT_HI(&gba->audio, value);
+		value &= 0xE03F;
+		break;
+	case REG_SOUND3CNT_X:
+		GBAAudioWriteSOUND3CNT_X(&gba->audio, value);
+		// TODO: The low bits need to not be readable, but still 8-bit writable
+		value &= 0x47FF;
+		break;
+	case REG_SOUND4CNT_LO:
+		GBAAudioWriteSOUND4CNT_LO(&gba->audio, value);
+		value &= 0xFF3F;
+		break;
+	case REG_SOUND4CNT_HI:
+		GBAAudioWriteSOUND4CNT_HI(&gba->audio, value);
+		value &= 0x40FF;
+		break;
+	case REG_SOUNDCNT_LO:
+		GBAAudioWriteSOUNDCNT_LO(&gba->audio, value);
+		value &= 0xFF77;
+		break;
+	case REG_SOUNDCNT_HI:
+		GBAAudioWriteSOUNDCNT_HI(&gba->audio, value);
+		value &= 0x770F;
+		break;
+	case REG_SOUNDCNT_X:
+		GBAAudioWriteSOUNDCNT_X(&gba->audio, value);
+		value &= 0x0080;
+		value |= gba->memory.io[REG_SOUNDCNT_X >> 1] & 0xF;
+		break;
+	case REG_SOUNDBIAS:
+		GBAAudioWriteSOUNDBIAS(&gba->audio, value);
+		break;
+
+	case REG_WAVE_RAM0_LO:
+	case REG_WAVE_RAM1_LO:
+	case REG_WAVE_RAM2_LO:
+	case REG_WAVE_RAM3_LO:
+		GBAIOWrite32(gba, address, (gba->memory.io[(address >> 1) + 1] << 16) | value);
+		break;
+
+	case REG_WAVE_RAM0_HI:
+	case REG_WAVE_RAM1_HI:
+	case REG_WAVE_RAM2_HI:
+	case REG_WAVE_RAM3_HI:
+		GBAIOWrite32(gba, address - 2, gba->memory.io[(address >> 1) - 1] | (value << 16));
+		break;
+
+	case REG_FIFO_A_LO:
+	case REG_FIFO_B_LO:
+		GBAIOWrite32(gba, address, (gba->memory.io[(address >> 1) + 1] << 16) | value);
+		return;
+
+	case REG_FIFO_A_HI:
+	case REG_FIFO_B_HI:
+		GBAIOWrite32(gba, address - 2, gba->memory.io[(address >> 1) - 1] | (value << 16));
+		return;
+
+	// DMA
+	case REG_DMA0SAD_LO:
+	case REG_DMA0DAD_LO:
+	case REG_DMA1SAD_LO:
+	case REG_DMA1DAD_LO:
+	case REG_DMA2SAD_LO:
+	case REG_DMA2DAD_LO:
+	case REG_DMA3SAD_LO:
+	case REG_DMA3DAD_LO:
+		GBAIOWrite32(gba, address, (gba->memory.io[(address >> 1) + 1] << 16) | value);
+		break;
+
+	case REG_DMA0SAD_HI:
+	case REG_DMA0DAD_HI:
+	case REG_DMA1SAD_HI:
+	case REG_DMA1DAD_HI:
+	case REG_DMA2SAD_HI:
+	case REG_DMA2DAD_HI:
+	case REG_DMA3SAD_HI:
+	case REG_DMA3DAD_HI:
+		GBAIOWrite32(gba, address - 2, gba->memory.io[(address >> 1) - 1] | (value << 16));
+		break;
+
+	case REG_DMA0CNT_LO:
+		GBADMAWriteCNT_LO(gba, 0, value & 0x3FFF);
+		break;
+	case REG_DMA0CNT_HI:
+		value = GBADMAWriteCNT_HI(gba, 0, value);
+		break;
+	case REG_DMA1CNT_LO:
+		GBADMAWriteCNT_LO(gba, 1, value & 0x3FFF);
+		break;
+	case REG_DMA1CNT_HI:
+		value = GBADMAWriteCNT_HI(gba, 1, value);
+		break;
+	case REG_DMA2CNT_LO:
+		GBADMAWriteCNT_LO(gba, 2, value & 0x3FFF);
+		break;
+	case REG_DMA2CNT_HI:
+		value = GBADMAWriteCNT_HI(gba, 2, value);
+		break;
+	case REG_DMA3CNT_LO:
+		GBADMAWriteCNT_LO(gba, 3, value);
+		break;
+	case REG_DMA3CNT_HI:
+		value = GBADMAWriteCNT_HI(gba, 3, value);
+		break;
+
+	// Timers
+	case REG_TM0CNT_LO:
+		GBATimerWriteTMCNT_LO(gba, 0, value);
+		return;
+	case REG_TM1CNT_LO:
+		GBATimerWriteTMCNT_LO(gba, 1, value);
+		return;
+	case REG_TM2CNT_LO:
+		GBATimerWriteTMCNT_LO(gba, 2, value);
+		return;
+	case REG_TM3CNT_LO:
+		GBATimerWriteTMCNT_LO(gba, 3, value);
+		return;
+
+	case REG_TM0CNT_HI:
+		value &= 0x00C7;
+		GBATimerWriteTMCNT_HI(gba, 0, value);
+		break;
+	case REG_TM1CNT_HI:
+		value &= 0x00C7;
+		GBATimerWriteTMCNT_HI(gba, 1, value);
+		break;
+	case REG_TM2CNT_HI:
+		value &= 0x00C7;
+		GBATimerWriteTMCNT_HI(gba, 2, value);
+		break;
+	case REG_TM3CNT_HI:
+		value &= 0x00C7;
+		GBATimerWriteTMCNT_HI(gba, 3, value);
+		break;
+
+	// SIO
+	case REG_SIOCNT:
+		GBASIOWriteSIOCNT(&gba->sio, value);
+		break;
+	case REG_RCNT:
+		value &= 0xC1FF;
+		GBASIOWriteRCNT(&gba->sio, value);
+		break;
+	case REG_JOY_TRANS_LO:
+	case REG_JOY_TRANS_HI:
+		gba->memory.io[REG_JOYSTAT >> 1] |= JOYSTAT_TRANS;
+		// Fall through
+	case REG_SIODATA32_LO:
+	case REG_SIODATA32_HI:
+	case REG_SIOMLT_SEND:
+	case REG_JOYCNT:
+	case REG_JOYSTAT:
+	case REG_JOY_RECV_LO:
+	case REG_JOY_RECV_HI:
+		value = GBASIOWriteRegister(&gba->sio, address, value);
+		break;
+
+	// Interrupts and misc
+	case REG_KEYCNT:
+		value &= 0xC3FF;
+		gba->memory.io[address >> 1] = value;
+		GBATestKeypadIRQ(gba);
+		return;
+	case REG_WAITCNT:
+		value &= 0x5FFF;
+		GBAAdjustWaitstates(gba, value);
+		break;
+	case REG_IE:
+		gba->memory.io[REG_IE >> 1] = value;
+		GBATestIRQ(gba, 1);
+		return;
+	case REG_IF:
+		value = gba->memory.io[REG_IF >> 1] & ~value;
+		gba->memory.io[REG_IF >> 1] = value;
+		GBATestIRQ(gba, 1);
+		return;
+	case REG_IME:
+		gba->memory.io[REG_IME >> 1] = value & 1;
+		GBATestIRQ(gba, 1);
+		return;
+	case REG_MAX:
+		// Some bad interrupt libraries will write to this
+		break;
+	case REG_DEBUG_ENABLE:
+		gba->debug = value == 0xC0DE;
+		return;
+	case REG_DEBUG_FLAGS:
+		if (gba->debug) {
+			GBADebug(gba, value);
+
 			return;
-
-		case REG_VCOUNT:
-			mLOG(GBA_IO, GAME_ERROR, "Write to read-only I/O register: %03X", address);
-			return;
-
-		// Audio
-		case REG_SOUND1CNT_LO:
-			GBAAudioWriteSOUND1CNT_LO(&gba->audio, value);
-			value &= 0x007F;
-			break;
-		case REG_SOUND1CNT_HI:
-			GBAAudioWriteSOUND1CNT_HI(&gba->audio, value);
-			break;
-		case REG_SOUND1CNT_X:
-			GBAAudioWriteSOUND1CNT_X(&gba->audio, value);
-			value &= 0x47FF;
-			break;
-		case REG_SOUND2CNT_LO:
-			GBAAudioWriteSOUND2CNT_LO(&gba->audio, value);
-			break;
-		case REG_SOUND2CNT_HI:
-			GBAAudioWriteSOUND2CNT_HI(&gba->audio, value);
-			value &= 0x47FF;
-			break;
-		case REG_SOUND3CNT_LO:
-			GBAAudioWriteSOUND3CNT_LO(&gba->audio, value);
-			value &= 0x00E0;
-			break;
-		case REG_SOUND3CNT_HI:
-			GBAAudioWriteSOUND3CNT_HI(&gba->audio, value);
-			value &= 0xE03F;
-			break;
-		case REG_SOUND3CNT_X:
-			GBAAudioWriteSOUND3CNT_X(&gba->audio, value);
-			// TODO: The low bits need to not be readable, but still 8-bit writable
-			value &= 0x47FF;
-			break;
-		case REG_SOUND4CNT_LO:
-			GBAAudioWriteSOUND4CNT_LO(&gba->audio, value);
-			value &= 0xFF3F;
-			break;
-		case REG_SOUND4CNT_HI:
-			GBAAudioWriteSOUND4CNT_HI(&gba->audio, value);
-			value &= 0x40FF;
-			break;
-		case REG_SOUNDCNT_LO:
-			GBAAudioWriteSOUNDCNT_LO(&gba->audio, value);
-			value &= 0xFF77;
-			break;
-		case REG_SOUNDCNT_HI:
-			GBAAudioWriteSOUNDCNT_HI(&gba->audio, value);
-			value &= 0x770F;
-			break;
-		case REG_SOUNDCNT_X:
-			GBAAudioWriteSOUNDCNT_X(&gba->audio, value);
-			value &= 0x0080;
-			value |= gba->memory.io[REG_SOUNDCNT_X >> 1] & 0xF;
-			break;
-		case REG_SOUNDBIAS:
-			GBAAudioWriteSOUNDBIAS(&gba->audio, value);
-			break;
-
-		case REG_WAVE_RAM0_LO:
-		case REG_WAVE_RAM1_LO:
-		case REG_WAVE_RAM2_LO:
-		case REG_WAVE_RAM3_LO:
-			GBAIOWrite32(gba, address, (gba->memory.io[(address >> 1) + 1] << 16) | value);
-			break;
-
-		case REG_WAVE_RAM0_HI:
-		case REG_WAVE_RAM1_HI:
-		case REG_WAVE_RAM2_HI:
-		case REG_WAVE_RAM3_HI:
-			GBAIOWrite32(gba, address - 2, gba->memory.io[(address >> 1) - 1] | (value << 16));
-			break;
-
-		case REG_FIFO_A_LO:
-		case REG_FIFO_B_LO:
-			GBAIOWrite32(gba, address, (gba->memory.io[(address >> 1) + 1] << 16) | value);
-			break;
-
-		case REG_FIFO_A_HI:
-		case REG_FIFO_B_HI:
-			GBAIOWrite32(gba, address - 2, gba->memory.io[(address >> 1) - 1] | (value << 16));
-			break;
-
-		// DMA
-		case REG_DMA0SAD_LO:
-		case REG_DMA0DAD_LO:
-		case REG_DMA1SAD_LO:
-		case REG_DMA1DAD_LO:
-		case REG_DMA2SAD_LO:
-		case REG_DMA2DAD_LO:
-		case REG_DMA3SAD_LO:
-		case REG_DMA3DAD_LO:
-			GBAIOWrite32(gba, address, (gba->memory.io[(address >> 1) + 1] << 16) | value);
-			break;
-
-		case REG_DMA0SAD_HI:
-		case REG_DMA0DAD_HI:
-		case REG_DMA1SAD_HI:
-		case REG_DMA1DAD_HI:
-		case REG_DMA2SAD_HI:
-		case REG_DMA2DAD_HI:
-		case REG_DMA3SAD_HI:
-		case REG_DMA3DAD_HI:
-			GBAIOWrite32(gba, address - 2, gba->memory.io[(address >> 1) - 1] | (value << 16));
-			break;
-
-		case REG_DMA0CNT_LO:
-			GBADMAWriteCNT_LO(gba, 0, value & 0x3FFF);
-			break;
-		case REG_DMA0CNT_HI:
-			value = GBADMAWriteCNT_HI(gba, 0, value);
-			break;
-		case REG_DMA1CNT_LO:
-			GBADMAWriteCNT_LO(gba, 1, value & 0x3FFF);
-			break;
-		case REG_DMA1CNT_HI:
-			value = GBADMAWriteCNT_HI(gba, 1, value);
-			break;
-		case REG_DMA2CNT_LO:
-			GBADMAWriteCNT_LO(gba, 2, value & 0x3FFF);
-			break;
-		case REG_DMA2CNT_HI:
-			value = GBADMAWriteCNT_HI(gba, 2, value);
-			break;
-		case REG_DMA3CNT_LO:
-			GBADMAWriteCNT_LO(gba, 3, value);
-			break;
-		case REG_DMA3CNT_HI:
-			value = GBADMAWriteCNT_HI(gba, 3, value);
-			break;
-
-		// Timers
-		case REG_TM0CNT_LO:
-			GBATimerWriteTMCNT_LO(gba, 0, value);
-			return;
-		case REG_TM1CNT_LO:
-			GBATimerWriteTMCNT_LO(gba, 1, value);
-			return;
-		case REG_TM2CNT_LO:
-			GBATimerWriteTMCNT_LO(gba, 2, value);
-			return;
-		case REG_TM3CNT_LO:
-			GBATimerWriteTMCNT_LO(gba, 3, value);
-			return;
-
-		case REG_TM0CNT_HI:
-			value &= 0x00C7;
-			GBATimerWriteTMCNT_HI(gba, 0, value);
-			break;
-		case REG_TM1CNT_HI:
-			value &= 0x00C7;
-			GBATimerWriteTMCNT_HI(gba, 1, value);
-			break;
-		case REG_TM2CNT_HI:
-			value &= 0x00C7;
-			GBATimerWriteTMCNT_HI(gba, 2, value);
-			break;
-		case REG_TM3CNT_HI:
-			value &= 0x00C7;
-			GBATimerWriteTMCNT_HI(gba, 3, value);
-			break;
-
-		// SIO
-		case REG_SIOCNT:
-			GBASIOWriteSIOCNT(&gba->sio, value);
-			break;
-		case REG_RCNT:
-			value &= 0xC1FF;
-			GBASIOWriteRCNT(&gba->sio, value);
-			break;
-		case REG_JOY_TRANS_LO:
-		case REG_JOY_TRANS_HI:
-			gba->memory.io[REG_JOYSTAT >> 1] |= JOYSTAT_TRANS_BIT;
-			// Fall through
-		case REG_SIOMLT_SEND:
-		case REG_JOYCNT:
-		case REG_JOYSTAT:
-		case REG_JOY_RECV_LO:
-		case REG_JOY_RECV_HI:
-			value = GBASIOWriteRegister(&gba->sio, address, value);
-			break;
-
-		// Interrupts and misc
-		case REG_KEYCNT:
-			value &= 0xC3FF;
-			gba->memory.io[address >> 1] = value;
-			GBATestKeypadIRQ(gba);
-			return;
-		case REG_WAITCNT:
-			value &= 0x5FFF;
-			GBAAdjustWaitstates(gba, value);
-			break;
-		case REG_IE:
-			gba->memory.io[REG_IE >> 1] = value;
-			GBATestIRQ(gba, 1);
-			return;
-		case REG_IF:
-			value = gba->memory.io[REG_IF >> 1] & ~value;
-			gba->memory.io[REG_IF >> 1] = value;
-			GBATestIRQ(gba, 1);
-			return;
-		case REG_IME:
-			gba->memory.io[REG_IME >> 1] = value;
-			GBATestIRQ(gba, 1);
-			return;
-		case REG_MAX:
-			// Some bad interrupt libraries will write to this
-			break;
-		case REG_DEBUG_ENABLE:
-			gba->debug = value == 0xC0DE;
-			return;
-		case REG_DEBUG_FLAGS:
-			if (gba->debug) {
-				GBADebug(gba, value);
-				return;
-			}
-			// Fall through
-		default:
-			if (address >= REG_DEBUG_STRING && address - REG_DEBUG_STRING < sizeof(gba->debugString)) {
-				STORE_16LE(value, address - REG_DEBUG_STRING, gba->debugString);
-				return;
-			}
-			mLOG(GBA_IO, STUB, "Stub I/O register write: %03X", address);
-			if (address >= REG_MAX) {
-				mLOG(GBA_IO, GAME_ERROR, "Write to unused I/O register: %03X", address);
-				return;
-			}
-			break;
 		}
+		// Fall through
+	default:
+		if (address >= REG_DEBUG_STRING && address - REG_DEBUG_STRING < sizeof(gba->debugString)) {
+			STORE_16LE(value, address - REG_DEBUG_STRING, gba->debugString);
+			return;
+		}
+		mLOG(GBA_IO, STUB, "Stub I/O register write: %03X", address);
+		if (address >= REG_MAX) {
+			mLOG(GBA_IO, GAME_ERROR, "Write to unused I/O register: %03X", address);
+			return;
+		}
+		break;
 	}
 	gba->memory.io[address >> 1] = value;
 }
@@ -615,6 +623,9 @@ void GBAIOWrite8(struct GBA* gba, uint32_t address, uint8_t value) {
 
 void GBAIOWrite32(struct GBA* gba, uint32_t address, uint32_t value) {
 	switch (address) {
+	// Wave RAM can be written and read even if the audio hardware is disabled.
+	// However, it is not possible to switch between the two banks because it
+	// isn't possible to write to register SOUND3CNT_LO.
 	case REG_WAVE_RAM0_LO:
 		GBAAudioWriteWaveRAM(&gba->audio, 0, value);
 		break;
@@ -629,7 +640,7 @@ void GBAIOWrite32(struct GBA* gba, uint32_t address, uint32_t value) {
 		break;
 	case REG_FIFO_A_LO:
 	case REG_FIFO_B_LO:
-		GBAAudioWriteFIFO(&gba->audio, address, value);
+		value = GBAAudioWriteFIFO(&gba->audio, address, value);
 		break;
 	case REG_DMA0SAD_LO:
 		value = GBADMAWriteSAD(gba, 0, value);
@@ -724,8 +735,7 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 		GBATimerUpdateRegister(gba, 3, 2);
 		break;
 
-	case REG_KEYINPUT:
-		{
+	case REG_KEYINPUT: {
 			size_t c;
 			for (c = 0; c < mCoreCallbacksListSize(&gba->coreCallbacks); ++c) {
 				struct mCoreCallbacks* callbacks = mCoreCallbacksListGetPointer(&gba->coreCallbacks, c);
@@ -733,10 +743,6 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 					callbacks->keysRead(callbacks->context);
 				}
 			}
-		}
-		if (gba->rr && gba->rr->isPlaying(gba->rr)) {
-			return 0x3FF ^ gba->rr->queryInput(gba->rr);
-		} else {
 			uint16_t input = 0;
 			if (gba->keyCallback) {
 				input = gba->keyCallback->readKeys(gba->keyCallback);
@@ -757,12 +763,9 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 					}
 				}
 			}
-			if (gba->rr && gba->rr->isRecording(gba->rr)) {
-				gba->rr->logInput(gba->rr, input);
-			}
-			return 0x3FF ^ input;
+			gba->memory.io[address >> 1] = 0x3FF ^ input;
 		}
-
+		break;
 	case REG_SIOCNT:
 		return gba->sio.siocnt;
 	case REG_RCNT:
@@ -834,13 +837,34 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 
 	case REG_JOY_RECV_LO:
 	case REG_JOY_RECV_HI:
-		gba->memory.io[REG_JOYSTAT >> 1] &= ~JOYSTAT_RECV_BIT;
+		gba->memory.io[REG_JOYSTAT >> 1] &= ~JOYSTAT_RECV;
 		break;
 
 	case REG_SOUNDBIAS:
 	case REG_POSTFLG:
 		mLOG(GBA_IO, STUB, "Stub I/O register read: %03x", address);
 		break;
+
+	// Wave RAM can be written and read even if the audio hardware is disabled.
+	// However, it is not possible to switch between the two banks because it
+	// isn't possible to write to register SOUND3CNT_LO.
+	case REG_WAVE_RAM0_LO:
+		return GBAAudioReadWaveRAM(&gba->audio, 0) & 0xFFFF;
+	case REG_WAVE_RAM0_HI:
+		return GBAAudioReadWaveRAM(&gba->audio, 0) >> 16;
+	case REG_WAVE_RAM1_LO:
+		return GBAAudioReadWaveRAM(&gba->audio, 1) & 0xFFFF;
+	case REG_WAVE_RAM1_HI:
+		return GBAAudioReadWaveRAM(&gba->audio, 1) >> 16;
+	case REG_WAVE_RAM2_LO:
+		return GBAAudioReadWaveRAM(&gba->audio, 2) & 0xFFFF;
+	case REG_WAVE_RAM2_HI:
+		return GBAAudioReadWaveRAM(&gba->audio, 2) >> 16;
+	case REG_WAVE_RAM3_LO:
+		return GBAAudioReadWaveRAM(&gba->audio, 3) & 0xFFFF;
+	case REG_WAVE_RAM3_HI:
+		return GBAAudioReadWaveRAM(&gba->audio, 3) >> 16;
+
 	case REG_SOUND1CNT_LO:
 	case REG_SOUND1CNT_HI:
 	case REG_SOUND1CNT_X:
@@ -858,6 +882,7 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 		}
 		// Fall through
 	case REG_DISPCNT:
+	case REG_GREENSWP:
 	case REG_DISPSTAT:
 	case REG_VCOUNT:
 	case REG_BG0CNT:
@@ -870,14 +895,6 @@ uint16_t GBAIORead(struct GBA* gba, uint32_t address) {
 	case REG_BLDALPHA:
 	case REG_SOUNDCNT_HI:
 	case REG_SOUNDCNT_X:
-	case REG_WAVE_RAM0_LO:
-	case REG_WAVE_RAM0_HI:
-	case REG_WAVE_RAM1_LO:
-	case REG_WAVE_RAM1_HI:
-	case REG_WAVE_RAM2_LO:
-	case REG_WAVE_RAM2_HI:
-	case REG_WAVE_RAM3_LO:
-	case REG_WAVE_RAM3_HI:
 	case REG_DMA0CNT_HI:
 	case REG_DMA1CNT_HI:
 	case REG_DMA2CNT_HI:
@@ -951,11 +968,15 @@ void GBAIOSerialize(struct GBA* gba, struct GBASerializedState* state) {
 	}
 
 	STORE_32(gba->memory.dmaTransferRegister, 0, &state->dmaTransferRegister);
+	STORE_32(gba->dmaPC, 0, &state->dmaBlockPC);
 
 	GBAHardwareSerialize(&gba->memory.hw, state);
 }
 
 void GBAIODeserialize(struct GBA* gba, const struct GBASerializedState* state) {
+	LOAD_16(gba->memory.io[REG_SOUNDCNT_X >> 1], REG_SOUNDCNT_X, state->io);
+	GBAAudioWriteSOUNDCNT_X(&gba->audio, gba->memory.io[REG_SOUNDCNT_X >> 1]);
+
 	int i;
 	for (i = 0; i < REG_MAX; i += 2) {
 		if (_isWSpecialRegister[i >> 1]) {
@@ -986,9 +1007,11 @@ void GBAIODeserialize(struct GBA* gba, const struct GBASerializedState* state) {
 		LOAD_32(gba->memory.dma[i].nextCount, 0, &state->dma[i].nextCount);
 		LOAD_32(gba->memory.dma[i].when, 0, &state->dma[i].when);
 	}
-	GBAAudioWriteSOUNDCNT_X(&gba->audio, gba->memory.io[REG_SOUNDCNT_X >> 1]);
+	gba->sio.siocnt = gba->memory.io[REG_SIOCNT >> 1];
+	GBASIOWriteRCNT(&gba->sio, gba->memory.io[REG_RCNT >> 1]);
 
 	LOAD_32(gba->memory.dmaTransferRegister, 0, &state->dmaTransferRegister);
+	LOAD_32(gba->dmaPC, 0, &state->dmaBlockPC);
 
 	GBADMAUpdate(gba);
 	GBAHardwareDeserialize(&gba->memory.hw, state);
