@@ -16,14 +16,19 @@ int GBAVideoRendererCleanOAM(struct GBAObj* oam, struct GBAVideoRendererSprite*
 		LOAD_16LE(obj.b, 0, &oam[i].b);
 		LOAD_16LE(obj.c, 0, &oam[i].c);
 		if (GBAObjAttributesAIsTransformed(obj.a) || !GBAObjAttributesAIsDisable(obj.a)) {
+			int width = GBAVideoObjSizes[GBAObjAttributesAGetShape(obj.a) * 4 + GBAObjAttributesBGetSize(obj.b)][0];
 			int height = GBAVideoObjSizes[GBAObjAttributesAGetShape(obj.a) * 4 + GBAObjAttributesBGetSize(obj.b)][1];
+			int cycles = width;
 			if (GBAObjAttributesAIsTransformed(obj.a)) {
 				height <<= GBAObjAttributesAGetDoubleSize(obj.a);
+				width <<= GBAObjAttributesAGetDoubleSize(obj.a);
+				cycles = 10 + width * 2;
 			}
 			if (GBAObjAttributesAGetY(obj.a) < GBA_VIDEO_VERTICAL_PIXELS || GBAObjAttributesAGetY(obj.a) + height >= VIDEO_VERTICAL_TOTAL_PIXELS) {
 				int y = GBAObjAttributesAGetY(obj.a) + offsetY;
 				sprites[oamMax].y = y;
 				sprites[oamMax].endY = y + height;
+				sprites[oamMax].cycles = cycles;
 				sprites[oamMax].obj = obj;
 				sprites[oamMax].index = i;
 				++oamMax;
