@@ -20,6 +20,7 @@
 #include <mgba-util/gui/file-select.h>
 #include <mgba-util/gui/font.h>
 #include <mgba-util/gui/menu.h>
+#include <mgba-util/math.h>
 #include <mgba-util/memory.h>
 
 #include <mgba-util/platform/3ds/3ds-vfs.h>
@@ -113,8 +114,13 @@ static bool _initGpu(void) {
 		return false;
 	}
 
-	topScreen[0] = C3D_RenderTargetCreate(240, 400, GPU_RB_RGB8, 0);
-	topScreen[1] = C3D_RenderTargetCreate(240, 400, GPU_RB_RGB8, 0);
+	if (gfxIsWide()) {
+		topScreen[0] = C3D_RenderTargetCreate(240, 800, GPU_RB_RGB8, 0);
+		topScreen[1] = C3D_RenderTargetCreate(240, 800, GPU_RB_RGB8, 0);
+	} else {
+		topScreen[0] = C3D_RenderTargetCreate(240, 400, GPU_RB_RGB8, 0);
+		topScreen[1] = C3D_RenderTargetCreate(240, 400, GPU_RB_RGB8, 0);
+	}
 	bottomScreen[0] = C3D_RenderTargetCreate(240, 320, GPU_RB_RGB8, 0);
 	bottomScreen[1] = C3D_RenderTargetCreate(240, 320, GPU_RB_RGB8, 0);
 	if (!topScreen[0] || !topScreen[1] || !bottomScreen[0] || !bottomScreen[1]) {
@@ -181,6 +187,7 @@ static void _cleanup(void) {
 	camExit();
 	ndspExit();
 	ptmuExit();
+	mcuHwcExit();
 }
 
 static void _map3DSKey(struct mInputMap* map, int ctrKey, enum GBAKey key) {
@@ -231,15 +238,20 @@ static void _drawEnd(void) {
 static int _batteryState(void) {
 	u8 charge;
 	u8 adapter;
-	PTMU_GetBatteryLevel(&charge);
 	PTMU_GetBatteryChargeState(&adapter);
+
 	int state = 0;
+	if (R_SUCCEEDED(MCUHWC_GetBatteryLevel(&charge))) {
+		charge |= BATTERY_PERCENTAGE_VALID;
+	} else {
+		PTMU_GetBatteryLevel(&charge);
+		if (charge > 0) {
+			--charge;
+		}
+	}
 	if (adapter) {
 		state |= BATTERY_CHARGING;
 	}
-	if (charge > 0) {
-		--charge;
-	}
 	return state | charge;
 }
 
@@ -289,7 +301,7 @@ static void _setup(struct mGUIRunner* runner) {
 	_map3DSKey(&runner->core->inputMap, KEY_L, GBA_KEY_L);
 	_map3DSKey(&runner->core->inputMap, KEY_R, GBA_KEY_R);
 
-	outputBuffer = linearMemAlign(256 * 224 * sizeof(color_t), 0x80);
+	memset(outputBuffer, 0, 256 * 224 * sizeof(color_t));
 	runner->core->setVideoBuffer(runner->core, outputBuffer, 256);
 
 	unsigned mode;
@@ -316,7 +328,7 @@ static void _gameLoaded(struct mGUIRunner* runner) {
 	switch (runner->core->platform(runner->core)) {
 #ifdef M_CORE_GBA
 		// TODO: Move these to callbacks
-	case PLATFORM_GBA:
+	case mPLATFORM_GBA:
 		if (((struct GBA*) runner->core->board)->memory.hw.devices & HW_TILT) {
 			HIDUSER_EnableAccelerometer();
 		}
@@ -326,7 +338,7 @@ static void _gameLoaded(struct mGUIRunner* runner) {
 		break;
 #endif
 #ifdef M_CORE_GB
-	case PLATFORM_GB:
+	case mPLATFORM_GB:
 		if (((struct GB*) runner->core->board)->memory.mbcType == GB_MBC7) {
 			HIDUSER_EnableAccelerometer();
 		}
@@ -400,7 +412,7 @@ static void _gameUnloaded(struct mGUIRunner* runner) {
 	switch (runner->core->platform(runner->core)) {
 #ifdef M_CORE_GBA
 		// TODO: Move these to callbacks
-	case PLATFORM_GBA:
+	case mPLATFORM_GBA:
 		if (((struct GBA*) runner->core->board)->memory.hw.devices & HW_TILT) {
 			HIDUSER_DisableAccelerometer();
 		}
@@ -410,7 +422,7 @@ static void _gameUnloaded(struct mGUIRunner* runner) {
 		break;
 #endif
 #ifdef M_CORE_GB
-	case PLATFORM_GB:
+	case mPLATFORM_GB:
 		if (((struct GB*) runner->core->board)->memory.mbcType == GB_MBC7) {
 			HIDUSER_DisableAccelerometer();
 		}
@@ -463,6 +475,7 @@ static u32 _setupTex(int out, bool faded) {
 
 static void _drawTex(struct mCore* core, bool faded, bool both) {
 	unsigned screen_w, screen_h;
+	bool isWide = screenMode >= SM_PA_TOP && gfxIsWide();
 	switch (screenMode) {
 	case SM_PA_BOTTOM:
 		C3D_FrameDrawOn(bottomScreen[doubleBuffer]);
@@ -471,7 +484,7 @@ static void _drawTex(struct mCore* core, bool faded, bool both) {
 		break;
 	case SM_PA_TOP:
 		C3D_FrameDrawOn(topScreen[doubleBuffer]);
-		screen_w = 400;
+		screen_w = isWide ? 800 : 400;
 		screen_h = 240;
 		break;
 	default:
@@ -480,6 +493,7 @@ static void _drawTex(struct mCore* core, bool faded, bool both) {
 		screen_h = 512;
 		break;
 	}
+	int wide = isWide ? 2 : 1;
 
 	unsigned corew, coreh;
 	core->desiredVideoDimensions(core, &corew, &coreh);
@@ -490,24 +504,16 @@ static void _drawTex(struct mCore* core, bool faded, bool both) {
 		w = GB_VIDEO_HORIZONTAL_PIXELS;
 		h = GB_VIDEO_VERTICAL_PIXELS;
 	}
-	int innerw = w;
-	int innerh = h;
-	// Get greatest common divisor
-	while (w != 0) {
-		int temp = h % w;
-		h = w;
-		w = temp;
-	}
-	int gcd = h;
-	unsigned aspectw = innerw / gcd;
-	unsigned aspecth = innerh / gcd;
+	int aspectw = w;
+	int aspecth = h;
+	int gcd = reduceFraction(&aspecth, &aspectw);
 	int x = 0;
 	int y = 0;
 
 	switch (screenMode) {
 	case SM_PA_TOP:
 	case SM_PA_BOTTOM:
-		w = corew;
+		w = corew * wide;
 		h = coreh;
 		x = (screen_w - w) / 2;
 		y = (screen_h - h) / 2;
@@ -537,8 +543,8 @@ static void _drawTex(struct mCore* core, bool faded, bool both) {
 	}
 	ctrFlushBatch();
 
-	innerw = corew;
-	innerh = coreh;
+	int innerw = corew;
+	int innerh = coreh;
 	corew = w;
 	coreh = h;
 	screen_h = 240;
@@ -547,7 +553,7 @@ static void _drawTex(struct mCore* core, bool faded, bool both) {
 		screen_w = 320;
 	} else {
 		C3D_FrameDrawOn(topScreen[doubleBuffer]);
-		screen_w = 400;
+		screen_w = isWide ? 800 : 400;
 	}
 	ctrSetViewportSize(screen_w, screen_h, true);
 
@@ -559,6 +565,7 @@ static void _drawTex(struct mCore* core, bool faded, bool both) {
 	case SM_AF_BOTTOM:
 		afw = screen_w / (float) aspectw;
 		afh = screen_h / (float) aspecth;
+		innerw *= wide;
 		if (afw * aspecth > screen_h) {
 			w = innerw * afh / gcd;
 			h = innerh * afh / gcd;
@@ -813,6 +820,7 @@ int main() {
 	camera.cam = SELECT_IN1;
 
 	ptmuInit();
+	mcuHwcInit();
 	camInit();
 
 	hasSound = NO_SOUND;
@@ -836,6 +844,14 @@ int main() {
 
 	gfxInit(GSP_BGR8_OES, GSP_BGR8_OES, true);
 
+	u8 model = 0;
+	cfguInit();
+	CFGU_GetSystemModel(&model);
+	if (model != 3 /* o2DS */) {
+		gfxSetWide(true);
+	}
+	cfguExit();
+
 	if (!_initGpu()) {
 		outputTexture[0].data = 0;
 		_cleanup();
@@ -868,6 +884,7 @@ int main() {
 		_cleanup();
 		return 1;
 	}
+	outputBuffer = linearMemAlign(256 * 224 * sizeof(color_t), 0x80);
 
 	struct mGUIRunner runner = {
 		.params = {
