@@ -15,13 +15,13 @@ namespace QGBA {
 
 class CoreController;
 
-class AssetTile : public QGroupBox {
+class AssetTile : public AssetInfo {
 Q_OBJECT
 
 public:
 	AssetTile(QWidget* parent = nullptr);
 	void setController(std::shared_ptr<CoreController>);
-	void addCustomProperty(const QString& id, const QString& visibleName);
+	QImage activeTile() const { return m_activeTile; }
 
 public slots:
 	void setPalette(int);
@@ -29,7 +29,9 @@ public slots:
 	void selectIndex(int);
 	void setFlip(bool h, bool v);
 	void selectColor(int);
-	void setCustomProperty(const QString& id, const QVariant& value);
+
+protected:
+	int customLocation(const QString& id = {}) override;
 
 private:
 	Ui::AssetTile m_ui;
@@ -47,6 +49,7 @@ private:
 	bool m_flipV = false;
 
 	QMap<QString, QLabel*> m_customProperties;
+	QImage m_activeTile;
 };
 
 }
