@@ -41,7 +41,7 @@ CheatsView::CheatsView(std::shared_ptr<CoreController> controller, QWidget* pare
 	QPushButton* add;
 	switch (controller->platform()) {
 #ifdef M_CORE_GBA
-	case PLATFORM_GBA:
+	case mPLATFORM_GBA:
 		connect(m_ui.add, &QPushButton::clicked, [this]() {
 			enterCheat(GBA_CHEAT_AUTODETECT);
 		});
@@ -66,7 +66,7 @@ CheatsView::CheatsView(std::shared_ptr<CoreController> controller, QWidget* pare
 		break;
 #endif
 #ifdef M_CORE_GB
-	case PLATFORM_GB:
+	case mPLATFORM_GB:
 		connect(m_ui.add, &QPushButton::clicked, [this]() {
 			enterCheat(GB_CHEAT_AUTODETECT);
 		});
@@ -126,6 +126,7 @@ void CheatsView::addSet() {
 	CoreController::Interrupter interrupter(m_controller);
 	mCheatSet* set = m_controller->cheatDevice()->createSet(m_controller->cheatDevice(), nullptr);
 	m_model.addSet(set);
+	m_ui.cheatList->selectionModel()->select(m_model.index(m_model.rowCount() - 1, 0, QModelIndex()), QItemSelectionModel::ClearAndSelect);
 }
 
 void CheatsView::removeSet() {
@@ -165,6 +166,8 @@ void CheatsView::enterCheat(int codeType) {
 		mCheatAddLine(set, string.toUtf8().constData(), codeType);
 		m_model.endAppendRow();
 	}
-	set->refresh(set, m_controller->cheatDevice());
+	if (set->refresh) {
+		set->refresh(set, m_controller->cheatDevice());
+	}
 	m_ui.codeEntry->clear();
 }
