@@ -16,21 +16,28 @@
 #endif
 #endif
 
+#include <QAtomicInt>
 #include <QElapsedTimer>
-#include <QOpenGLContext>
+#include <QHash>
 #include <QList>
 #include <QMouseEvent>
+#include <QOpenGLContext>
 #include <QPainter>
 #include <QQueue>
 #include <QThread>
 #include <QTimer>
 
+#include <array>
+
+#include "CoreController.h"
 #include "VideoProxy.h"
 
 #include "platform/video-backend.h"
 
 class QOpenGLPaintDevice;
 
+uint qHash(const QSurfaceFormat&, uint seed = 0);
+
 namespace QGBA {
 
 class PainterGL;
@@ -48,6 +55,8 @@ public:
 	void setVideoProxy(std::shared_ptr<VideoProxy>) override;
 	int framebufferHandle() override;
 
+	static bool supportsFormat(const QSurfaceFormat&);
+
 public slots:
 	void stopDrawing() override;
 	void pauseDrawing() override;
@@ -71,10 +80,12 @@ protected:
 private:
 	void resizePainter();
 
+	static QHash<QSurfaceFormat, bool> s_supports;
+
 	bool m_isDrawing = false;
-	QOpenGLContext* m_gl;
-	PainterGL* m_painter;
-	QThread* m_drawThread = nullptr;
+	bool m_hasStarted = false;
+	std::unique_ptr<PainterGL> m_painter;
+	QThread m_drawThread;
 	std::shared_ptr<CoreController> m_context;
 };
 
@@ -82,18 +93,24 @@ class PainterGL : public QObject {
 Q_OBJECT
 
 public:
-	PainterGL(QWindow* surface, QOpenGLContext* parent, int forceVersion = 0);
+	PainterGL(QWindow* surface, const QSurfaceFormat& format);
 	~PainterGL();
 
+	void setThread(QThread*);
 	void setContext(std::shared_ptr<CoreController>);
 	void setMessagePainter(MessagePainter*);
 	void enqueue(const uint32_t* backing);
 
 	bool supportsShaders() const { return m_supportsShaders; }
+	int glTex();
 
 	void setVideoProxy(std::shared_ptr<VideoProxy>);
+	void interrupt();
 
 public slots:
+	void create();
+	void destroy();
+
 	void forceDraw();
 	void draw();
 	void start();
@@ -112,29 +129,36 @@ public slots:
 	void clearShaders();
 	VideoShader* shaders();
 
-	int glTex();
+signals:
+	void started();
 
 private:
+	void makeCurrent();
 	void performDraw();
 	void dequeue();
-	void dequeueAll();
+	void dequeueAll(bool keep = false);
 
+	std::array<std::array<uint32_t, 0x100000>, 3> m_buffers;
 	QList<uint32_t*> m_free;
 	QQueue<uint32_t*> m_queue;
-	uint32_t* m_buffer;
+	uint32_t* m_buffer = nullptr;
 	QPainter m_painter;
 	QMutex m_mutex;
 	QWindow* m_surface;
-	QOpenGLPaintDevice* m_window;
-	QOpenGLContext* m_gl;
+	QSurfaceFormat m_format;
+	std::unique_ptr<QOpenGLPaintDevice> m_window;
+	std::unique_ptr<QOpenGLContext> m_gl;
 	bool m_active = false;
 	bool m_started = false;
-	std::shared_ptr<CoreController> m_context = nullptr;
+	QTimer m_drawTimer;
+	std::shared_ptr<CoreController> m_context;
+	CoreController::Interrupter m_interrupter;
 	bool m_supportsShaders;
 	bool m_showOSD;
 	VideoShader m_shader{};
 	VideoBackend* m_backend = nullptr;
 	QSize m_size;
+	QSize m_dims;
 	MessagePainter* m_messagePainter = nullptr;
 	QElapsedTimer m_delayTimer;
 	std::shared_ptr<VideoProxy> m_videoProxy;
