@@ -25,7 +25,6 @@
 #include <QAction>
 #include <QButtonGroup>
 #include <QClipboard>
-#include <QFontDatabase>
 #include <QMouseEvent>
 #include <QRadioButton>
 #include <QTimer>
@@ -41,7 +40,7 @@ MapView::MapView(std::shared_ptr<CoreController> controller, QWidget* parent)
 
 	switch (m_controller->platform()) {
 #ifdef M_CORE_GBA
-	case PLATFORM_GBA:
+	case mPLATFORM_GBA:
 		m_boundary = 2048;
 		m_addressBase = BASE_VRAM;
 		m_addressWidth = 8;
@@ -54,7 +53,7 @@ MapView::MapView(std::shared_ptr<CoreController> controller, QWidget* parent)
 		break;
 #endif
 #ifdef M_CORE_GB
-	case PLATFORM_GB:
+	case mPLATFORM_GB:
 		m_boundary = 1024;
 		m_addressBase = GB_BASE_VRAM;
 		m_addressWidth = 4;
@@ -110,19 +109,25 @@ MapView::MapView(std::shared_ptr<CoreController> controller, QWidget* parent)
 }
 
 void MapView::selectMap(int map) {
-	if (map >= mMapCacheSetSize(&m_cacheSet->maps)) {
+	if (map == m_map || map < 0) {
 		return;
 	}
-	if (map == m_map) {
+	if (static_cast<unsigned>(map) >= mMapCacheSetSize(&m_cacheSet->maps)) {
 		return;
 	}
 	m_map = map;
+	m_mapStatus.fill({});
 	updateTiles(true);
 }
 
 void MapView::selectTile(int x, int y) {
 	CoreController::Interrupter interrupter(m_controller);
 	mMapCache* mapCache = mMapCacheSetGetPointer(&m_cacheSet->maps, m_map);
+	int tiles = mMapCacheTileCount(mapCache);
+	if (m_mapStatus.size() != tiles) {
+		m_mapStatus.resize(tiles);
+		m_mapStatus.fill({});
+	}
 	size_t tileCache = mTileCacheSetIndex(&m_cacheSet->tiles, mapCache->tileCache);
 	m_ui.tile->setBoundary(m_boundary, tileCache, tileCache);
 	uint32_t location = mMapCacheTileId(mapCache, x, y);
@@ -147,7 +152,7 @@ void MapView::selectTile(int x, int y) {
 		.arg(location, m_addressWidth, 16, QChar('0')));
 }
 
-bool MapView::eventFilter(QObject* obj, QEvent* event) {
+bool MapView::eventFilter(QObject*, QEvent* event) {
 	if (event->type() != QEvent::MouseButtonPress) {
 		return false;
 	}
@@ -159,7 +164,7 @@ bool MapView::eventFilter(QObject* obj, QEvent* event) {
 	return true;
 }
 
-void MapView::updateTilesGBA(bool force) {
+void MapView::updateTilesGBA(bool) {
 	{
 		CoreController::Interrupter interrupter(m_controller);
 		int bitmap = -1;
@@ -168,7 +173,7 @@ void MapView::updateTilesGBA(bool force) {
 		QString offset(tr("N/A"));
 		QString transform(tr("N/A"));
 #ifdef M_CORE_GBA
-		if (m_controller->platform() == PLATFORM_GBA) {
+		if (m_controller->platform() == mPLATFORM_GBA) {
 			uint16_t* io = static_cast<GBA*>(m_controller->thread()->core->board)->memory.io;
 			int mode = GBARegisterDISPCNTGetMode(io[REG_DISPCNT >> 1]);
 			if (m_map == 2 && mode > 2) {
@@ -202,7 +207,7 @@ void MapView::updateTilesGBA(bool force) {
 		}
 #endif
 #ifdef M_CORE_GB
-		if (m_controller->platform() == PLATFORM_GB) {
+		if (m_controller->platform() == mPLATFORM_GB) {
 			uint8_t* io = static_cast<GB*>(m_controller->thread()->core->board)->memory.io;
 			int x = io[m_map == 0 ? 0x42 : 0x4A];
 			int y = io[m_map == 0 ? 0x43 : 0x4B];
@@ -222,7 +227,7 @@ void MapView::updateTilesGBA(bool force) {
 			m_rawMap = QImage(QSize(width, height), QImage::Format_ARGB32);
 			uchar* bgBits = m_rawMap.bits();
 			for (int j = 0; j < height; ++j) {
-				mBitmapCacheCleanRow(bitmapCache, m_bitmapStatus, j);
+				mBitmapCacheCleanRow(bitmapCache, m_bitmapStatus.data(), j);
 				memcpy(static_cast<void*>(&bgBits[width * j * 4]), mBitmapCacheGetRow(bitmapCache, j), width * 4);
 			}
 			m_rawMap = m_rawMap.convertToFormat(QImage::Format_RGB32).rgbSwapped();
@@ -240,7 +245,7 @@ void MapView::updateTilesGBA(bool force) {
 			m_ui.bgInfo->setCustomProperty("priority", priority);
 			m_ui.bgInfo->setCustomProperty("offset", offset);
 			m_ui.bgInfo->setCustomProperty("transform", transform);
-			m_rawMap = compositeMap(m_map, m_mapStatus);
+			m_rawMap = compositeMap(m_map, &m_mapStatus);
 		}
 	}
 	QPixmap map = QPixmap::fromImage(m_rawMap.convertToFormat(QImage::Format_RGB32));
