@@ -10,7 +10,6 @@
 
 #include <QAction>
 #include <QClipboard>
-#include <QFontDatabase>
 #include <QListWidgetItem>
 #include <QTimer>
 
@@ -34,7 +33,7 @@ ObjView::ObjView(std::shared_ptr<CoreController> controller, QWidget* parent)
 	m_ui.setupUi(this);
 	m_ui.tile->setController(controller);
 
-	const QFont font = QFontDatabase::systemFont(QFontDatabase::FixedFont);
+	const QFont font = GBAApp::app()->monospaceFont();
 
 	m_ui.x->setFont(font);
 	m_ui.y->setFont(font);
@@ -44,6 +43,10 @@ ObjView::ObjView(std::shared_ptr<CoreController> controller, QWidget* parent)
 	m_ui.priority->setFont(font);
 	m_ui.palette->setFont(font);
 	m_ui.transform->setFont(font);
+	m_ui.xformPA->setFont(font);
+	m_ui.xformPB->setFont(font);
+	m_ui.xformPC->setFont(font);
+	m_ui.xformPD->setFont(font);
 	m_ui.mode->setFont(font);
 
 	connect(m_ui.tiles, &TilePainter::indexPressed, this, &ObjView::translateIndex);
@@ -128,15 +131,19 @@ void ObjView::updateTilesGBA(bool force) {
 	m_objInfo = newInfo;
 	m_tileOffset = newInfo.tile;
 	mTileCache* tileCache = mTileCacheSetGetPointer(&m_cacheSet->tiles, newInfo.paletteSet);
-
+	unsigned maxTiles = mTileCacheSystemInfoGetMaxTiles(tileCache->sysConfig);
 	int i = 0;
-	for (int y = 0; y < newInfo.height; ++y) {
-		for (int x = 0; x < newInfo.width; ++x, ++i, ++tile, ++tileBase) {
-			const color_t* data = mTileCacheGetTileIfDirty(tileCache, &m_tileStatus[16 * tileBase], tile, newInfo.paletteId);
-			if (data) {
-				m_ui.tiles->setTile(i, data);
-			} else if (force) {
-				m_ui.tiles->setTile(i, mTileCacheGetTile(tileCache, tile, newInfo.paletteId));
+	for (unsigned y = 0; y < newInfo.height; ++y) {
+		for (unsigned x = 0; x < newInfo.width; ++x, ++i, ++tile, ++tileBase) {
+			if (tile < maxTiles) {
+				const color_t* data = mTileCacheGetTileIfDirty(tileCache, &m_tileStatus[16 * tileBase], tile, newInfo.paletteId);
+				if (data) {
+					m_ui.tiles->setTile(i, data);
+				} else if (force) {
+					m_ui.tiles->setTile(i, mTileCacheGetTile(tileCache, tile, newInfo.paletteId));
+				}
+			} else {
+				m_ui.tiles->clearTile(i);
 			}
 		}
 		tile += newInfo.stride - newInfo.width;
@@ -157,9 +164,23 @@ void ObjView::updateTilesGBA(bool force) {
 	m_ui.mosaic->setChecked(GBAObjAttributesAIsMosaic(obj->a));
 
 	if (GBAObjAttributesAIsTransformed(obj->a)) {
-		m_ui.transform->setText(QString::number(GBAObjAttributesBGetMatIndex(obj->b)));
+		int mtxId = GBAObjAttributesBGetMatIndex(obj->b);
+		struct GBAOAMMatrix mat;
+		LOAD_16LE(mat.a, 0, &gba->video.oam.mat[mtxId].a);
+		LOAD_16LE(mat.b, 0, &gba->video.oam.mat[mtxId].b);
+		LOAD_16LE(mat.c, 0, &gba->video.oam.mat[mtxId].c);
+		LOAD_16LE(mat.d, 0, &gba->video.oam.mat[mtxId].d);
+		m_ui.transform->setText(QString::number(mtxId));
+		m_ui.xformPA->setText(QString("%0").arg(mat.a / 256., 5, 'f', 2));
+		m_ui.xformPB->setText(QString("%0").arg(mat.b / 256., 5, 'f', 2));
+		m_ui.xformPC->setText(QString("%0").arg(mat.c / 256., 5, 'f', 2));
+		m_ui.xformPD->setText(QString("%0").arg(mat.d / 256., 5, 'f', 2));
 	} else {
 		m_ui.transform->setText(tr("Off"));
+		m_ui.xformPA->setText(tr("---"));
+		m_ui.xformPB->setText(tr("---"));
+		m_ui.xformPC->setText(tr("---"));
+		m_ui.xformPD->setText(tr("---"));
 	}
 
 	switch (GBAObjAttributesAGetMode(obj->a)) {
@@ -207,7 +228,7 @@ void ObjView::updateTilesGB(bool force) {
 
 	int i = 0;
 	m_ui.tile->setPalette(newInfo.paletteId);
-	for (int y = 0; y < newInfo.height; ++y, ++i) {
+	for (unsigned y = 0; y < newInfo.height; ++y, ++i) {
 		unsigned t = tile + i;
 		const color_t* data = mTileCacheGetTileIfDirty(tileCache, &m_tileStatus[8 * t], t, newInfo.paletteId);
 		if (data) {
@@ -230,6 +251,10 @@ void ObjView::updateTilesGB(bool force) {
 	m_ui.doubleSize->setChecked(false);
 	m_ui.mosaic->setChecked(false);
 	m_ui.transform->setText(tr("N/A"));
+	m_ui.xformPA->setText(tr("---"));
+	m_ui.xformPB->setText(tr("---"));
+	m_ui.xformPC->setText(tr("---"));
+	m_ui.xformPD->setText(tr("---"));
 	m_ui.mode->setText(tr("N/A"));
 }
 #endif
@@ -250,7 +275,7 @@ void ObjView::updateObjList(int maxObj) {
 		QListWidgetItem* item = m_objs[i];
 		ObjInfo info;
 		lookupObj(i, &info);
-		item->setIcon(QPixmap::fromImage(std::move(compositeObj(info))));
+		item->setIcon(QPixmap::fromImage(compositeObj(info)));
 	}
 }
 
