@@ -6,6 +6,7 @@
 #pragma once
 
 #include <QDialog>
+#include <QMap>
 
 #include "ColorPicker.h"
 #include "LogConfigModel.h"
@@ -29,6 +30,21 @@ class SettingsView : public QDialog {
 Q_OBJECT
 
 public:
+	enum class Page {
+		AV,
+		INTERFACE,
+		EMULATION,
+		ENHANCEMENTS,
+		BIOS,
+		PATHS,
+		LOGGING,
+		GB,
+		KEYBOARD,
+		CONTROLLERS,
+		SHORTCUTS,
+		SHADERS,
+	};
+
 	SettingsView(ConfigController* controller, InputController* inputController, ShortcutController* shortcutController, LogController* logController, QWidget* parent = nullptr);
 	~SettingsView();
 
@@ -45,8 +61,12 @@ signals:
 	void languageChanged();
 	void libraryCleared();
 
+public slots:
+	void selectPage(Page);
+
 private slots:
 	void selectBios(QLineEdit*);
+	void selectPath(QLineEdit*, QCheckBox*);
 	void updateConfig();
 	void reloadConfig();
 
@@ -61,9 +81,14 @@ private:
 #ifdef M_CORE_GB
 	uint32_t m_gbColors[12]{};
 	ColorPicker m_colorPickers[12];
-	static QList<enum GBModel> s_gbModelList;
 #endif
 
+	QMap<Page, int> m_pageIndex;
+
+	QString makePortablePath(const QString& path);
+
+	void addPage(const QString& name, QWidget* view, Page index);
+
 	void saveSetting(const char* key, const QAbstractButton*);
 	void saveSetting(const char* key, const QComboBox*);
 	void saveSetting(const char* key, const QDoubleSpinBox*);
