@@ -10,6 +10,7 @@
 #include "ShortcutController.h"
 #include "ShortcutModel.h"
 
+#include <QFontMetrics>
 #include <QKeyEvent>
 
 using namespace QGBA;
@@ -87,7 +88,7 @@ void ShortcutView::clear() {
 	QModelIndex index = m_ui.shortcutTable->selectionModel()->currentIndex();
 	QString name = m_model->name(index);
 	const Shortcut* item = m_controller->shortcut(name);
-	if (!item->action()) {
+	if (!item || !item->action()) {
 		return;
 	}
 	if (m_ui.gamepadButton->isChecked()) {
@@ -106,7 +107,7 @@ void ShortcutView::updateButton(int button) {
 	}
 	QString name = m_model->name(m_ui.shortcutTable->selectionModel()->currentIndex());
 	const Shortcut* item = m_controller->shortcut(name);
-	if (!item->action()) {
+	if (!item || !item->action()) {
 		return;
 	}
 	if (m_ui.gamepadButton->isChecked()) {
@@ -122,7 +123,7 @@ void ShortcutView::updateAxis(int axis, int direction) {
 	}
 	QString name = m_model->name(m_ui.shortcutTable->selectionModel()->currentIndex());
 	const Shortcut* item = m_controller->shortcut(name);
-	if (!item->action()) {
+	if (!item || !item->action()) {
 		return;
 	}
 	m_controller->updateAxis(name, axis, static_cast<GamepadAxisEvent::Direction>(direction));
@@ -134,6 +135,15 @@ void ShortcutView::closeEvent(QCloseEvent*) {
 	}
 }
 
+void ShortcutView::showEvent(QShowEvent*) {
+	QString longString("Ctrl+Alt+Shift+Tab");
+	int width = QFontMetrics(QFont()).width(longString);
+	QHeaderView* header = m_ui.shortcutTable->header();
+	header->resizeSection(0, header->length() - width * 2);
+	header->resizeSection(1, width);
+	header->resizeSection(2, width);
+}
+
 bool ShortcutView::event(QEvent* event) {
 	if (m_input) {
 		QEvent::Type type = event->type();
