@@ -1,5 +1,5 @@
 /* Copyright (c) 2014-2017 waddlesplash
- * Copyright (c) 2014-2020 Jeffrey Pfau
+ * Copyright (c) 2014-2021 Jeffrey Pfau
  *
  * This Source Code Form is subject to the terms of the Mozilla Public
  * License, v. 2.0. If a copy of the MPL was not distributed with this
@@ -8,8 +8,9 @@
 
 #include <memory>
 
+#include <QAtomicInteger>
+#include <QHash>
 #include <QList>
-#include <QMap>
 #include <QStackedWidget>
 
 #include <mgba/core/library.h>
@@ -28,40 +29,18 @@ enum class LibraryStyle {
 	STYLE_ICON
 };
 
-class LibraryEntry final {
-public:
-	LibraryEntry(mLibraryEntry* entry);
-
-	QString displayTitle() const { return title().isNull() ? filename() : title(); }
-
-	QString base() const { return QString(entry->base); }
-	QString filename() const { return QString(entry->filename); }
-	QString fullpath() const { return m_fullpath; }
-	QString title() const { return QString(entry->title); }
-	QByteArray internalTitle() const { return QByteArray(entry->internalTitle); }
-	QByteArray internalCode() const { return QByteArray(entry->internalCode); }
-	mPlatform platform() const { return entry->platform; }
-	size_t filesize() const { return entry->filesize; }
-	uint32_t crc32() const { return entry->crc32; }
-
-	const mLibraryEntry* entry;
-private:
-	const QString m_fullpath;
-};
-typedef std::shared_ptr<LibraryEntry> LibraryEntryRef;
-
 class AbstractGameList {
 public:
-	virtual LibraryEntryRef selectedEntry() = 0;
-	virtual void selectEntry(LibraryEntryRef game) = 0;
+	virtual mLibraryEntry* selectedEntry() = 0;
+	virtual void selectEntry(mLibraryEntry* game) = 0;
 
 	virtual void setViewStyle(LibraryStyle newStyle) = 0;
 
-	virtual void addEntry(LibraryEntryRef item) = 0;
-	virtual void addEntries(QList<LibraryEntryRef> items);
+	virtual void addEntry(mLibraryEntry* item) = 0;
+	virtual void addEntries(QList<mLibraryEntry*> items);
 
-	virtual void removeEntry(LibraryEntryRef item) = 0;
-	virtual void removeEntries(QList<LibraryEntryRef> items);
+	virtual void removeEntry(mLibraryEntry* item) = 0;
+	virtual void removeEntries(QList<mLibraryEntry*> items);
 
 	virtual QWidget* widget() = 0;
 };
@@ -77,14 +56,14 @@ public:
 	LibraryStyle viewStyle() const { return m_currentStyle; }
 	void setViewStyle(LibraryStyle newStyle);
 
-	void selectEntry(LibraryEntryRef entry);
-	LibraryEntryRef selectedEntry();
+	void selectEntry(mLibraryEntry* entry);
+	mLibraryEntry* selectedEntry();
 	VFile* selectedVFile();
 	QPair<QString, QString> selectedPath();
 
 	void selectLastBootedGame();
 
-	void addDirectory(const QString& dir);
+	void addDirectory(const QString& dir, bool recursive = true);
 
 public slots:
 	void clear();
@@ -97,14 +76,14 @@ private slots:
 	void refresh();
 
 private:
-	void loadDirectory(const QString&); // Called on separate thread
+	void loadDirectory(const QString&, bool recursive = true); // Called on separate thread
 	void freeLibrary();
 
 	ConfigController* m_config = nullptr;
 	std::shared_ptr<mLibrary> m_library;
-	qint64 m_libraryJob = -1;
+	QAtomicInteger<qint64> m_libraryJob = -1;
 	mLibraryListing m_listing;
-	QMap<QString, LibraryEntryRef> m_entries;
+	QHash<QString, mLibraryEntry*> m_entries;
 
 	LibraryStyle m_currentStyle;
 	AbstractGameList* m_currentList = nullptr;
