@@ -28,17 +28,22 @@
 #include <mgba-util/memory.h>
 #include <mgba-util/vfs.h>
 
+#ifdef WIIDRC
+#include <wiidrc/wiidrc.h>
+#endif
+
 #define GCN1_INPUT 0x47434E31
 #define GCN2_INPUT 0x47434E32
 #define WIIMOTE_INPUT 0x5749494D
 #define CLASSIC_INPUT 0x57494943
+#define DRC_INPUT 0x44524355
 
 #define TEX_W 256
 #define TEX_H 224
 
 #define ANALOG_DEADZONE 0x30
 
-static void _mapKey(struct mInputMap* map, uint32_t binding, int nativeKey, enum GBAKey key) {
+static void _mapKey(struct mInputMap* map, uint32_t binding, int nativeKey, int key) {
 	mInputBindKey(map, binding, __builtin_ctz(nativeKey), key);
 }
 
@@ -119,6 +124,7 @@ static bool sgbCrop = false;
 static int32_t tiltX;
 static int32_t tiltY;
 static int32_t gyroZ;
+static float gyroSensitivity = 1.f;
 static uint32_t retraceCount;
 static uint32_t referenceRetraceCount;
 static bool frameLimiter = true;
@@ -254,6 +260,9 @@ int main(int argc, char* argv[]) {
 	PAD_Init();
 	WPAD_Init();
 	WPAD_SetDataFormat(0, WPAD_FMT_BTNS_ACC_IR);
+#ifdef WIIDRC
+	WiiDRC_Init();
+#endif
 	AUDIO_Init(0);
 	AUDIO_SetDSPSampleRate(AI_SAMPLERATE_48KHZ);
 	AUDIO_RegisterDMACallback(_audioDMA);
@@ -452,6 +461,31 @@ int main(int argc, char* argv[]) {
 				},
 				.nKeys = 32
 			},
+#ifdef WIIDRC
+			{
+				.name = "Wii U GamePad Input",
+				.id = DRC_INPUT,
+				.keyNames = (const char*[]) {
+					0, // Sync
+					"\1\xE",
+					"-",
+					"+",
+					"R",
+					"L",
+					"ZR",
+					"ZL",
+					"Down",
+					"Up",
+					"Right",
+					"Left",
+					"Y",
+					"X",
+					"B",
+					"A",
+				},
+				.nKeys = 16
+			},
+#endif
 			{ .id = 0 }
 		},
 		.configExtra = (struct GUIMenuItem[]) {
@@ -531,8 +565,26 @@ int main(int argc, char* argv[]) {
 				},
 				.nStates = 8
 			},
+			{
+				.title = "Gyroscope sensitivity",
+				.data = "gyroSensitivity",
+				.submenu = 0,
+				.state = 0,
+				.validStates = (const char*[]) {
+					"1x", "1x flipped", "2x", "2x flipped", "1/2x", "1/2x flipped"
+				},
+				.stateMappings = (const struct GUIVariant[]) {
+					GUI_V_F(1.f),
+					GUI_V_F(-1.f),
+					GUI_V_F(2.f),
+					GUI_V_F(-2.f),
+					GUI_V_F(0.5f),
+					GUI_V_F(-0.5f),
+				},
+				.nStates = 6
+			},
 		},
-		.nConfigExtra = 5,
+		.nConfigExtra = 6,
 		.setup = _setup,
 		.teardown = 0,
 		.gameLoaded = _gameLoaded,
@@ -576,6 +628,15 @@ int main(int argc, char* argv[]) {
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_LEFT, GUI_INPUT_LEFT);
 	_mapKey(&runner.params.keyMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_RIGHT, GUI_INPUT_RIGHT);
 
+#ifdef WIIDRC
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_A, GUI_INPUT_SELECT);
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_B, GUI_INPUT_BACK);
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_X, GUI_INPUT_CANCEL);
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_UP, GUI_INPUT_UP);
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_DOWN, GUI_INPUT_DOWN);
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_LEFT, GUI_INPUT_LEFT);
+	_mapKey(&runner.params.keyMap, DRC_INPUT, WIIDRC_BUTTON_RIGHT, GUI_INPUT_RIGHT);
+#endif
 
 	float stretch = 0;
 	if (mCoreConfigGetFloatValue(&runner.config, "stretchWidth", &stretch)) {
@@ -701,29 +762,48 @@ static uint32_t _pollInput(const struct mInputMap* map) {
 	u32 ext = 0;
 	WPAD_Probe(0, &ext);
 
+#ifdef WIIDRC
+	u32 drckeys = 0;
+	if (WiiDRC_ScanPads()) {
+		drckeys = WiiDRC_ButtonsHeld();
+	}
+#endif
+
 	int keys = 0;
 	keys |= mInputMapKeyBits(map, GCN1_INPUT, padkeys, 0);
 	keys |= mInputMapKeyBits(map, GCN2_INPUT, padkeys, 0);
 	keys |= mInputMapKeyBits(map, WIIMOTE_INPUT, wiiPad, 0);
+#ifdef WIIDRC
+	keys |= mInputMapKeyBits(map, DRC_INPUT, drckeys, 0);
+#endif
 	if (ext == WPAD_EXP_CLASSIC) {
 		keys |= mInputMapKeyBits(map, CLASSIC_INPUT, wiiPad, 0);
 	}
 	int x = PAD_StickX(0);
 	int y = PAD_StickY(0);
-	int w_x = WPAD_StickX(0, 0);
-	int w_y = WPAD_StickY(0, 0);
-	if (x < -ANALOG_DEADZONE || w_x < -ANALOG_DEADZONE) {
+	int wX = WPAD_StickX(0, 0);
+	int wY = WPAD_StickY(0, 0);
+	ATTRIBUTE_UNUSED int drcX = 0;
+	ATTRIBUTE_UNUSED int drcY = 0;
+#ifdef WIIDRC
+	if (WiiDRC_Connected()) {
+		drcX = WiiDRC_lStickX();
+		drcY = WiiDRC_lStickY();
+	}
+#endif
+	if (x < -ANALOG_DEADZONE || wX < -ANALOG_DEADZONE || drcX < -ANALOG_DEADZONE) {
 		keys |= 1 << GUI_INPUT_LEFT;
 	}
-	if (x > ANALOG_DEADZONE || w_x > ANALOG_DEADZONE) {
+	if (x > ANALOG_DEADZONE || wX > ANALOG_DEADZONE || drcX > ANALOG_DEADZONE) {
 		keys |= 1 << GUI_INPUT_RIGHT;
 	}
-	if (y < -ANALOG_DEADZONE || w_y < -ANALOG_DEADZONE) {
+	if (y < -ANALOG_DEADZONE || wY < -ANALOG_DEADZONE || drcY < -ANALOG_DEADZONE) {
 		keys |= 1 << GUI_INPUT_DOWN;
 	}
-	if (y > ANALOG_DEADZONE || w_y > ANALOG_DEADZONE) {
+	if (y > ANALOG_DEADZONE || wY > ANALOG_DEADZONE || drcY > ANALOG_DEADZONE) {
 		keys |= 1 << GUI_INPUT_UP;
 	}
+
 	return keys;
 }
 
@@ -803,12 +883,27 @@ void _setup(struct mGUIRunner* runner) {
 	_mapKey(&runner->core->inputMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_FULL_L, GBA_KEY_L);
 	_mapKey(&runner->core->inputMap, CLASSIC_INPUT, WPAD_CLASSIC_BUTTON_FULL_R, GBA_KEY_R);
 
+#ifdef WIIDRC
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_A, GBA_KEY_A);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_B, GBA_KEY_B);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_PLUS, GBA_KEY_START);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_MINUS, GBA_KEY_SELECT);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_UP, GBA_KEY_UP);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_DOWN, GBA_KEY_DOWN);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_LEFT, GBA_KEY_LEFT);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_RIGHT, GBA_KEY_RIGHT);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_L, GBA_KEY_L);
+	_mapKey(&runner->core->inputMap, DRC_INPUT, WIIDRC_BUTTON_R, GBA_KEY_R);
+#endif
+
 	struct mInputAxis desc = { GBA_KEY_RIGHT, GBA_KEY_LEFT, ANALOG_DEADZONE, -ANALOG_DEADZONE };
 	mInputBindAxis(&runner->core->inputMap, GCN1_INPUT, 0, &desc);
 	mInputBindAxis(&runner->core->inputMap, CLASSIC_INPUT, 0, &desc);
+	mInputBindAxis(&runner->core->inputMap, DRC_INPUT, 0, &desc);
 	desc = (struct mInputAxis) { GBA_KEY_UP, GBA_KEY_DOWN, ANALOG_DEADZONE, -ANALOG_DEADZONE };
 	mInputBindAxis(&runner->core->inputMap, GCN1_INPUT, 1, &desc);
 	mInputBindAxis(&runner->core->inputMap, CLASSIC_INPUT, 1, &desc);
+	mInputBindAxis(&runner->core->inputMap, DRC_INPUT, 1, &desc);
 
 	outputBuffer = memalign(32, TEX_W * TEX_H * BYTES_PER_PIXEL);
 	runner->core->setVideoBuffer(runner->core, outputBuffer, TEX_W);
@@ -839,7 +934,7 @@ void _gameUnloaded(struct mGUIRunner* runner) {
 
 void _gameLoaded(struct mGUIRunner* runner) {
 	reconfigureScreen(runner);
-	if (runner->core->platform(runner->core) == PLATFORM_GBA && ((struct GBA*) runner->core->board)->memory.hw.devices & HW_GYRO) {
+	if (runner->core->platform(runner->core) == mPLATFORM_GBA && ((struct GBA*) runner->core->board)->memory.hw.devices & HW_GYRO) {
 		int i;
 		for (i = 0; i < 6; ++i) {
 			u32 result = WPAD_SetMotionPlus(0, 1);
@@ -900,6 +995,7 @@ void _unpaused(struct mGUIRunner* runner) {
 	if (mCoreConfigGetFloatValue(&runner->config, "stretchHeight", &stretch)) {
 		hStretch = fminf(1.0f, fmaxf(0.5f, stretch));
 	}
+	mCoreConfigGetFloatValue(&runner->config, "gyroSensitivity", &gyroSensitivity);
 }
 
 void _prepareForFrame(struct mGUIRunner* runner) {
@@ -1033,9 +1129,18 @@ uint16_t _pollGameInput(struct mGUIRunner* runner) {
 	u32 wiiPad = WPAD_ButtonsHeld(0);
 	u32 ext = 0;
 	WPAD_Probe(0, &ext);
+#ifdef WIIDRC
+	u32 drckeys = 0;
+	if (WiiDRC_ScanPads()) {
+		drckeys = WiiDRC_ButtonsHeld();
+	}
+#endif
 	uint16_t keys = mInputMapKeyBits(&runner->core->inputMap, GCN1_INPUT, padkeys, 0);
 	keys |= mInputMapKeyBits(&runner->core->inputMap, GCN2_INPUT, padkeys, 0);
 	keys |= mInputMapKeyBits(&runner->core->inputMap, WIIMOTE_INPUT, wiiPad, 0);
+#ifdef WIIDRC
+	keys |= mInputMapKeyBits(&runner->core->inputMap, DRC_INPUT, drckeys, 0);
+#endif
 
 	enum GBAKey angles = mInputMapAxis(&runner->core->inputMap, GCN1_INPUT, 0, PAD_StickX(0));
 	if (angles != GBA_KEY_NONE) {
@@ -1056,6 +1161,19 @@ uint16_t _pollGameInput(struct mGUIRunner* runner) {
 			keys |= 1 << angles;
 		}
 	}
+#ifdef WIIDRC
+	if (WiiDRC_Connected()) {
+		keys |= mInputMapKeyBits(&runner->core->inputMap, DRC_INPUT, drckeys, 0);
+		angles = mInputMapAxis(&runner->core->inputMap, DRC_INPUT, 0, WiiDRC_lStickX());
+		if (angles != GBA_KEY_NONE) {
+			keys |= 1 << angles;
+		}
+		angles = mInputMapAxis(&runner->core->inputMap, DRC_INPUT, 1, WiiDRC_lStickY());
+		if (angles != GBA_KEY_NONE) {
+			keys |= 1 << angles;
+		}
+	}
+#endif
 
 	return keys;
 }
@@ -1122,7 +1240,7 @@ int32_t _readTiltY(struct mRotationSource* source) {
 
 int32_t _readGyroZ(struct mRotationSource* source) {
 	UNUSED(source);
-	return gyroZ;
+	return gyroZ * gyroSensitivity;
 }
 
 static s8 WPAD_StickX(u8 chan, u8 right) {
